// 
// Author:
//       smdn <smdn@smdn.jp>
// 
// Copyright (c) 2009-2017 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.ComponentModel;

using Smdn.Collections;
using Smdn.Formats.Thbgm;

namespace Smdn.Windows.Forms.ThbgmPlayer {
  public class Effect : ICloneable, INotifyPropertyChanged {
    public static readonly TimeSpan TimeSpanUIMaxValue = TimeSpan.FromHours(6.0);
    public static readonly IDictionary<FadingCurveType, decimal> FadingCurveCoefficientUIMaxValue = (new Dictionary<FadingCurveType, decimal>() {
      {FadingCurveType.None,            1.0m},
      {FadingCurveType.Linear,          1.0m},
      {FadingCurveType.Exponential,     100.0m},
      {FadingCurveType.Logarithmic,     100.0m},
      {FadingCurveType.Trigonometrical, 1.0m},
    }).AsReadOnly();
    public static readonly IDictionary<FadingCurveType, decimal> FadingCurveCoefficientUIMinValue = (new Dictionary<FadingCurveType, decimal>() {
      {FadingCurveType.None,            0.0m},
      {FadingCurveType.Linear,          0.0m},
      {FadingCurveType.Exponential,     1.0m},
      {FadingCurveType.Logarithmic,     1.0m},
      {FadingCurveType.Trigonometrical, 0.0m},
    }).AsReadOnly();
    public static readonly IDictionary<FadingCurveType, decimal> FadingCurveCoefficientUIIncrementValue = (new Dictionary<FadingCurveType, decimal>() {
      {FadingCurveType.None,            0.1m},
      {FadingCurveType.Linear,          0.1m},
      {FadingCurveType.Exponential,     0.25m},
      {FadingCurveType.Logarithmic,     0.25m},
      {FadingCurveType.Trigonometrical, 0.05m},
    }).AsReadOnly();

    public static string TimeSpanToStringConverter(TimeSpan val)
    {
      return string.Format("{0}:{1:D2}:{2:D2}.{3:D3}", (int)val.TotalHours, val.Minutes, val.Seconds, val.Milliseconds);
    }

    public static TimeSpan StringToTimeSpanConverter(string val)
    {
      var values = val.Split(new[] {System.Globalization.CultureInfo.CurrentCulture.DateTimeFormat.TimeSeparator}, StringSplitOptions.None);

      if (values.Length <= 0)
        throw new FormatException();

      var secs = double.Parse(values[values.Length - 1]);

      switch (values.Length) {
        case 1: { // sss.fff
          return TimeSpan.FromSeconds(secs);
        }

        case 2: { // mm:ss.fff
          var mins = (double)int.Parse(values[0]);
          return TimeSpan.FromSeconds(mins * 60.0 + secs);
        }

        case 3: { // hh:mm:ss.ff
          var hours = (double)int.Parse(values[0]);
          var mins  = (double)int.Parse(values[1]);
          return TimeSpan.FromSeconds(hours * 3600.0 + mins * 60.0 + secs);
        }

        default:
          throw new FormatException();
      }
    }

    public static double CalculateVolumeExponential(double volume, double @base)
    {
      // f(x) = (n^x - 1.0) / (n - 1.0)
      return (@base <= 1.0) ? volume : (Math.Pow(@base, volume) - 1.0) / (@base - 1.0);
    }

    public static double CalculateVolumeLogarithmic(double volume, double @base)
    {
      // f(x) = log_n(x * (n - 1.0) + 1.0)
      return (@base <= 1.0) ? volume : Math.Log(volume * (@base - 1.0) + 1.0, @base);
    }

    public static double CalculateVolumeTrigonometical(double volume, double exponent)
    {
      // f(x) = (1.0 - (+cos(x * π))^n) / 2.0 where 0 <= x < 0.5
      //      = (1.0 + (-cos(x * π))^n) / 2.0 where 0.5 <= x <= 1.0
      volume = Math.Cos(Math.PI * volume);
      volume = (0.0 < volume) ? Math.Pow(volume, exponent) : -Math.Pow(-volume, exponent);
      return (1.0 - volume) * 0.5;
    }

    /*
     * instance members
     */
    public event PropertyChangedEventHandler PropertyChanged;

    public TimeSpan FadingStart {
      get { return fadingStart; }
      set
      {
        if (value < TimeSpan.Zero)
          throw new ArgumentOutOfRangeException("FadingStart", "must be zero or positive value");

        if (fadingStart == value)
          return;

        fadingStart = value;

        OnPropertyChanged(new PropertyChangedEventArgs("FadingStart"));
      }
    }

    public TimeSpan FadingLength {
      get { return fadingLength; }
      set
      {
        if (value < TimeSpan.Zero)
          throw new ArgumentOutOfRangeException("FadingLength", "must be zero or positive value");

        if (fadingLength == value)
          return;

        fadingLength = value;

        OnPropertyChanged(new PropertyChangedEventArgs("FadingLength"));
      }
    }

    public TimeSpan TrailingSilenceLength {
      get { return trailingSilenceLength; }
      set
      {
        if (value < TimeSpan.Zero)
          throw new ArgumentOutOfRangeException("TrailingSilenceLength", "must be zero or positive value");

        if (trailingSilenceLength == value)
          return;

        trailingSilenceLength = value;

        OnPropertyChanged(new PropertyChangedEventArgs("TrailingSilenceLength"));
      }
    }

    public FadingCurveType FadingCurveType {
      get { return fadingCurveType; }
      set
      {
        if (fadingCurveType == value)
          return;

        fadingCurveType = value;

        OnPropertyChanged(new PropertyChangedEventArgs("FadingCurveType"));
      }
    }

    public double FadingCurveCoefficient {
      get { return fadingCurveCoefficient; }
      set
      {
        if (fadingCurveCoefficient == value)
          return;

        fadingCurveCoefficient = value;

        OnPropertyChanged(new PropertyChangedEventArgs("FadingCurveCoefficient"));
      }
    }

    public TimeSpan TotalLength {
      get { return fadingStart + fadingLength + trailingSilenceLength; }
    }

    // 2 repeats, 20sec fading, 0sec silence
    public static readonly int DefaultTimesToRepeat = 2;
    public static readonly TimeSpan DefaultFadingLength = TimeSpan.FromSeconds(20.0);
    public static readonly TimeSpan DefaultFadingOffset = -DefaultFadingLength;
    public static readonly TimeSpan DefaultTrailingSilenceLength = TimeSpan.Zero;
    public static readonly FadingCurveType DefaultFadingCurveType = FadingCurveType.Linear;

    public static Effect CreateDefault(TrackInfo track)
    {
      if (track == null)
        throw new ArgumentNullException("track");

      var effect = new Effect();

      effect.SetLength(track, DefaultTimesToRepeat, DefaultFadingLength, DefaultFadingOffset, DefaultTrailingSilenceLength);

      // linear curve fading
      effect.FadingCurveType = DefaultFadingCurveType;

      effect.FadingCurveCoefficient = 1.0;

      return effect;
    }

    object ICloneable.Clone()
    {
      return Clone();
    }

    public virtual Effect Clone()
    {
      var cloned = (Effect)MemberwiseClone();

      cloned.PropertyChanged = null;

      return cloned;
    }

    public void SetLength(TimeSpan fadingStart, TimeSpan fadingLength, TimeSpan trailingSilenceLength)
    {
      FadingStart = fadingStart;
      FadingLength = fadingLength;
      TrailingSilenceLength = trailingSilenceLength;
    }

    public void SetTotalLength(TimeSpan totalLength, TimeSpan fadingLength, TimeSpan trailingSilenceLength)
    {
      var start = totalLength - fadingLength - trailingSilenceLength;

      if (start < TimeSpan.Zero)
        start = TimeSpan.Zero;

      FadingStart = start;
      FadingLength = fadingLength;
      TrailingSilenceLength = trailingSilenceLength;
    }

    public void SetTimesToRepeat(TrackInfo track, int timesToRepeat)
    {
      var fadingOffset = fadingStart - track.CalculateRepeatPosition(track.CalculateRepeatedTimes(TotalLength));

      SetLength(track, timesToRepeat, fadingLength, fadingOffset, trailingSilenceLength);
    }

    public void SetLength(TrackInfo track, int timesToRepeat, TimeSpan fadingLength, TimeSpan fadingOffset, TimeSpan trailingSilenceLength)
    {
      var start = track.CalculateRepeatPosition(timesToRepeat) + fadingOffset;

      if (start < TimeSpan.Zero)
        start = TimeSpan.Zero;

      FadingStart = start;
      FadingLength = fadingLength;
      TrailingSilenceLength = trailingSilenceLength;
    }

    public void SetFadingCurve(FadingCurveType newCurveType, bool roundCoefficientToAppropriateRange)
    {
      SetFadingCurve(newCurveType, FadingCurveCoefficient, roundCoefficientToAppropriateRange);
    }

    public void SetFadingCurve(double newCurveCoefficient, bool roundCoefficientToAppropriateRange)
    {
      SetFadingCurve(FadingCurveType, newCurveCoefficient, roundCoefficientToAppropriateRange);
    }

    public void SetFadingCurve(FadingCurveType newCurveType, double newCurveCoefficient, bool roundCoefficientToAppropriateRange)
    {
      FadingCurveType = newCurveType;

      if (!roundCoefficientToAppropriateRange)
        return;

      switch (newCurveType) {
        case FadingCurveType.Exponential:
        case FadingCurveType.Logarithmic:
        case FadingCurveType.Trigonometrical:
          if (double.IsNaN(newCurveCoefficient))
            newCurveCoefficient = double.MinValue;
          else if (double.IsNegativeInfinity(newCurveCoefficient))
            newCurveCoefficient = double.MinValue;
          else if (double.IsPositiveInfinity(newCurveCoefficient))
            newCurveCoefficient = double.MaxValue;

          if ((double)FadingCurveCoefficientUIMaxValue[newCurveType] < newCurveCoefficient)
            FadingCurveCoefficient = (double)FadingCurveCoefficientUIMaxValue[newCurveType];
          else if ((double)FadingCurveCoefficientUIMinValue[newCurveType] > newCurveCoefficient)
            FadingCurveCoefficient = (double)FadingCurveCoefficientUIMinValue[newCurveType];
          else
            FadingCurveCoefficient = newCurveCoefficient;
          break;

        default:
          break; // do nothing
      }
    }

    protected virtual void OnPropertyChanged(PropertyChangedEventArgs e)
    {
      var ev = this.PropertyChanged;

      if (ev != null)
        ev(this, e);
    }

    private TimeSpan fadingStart;
    private TimeSpan fadingLength;
    private TimeSpan trailingSilenceLength;
    private FadingCurveType fadingCurveType;
    private double fadingCurveCoefficient = 1.0;
  }
}
