// 
// Author:
//       smdn <smdn@smdn.jp>
// 
// Copyright (c) 2009-2017 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.ComponentModel;
using System.Linq;
using System.Windows.Forms;

using Smdn.Collections;
using Smdn.Formats.Thbgm;

namespace Smdn.Windows.Forms.ThbgmPlayer {
  internal interface ITrackListView {
    void UpdateAllTracksItemSubItems();
  }

  public class TrackListView : TrackListView<TrackListViewItem> {
  }

  public class TrackListView<TTrackListViewItem> : EditableListView, ITrackListView where TTrackListViewItem : TrackListViewItem, new() {
    public event CancelEventHandler ProductChanging;
    public event EventHandler ProductChanged;

    public ProductInfo Product {
      get { return product; }
      set
      {
        if (product == value)
          return;

        var e = new CancelEventArgs();

        OnProductChanging(e);

        if (e.Cancel)
          return;

        product = value;

        OnProductChanged();
      }
    }

    public Track SelectedTrack {
      get
      {
        if (SelectedItems.Count <= 0)
          return null;
        else
          return (SelectedItems[0] as TrackListViewItem).Track;
      }
    }

    public ReadOnlyCollection<Track> Tracks {
      get { return tracks; }
    }

    public ReadOnlyCollection<TTrackListViewItem> TrackItems {
      get { return trackItems.AsReadOnly(); }
    }

    protected ListViewItem AllTracksItem {
      get { return allTracksItem; }
    }

    public TrackListView()
      : base()
    {
      CheckBoxes = true;
      View = View.Details;
      HeaderStyle = ColumnHeaderStyle.Nonclickable;
      MultiSelect = false;
      FullRowSelect = true;
      HideSelection = false;

      Columns.Add(new TrackListViewColumnHeader("TrackNumber", Texts.GetText("TrackNumber"), 45));
      Columns.Add(new TrackListViewColumnHeader("Title", Texts.GetText("Title"), 260));
#if DEBUG
      Columns.Add(new TrackListViewColumnHeader("IsDirty", "IsDirty", 20));
#endif
      Columns.Add(new TrackListViewColumnHeader("TotalLength", Texts.GetText("TotalLength"), 100));
      Columns.Add(new TrackListViewColumnHeader("RepeatedTimes", Texts.GetText("RepeatedTimes"), 70));
      Columns.Add(new TrackListViewColumnHeader("FadingStart", Texts.GetText("FadingStart"), 100));
      Columns.Add(new TrackListViewColumnHeader("FadingLength", Texts.GetText("FadingLength"), 140));
      Columns.Add(new TrackListViewColumnHeader("FadingCurveType", Texts.GetText("FadingCurveType"), 120));
      Columns.Add(new TrackListViewColumnHeader("FadingCurveCoefficient", Texts.GetText("FadingCurveCoefficient"), 45));
      Columns.Add(new TrackListViewColumnHeader("TrailingSilenceLength", Texts.GetText("TrailingSilenceLength"), 100));

      allTracksItem = new ListViewItem();
    }

    public virtual /* virtual required, Mono BUG? */ void UpdateAllTracksItemSubItems()
    {
      if (allTracksItem.SubItems.Count != Columns.Count) {
        allTracksItem.SubItems.Clear();

        for (var i = 0; i < Columns.Count; i++) {
          allTracksItem.SubItems.Add(string.Empty);
        }
      }

      for (var i = 0; i < Columns.Count; i++) {
        allTracksItem.SubItems[i].Text = GetAllTracksSubItemText(Columns[i] as TrackListViewColumnHeader);
      }
    }

    protected virtual string GetAllTracksSubItemText(TrackListViewColumnHeader header)
    {
      switch (header.BindingValueName) {
        case "TrackNumber":
          return Texts.GetText("AllTracks");

        case "Title":
          return product == null ? string.Empty : product.Title;

        case "IsDirty":
          return string.Empty;

        case "TotalLength": {
          var totalLength = TimeSpan.Zero;

          foreach (var item in trackItems) {
            if (item.Track != null && item.Track.Checked)  // can't refer item.Checked; will cause ArgumentOutOfRangeException on init
              totalLength += item.Track.Effect.TotalLength;
          }

          return TrackListViewItem.ConvertTimeSpanToString(totalLength);
        }
      }

      if (trackItems.Count == 0)
        return string.Empty;

      var firstTrack = trackItems[0].Track;

      if (firstTrack == null)
        return string.Empty;

      switch (header.BindingValueName) {
        case "RepeatedTimes": {
          var repeatedTimes = firstTrack.TrackInfo.CalculateRepeatedTimes(firstTrack.Effect.TotalLength);

          if (trackItems.TrueForAll(delegate(TTrackListViewItem item) {
            return repeatedTimes == item.Track.TrackInfo.CalculateRepeatedTimes(item.Track.Effect.TotalLength);
          }))
            return repeatedTimes.ToString();
          else
            return Texts.GetText("MultipleValues");
        }

        case "FadingStart":
          if (trackItems.TrueForAll(delegate(TTrackListViewItem item) {
            return firstTrack.Effect.FadingStart == item.Track.Effect.FadingStart;
          }))
            return TrackListViewItem.ConvertTimeSpanToString(firstTrack.Effect.FadingStart);
          else
            return Texts.GetText("MultipleValues");

        case "FadingLength":
          if (trackItems.TrueForAll(delegate(TTrackListViewItem item) {
            return firstTrack.Effect.FadingLength == item.Track.Effect.FadingLength;
          }))
            return TrackListViewItem.ConvertTimeSpanToString(firstTrack.Effect.FadingLength);
          else
            return Texts.GetText("MultipleValues");

        case "FadingCurveType":
          if (trackItems.TrueForAll(delegate(TTrackListViewItem item) {
            return firstTrack.Effect.FadingCurveType == item.Track.Effect.FadingCurveType;
          }))
            return Texts.GetText(firstTrack.Effect.FadingCurveType);
          else
            return Texts.GetText("MultipleValues");

        case "FadingCurveCoefficient":
          if (trackItems.TrueForAll(delegate(TTrackListViewItem item) {
            return firstTrack.Effect.FadingCurveCoefficient == item.Track.Effect.FadingCurveCoefficient;
          }))
            return firstTrack.Effect.FadingCurveCoefficient.ToString("F2");
          else
            return Texts.GetText("MultipleValues");

        case "TrailingSilenceLength":
          if (trackItems.TrueForAll(delegate(TTrackListViewItem item) {
            return firstTrack.Effect.TrailingSilenceLength == item.Track.Effect.TrailingSilenceLength;
          }))
            return TrackListViewItem.ConvertTimeSpanToString(firstTrack.Effect.TrailingSilenceLength);
          else
            return Texts.GetText("MultipleValues");

        default:
          return string.Empty;
      }
    }

    public void SetProductAndTracks(ProductInfo product, ReadOnlyCollection<Track> tracks)
    {
      SetTracks(tracks);

      if (this.product != product) {
        this.product = product;

        // raise event
        OnProductChanged(EventArgs.Empty);
      }
    }

    public void CheckAll()
    {
      CheckAll(true);
    }

    public void UncheckAll()
    {
      CheckAll(false);
    }

    private void CheckAll(bool check)
    {
      itemsUpdating = true;

      allTracksItem.Checked = !check; // uncheck all items first to invoke OnItemChecked()

      itemsUpdating = false;

      allTracksItem.Checked = check; // now, check all items
    }

    private void SetTracks(ReadOnlyCollection<Track> tracks)
    {
      this.tracks = tracks;

      itemsUpdating = true;

      BeginUpdate();

      trackItems.Clear();

      Items.Clear();

      foreach (var track in tracks) {
        var item = new TTrackListViewItem();

        trackItems.Add(item);

        Items.Add(item);

        item.Track = track;
        item.Checked = track.Checked;
      }

      Items.Add(allTracksItem);

      UpdateAllTracksItemSubItems();

      EndUpdate();

      itemsUpdating = false;
    }

    protected override void OnItemChecked(ItemCheckedEventArgs e)
    {
      if (e.Item == allTracksItem) {
        // check/uncheck all
        if (!itemsUpdating) {
          foreach (var item in trackItems) {
            item.Checked = allTracksItem.Checked;
          }
        }
      }
      else {
        var item = (e.Item as TTrackListViewItem);

        if (item.Track != null) {
          item.Track.Checked = item.Checked;

          if (!itemsUpdating) {
            // edit by user operation
            item.Track.SetDirty();

            UpdateAllTracksItemSubItems();
          }
        }
      }

      base.OnItemChecked(e);
    }

    protected override void OnItemSelectionChanged(ListViewItemSelectionChangedEventArgs e)
    {
      if (e.Item == allTracksItem && e.IsSelected)
        e.Item.Selected = false;

      base.OnItemSelectionChanged(e);
    }

    private void OnProductChanged()
    {
      if (product == null)
        SetTracks(emptyTrackCollection);
      else
        SetTracks(product.Tracks.Select(delegate(TrackInfo trackInfo) {
          return new Track(trackInfo);
        }).ToList().AsReadOnly());

      allTracksItem.Checked = true;

      OnProductChanged(EventArgs.Empty);
    }

    protected virtual void OnProductChanging(CancelEventArgs e)
    {
      var ev = this.ProductChanging;

      if (ev != null)
        ev(this, e);
    }

    protected virtual void OnProductChanged(EventArgs e)
    {
      var ev = this.ProductChanged;

      if (ev != null)
        ev(this, e);
    }

    protected override ListViewSubItemEditorBase CreateSubItemEditor(ListViewItem item, ListViewItem.ListViewSubItem subItem)
    {
      return CreateSubItemEditor(item as TrackListViewItem,
                                 Columns[item.SubItems.IndexOf(subItem)] as TrackListViewColumnHeader,
                                 subItem);
    }

    protected class TimeSpanUpDownListViewSubItemEditor : GenericListViewSubItemEditor<TimeSpanUpDown> {
      public Track Track {
        get { return (Item as TrackListViewItem).Track; }
      }

      public TimeSpanUpDownListViewSubItemEditor(ListViewItem item, ListViewItem.ListViewSubItem subItem, EventHandler onEditEntered)
        : base(item, subItem)
      {
        EditEntered += onEditEntered;
        EditLeave += onEditEntered;
      }

      protected override Control BeginEdit()
      {
        EditControl.SelectAll();

        return EditControl;
      }
    }

    protected class NumericUpDownListViewSubItemEditor : GenericListViewSubItemEditor<NumericUpDown> {
      public Track Track {
        get { return (Item as TrackListViewItem).Track; }
      }

      public NumericUpDownListViewSubItemEditor(ListViewItem item, ListViewItem.ListViewSubItem subItem, EventHandler onEditEntered)
        : base(item, subItem)
      {
        EditEntered += onEditEntered;
        EditLeave += onEditEntered;
      }

      protected override Control BeginEdit()
      {
        if (EditControl.Text != null)
          EditControl.Select(0, EditControl.Text.Length);

        return EditControl;
      }
    }

    protected class FadingCurveTypeListViewSubItemEditor : ListControlListViewSubItemEditor<FadingCurveTypeDropDownList> {
      public Track Track {
        get { return (Item as TrackListViewItem).Track; }
      }

      public FadingCurveTypeListViewSubItemEditor(ListViewItem item, ListViewItem.ListViewSubItem subItem, EventHandler onEditEntered)
        : base(item, subItem)
      {
        EditEntered += onEditEntered;
        EditLeave += onEditEntered;
      }
    }

    protected virtual ListViewSubItemEditorBase CreateSubItemEditor(TrackListViewItem item, TrackListViewColumnHeader header, ListViewItem.ListViewSubItem subItem)
    {
      var track = item.Track;

      switch (header.BindingValueName) {
        case "TotalLength": {
          var editor = new TimeSpanUpDownListViewSubItemEditor(item, subItem, delegate(object sender, EventArgs e) {
            var editControl = (sender as TimeSpanUpDownListViewSubItemEditor).EditControl;
            var t = (sender as TimeSpanUpDownListViewSubItemEditor).Track;

            t.Effect.SetTotalLength(editControl.Value, t.Effect.FadingLength, t.Effect.TrailingSilenceLength);
          });

          editor.EditControl.Value = track.Effect.TotalLength;

          SetEditControlDefaults(header.BindingValueName, editor.EditControl);

          return editor;
        }

        case "RepeatedTimes": {
          var editor = new NumericUpDownListViewSubItemEditor(item, subItem, delegate(object sender, EventArgs e) {
            var editControl = (sender as NumericUpDownListViewSubItemEditor).EditControl;
            var t = (sender as NumericUpDownListViewSubItemEditor).Track;

            t.Effect.SetTimesToRepeat(t.TrackInfo, (int)editControl.Value);
          });

          editor.EditControl.Value = (decimal)track.TrackInfo.CalculateRepeatedTimes(track.Effect.TotalLength);
          editor.EditControl.Increment = 1.0m;
          editor.EditControl.Minimum = 0.0m;
          editor.EditControl.Maximum = 100.0m;

          return editor;
        }

        case "FadingStart": {
          var editor = new TimeSpanUpDownListViewSubItemEditor(item, subItem, delegate(object sender, EventArgs e) {
            var editControl = (sender as TimeSpanUpDownListViewSubItemEditor).EditControl;
            var t = (sender as TimeSpanUpDownListViewSubItemEditor).Track;

            t.Effect.FadingStart = editControl.Value;
          });

          editor.EditControl.Value = track.Effect.FadingStart;

          SetEditControlDefaults(header.BindingValueName, editor.EditControl);

          return editor;
        }

        case "FadingLength": {
          var editor = new TimeSpanUpDownListViewSubItemEditor(item, subItem, delegate(object sender, EventArgs e) {
            var editControl = (sender as TimeSpanUpDownListViewSubItemEditor).EditControl;
            var t = (sender as TimeSpanUpDownListViewSubItemEditor).Track;

            t.Effect.FadingLength = editControl.Value;
          });

          editor.EditControl.Value = track.Effect.FadingLength;

          SetEditControlDefaults(header.BindingValueName, editor.EditControl);

          return editor;
        }

        case "FadingCurveType": {
          var editor = new FadingCurveTypeListViewSubItemEditor(item, subItem, delegate(object sender, EventArgs e) {
            var editControl = (sender as FadingCurveTypeListViewSubItemEditor).EditControl;
            var t = (sender as FadingCurveTypeListViewSubItemEditor).Track;

            t.Effect.SetFadingCurve(editControl.SelectedItem, true);
          });

          editor.EditControl.SelectedItem = track.Effect.FadingCurveType;

          return editor;
        }

        case "FadingCurveCoefficient": {
          switch (track.Effect.FadingCurveType) {
            case FadingCurveType.Exponential:
            case FadingCurveType.Logarithmic:
            case FadingCurveType.Trigonometrical:
              break;

            case FadingCurveType.None:
            case FadingCurveType.Linear:
            default:
              return null; // not editable
          }

          var editor = new NumericUpDownListViewSubItemEditor(item, subItem, delegate(object sender, EventArgs e) {
            var editControl = (sender as NumericUpDownListViewSubItemEditor).EditControl;
            var t = (sender as NumericUpDownListViewSubItemEditor).Track;

            t.Effect.SetFadingCurve((double)editControl.Value, true);
          });

          editor.EditControl.Value = (decimal)track.Effect.FadingCurveCoefficient;
          editor.EditControl.Increment  = Effect.FadingCurveCoefficientUIIncrementValue[track.Effect.FadingCurveType];
          editor.EditControl.Minimum    = Effect.FadingCurveCoefficientUIMinValue[track.Effect.FadingCurveType];
          editor.EditControl.Maximum    = Effect.FadingCurveCoefficientUIMaxValue[track.Effect.FadingCurveType];
          editor.EditControl.DecimalPlaces = 2;

          return editor;
        }

        case "TrailingSilenceLength": {
          var editor = new TimeSpanUpDownListViewSubItemEditor(item, subItem, delegate(object sender, EventArgs e) {
            var editControl = (sender as TimeSpanUpDownListViewSubItemEditor).EditControl;
            var t = (sender as TimeSpanUpDownListViewSubItemEditor).Track;

            t.Effect.TrailingSilenceLength = editControl.Value;
          });

          editor.EditControl.Value = track.Effect.TrailingSilenceLength;

          SetEditControlDefaults(header.BindingValueName, editor.EditControl);

          return editor;
        }

        case "TrackNumber":
        case "Title":
        case "IsDirty":
        default:
          return null;
      }
    }

    private void SetEditControlDefaults(string valueName, TimeSpanUpDown timeSpanUpDown)
    {
      timeSpanUpDown.TimeSpanToStringConverter = Effect.TimeSpanToStringConverter;
      timeSpanUpDown.StringToTimeSpanConverter = Effect.StringToTimeSpanConverter;
      timeSpanUpDown.Minimum = TimeSpan.Zero;
      timeSpanUpDown.Maximum = Effect.TimeSpanUIMaxValue;

      switch (valueName) {
        case "FadingStart":
          timeSpanUpDown.Increment = TimeSpan.FromSeconds(10.0);
          break;
        case "FadingLength":
          timeSpanUpDown.Increment = TimeSpan.FromSeconds(0.2);
          break;
        case "FadingOffset":
          timeSpanUpDown.Increment = TimeSpan.FromSeconds(0.2);
          break;
        case "TrailingSilenceLength":
          timeSpanUpDown.Increment = TimeSpan.FromSeconds(0.5);
          break;
        case "TotalLength":
          timeSpanUpDown.Increment = TimeSpan.FromSeconds(10.0);
          break;
      }
    }

    private ProductInfo product = null;
    private static readonly ReadOnlyCollection<Track> emptyTrackCollection = Array.AsReadOnly(EmptyArray<Track>.Instance);
    private ReadOnlyCollection<Track> tracks = emptyTrackCollection;
    private readonly List<TTrackListViewItem> trackItems = new List<TTrackListViewItem>();
    private bool itemsUpdating = false;
    private ListViewItem allTracksItem;
  }
}
