// 
// Author:
//       smdn <smdn@smdn.jp>
// 
// Copyright (c) 2009-2014 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.ComponentModel;
using System.Windows.Forms;

namespace Smdn.Windows.Forms.ThbgmPlayer {
  public class TrackListViewItem : ListViewItem {
    public Track Track {
      get { return track; }
      internal set
      {
        if (value == null)
          throw new ArgumentNullException();

        if (track != null)
          track.Effect.PropertyChanged -= Effect_PropertyChanged;

        track = value;
        track.Effect.PropertyChanged += Effect_PropertyChanged;

        UpdateSubItems();
      }
    }

    public TrackListViewItem()
    {
    }

    public TrackListViewItem(Track track)
    {
      if (track == null)
        throw new ArgumentNullException("track");

      Track = track;
    }

    protected virtual void UpdateSubItems()
    {
      if (SubItems.Count != ListView.Columns.Count) {
        SubItems.Clear();

        for (var i = 0; i < ListView.Columns.Count; i++) {
          SubItems.Add(string.Empty);
        }
      }

      for (var i = 0; i < ListView.Columns.Count; i++) {
        SubItems[i].Text = GetSubItemText(ListView.Columns[i] as TrackListViewColumnHeader);
      }

      Checked = Track.Checked;
    }

    protected virtual string GetSubItemText(TrackListViewColumnHeader header)
    {
      switch (header.BindingValueName) {
        case "TrackNumber":             return Track.TrackInfo.TrackNumber.ToString();
        case "Title":                   return Track.TrackInfo.Title;
        case "IsDirty":                 return Track.IsDirty.ToString();
        case "TotalLength":             return ConvertTimeSpanToString(Track.Effect.TotalLength);
        case "RepeatedTimes":           return Track.TrackInfo.CalculateRepeatedTimes(Track.Effect.TotalLength).ToString();
        case "FadingStart":             return ConvertTimeSpanToString(Track.Effect.FadingStart);
        case "FadingLength":            return ConvertTimeSpanToString(Track.Effect.FadingLength);
        case "FadingCurveType":         return Texts.GetText(Track.Effect.FadingCurveType);
        case "FadingCurveCoefficient": {
          switch (Track.Effect.FadingCurveType) {
            case FadingCurveType.Exponential:
            case FadingCurveType.Logarithmic:
            case FadingCurveType.Trigonometrical:
              return Track.Effect.FadingCurveCoefficient.ToString("F2");
            default:
              return "-";
          }
        }
        case "TrailingSilenceLength":   return ConvertTimeSpanToString(Track.Effect.TrailingSilenceLength);
        default: return string.Empty;
      }
    }

    protected internal static string ConvertTimeSpanToString(TimeSpan t)
    {
      return string.Format("{0}:{1:D2}.{2:D3}", (int)t.TotalMinutes, t.Seconds, t.Milliseconds);
    }

    private void Effect_PropertyChanged(object sender, PropertyChangedEventArgs e)
    {
      if (sender == track.Effect)
        OnEffectPropertyChanged(e);
    }

    protected virtual void OnEffectPropertyChanged(PropertyChangedEventArgs e)
    {
      UpdateSubItems();

      if (ListView is ITrackListView)
        (ListView as ITrackListView).UpdateAllTracksItemSubItems();
    }

    private Track track;
  }
}
