// 
// Author:
//       smdn <smdn@smdn.jp>
// 
// Copyright (c) 2009-2017 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

//#if !DEBUG
#define TRANSPARENT_ANCHOR
//#endif

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Drawing.Imaging;
using System.Drawing.Text;
using System.IO;
using System.Threading;
using System.Threading.Tasks;
using System.Windows.Forms;

using Smdn.Formats.Thbgm;

#if DEBUG
using System.Diagnostics;
#endif

namespace Smdn.Windows.Forms.ThbgmPlayer {
  public class WaveFormView : Control {
    private class WaveForm : IDisposable {
      public const int WaveFormPartLengthInMilliseconds = 25;

      public event EventHandler Refreshing;
      public event EventHandler Refreshed;
      public event EventHandler Modified;

      public static WaveForm Create(string bgmSourcePath, Track track)
      {
        if (track == null)
          throw new ArgumentNullException("track");

        if (string.IsNullOrEmpty(bgmSourcePath) || !File.Exists(bgmSourcePath))
          // construct with effect only
          return new WaveForm(track.Effect);

        try {
          return new WaveForm(track.GetStream(bgmSourcePath, true, track.TrackInfo.StreamFormat.BlockAlign));
        }
        catch /* ignore any exceptions */ {
          // construct with effect only
          return new WaveForm(track.Effect);
        }
      }

      private WaveForm(EffectAppliedBgmStream stream)
        : this(stream.Effect)
      {
        this.stream = stream;
        this.samplesPerWaveFormPart = (stream.Format.SamplesPerSecond * WaveFormPartLengthInMilliseconds) / 1000;
      }

      private WaveForm(Effect effect)
      {
        if (effect == null)
          throw new ArgumentNullException("effect");

        this.effect = effect;
        this.effect.PropertyChanged += Effect_PropertyChanged;
        this.effectPreviousFadingStartTime = effect.FadingStart;
        this.samplesPerWaveFormPart = 0;
        this.totalWaveFormLength = 0;
        this.availableWaveFormLength = 0;
        this.waveFormLeft  = EmptyArray<float>.Instance;
        this.waveFormRight = EmptyArray<float>.Instance;
      }

      public void Dispose()
      {
        Refreshed = null;
        Modified = null;

        AbortAndWait();

        if (effect != null) {
          effect.PropertyChanged -= Effect_PropertyChanged;
          effect = null;
        }

        if (stream != null) {
          stream.Close();
          stream = null;
        }

        availableWaveFormLength = 0;

        waveFormLeft = null;
        waveFormRight = null;
      }

      private void Effect_PropertyChanged(object sender, PropertyChangedEventArgs e)
      {
        if (effect == null)
          return; // disposed

        TimeSpan modifyingPoint;

        if (string.Equals(e.PropertyName, "FadingStart", StringComparison.Ordinal)) {
          var fadingStart = (sender as Effect).FadingStart;

          modifyingPoint = (fadingStart < effectPreviousFadingStartTime) ? fadingStart : effectPreviousFadingStartTime;

          effectPreviousFadingStartTime = fadingStart;
        }
        else if (string.Equals(e.PropertyName, "FadingLength", StringComparison.Ordinal)) {
          modifyingPoint = (sender as Effect).FadingStart;
        }
        else if (string.Equals(e.PropertyName, "TrailingSilenceLength", StringComparison.Ordinal)) {
          modifyingPoint = (sender as Effect).FadingStart + (sender as Effect).FadingLength;
        }
        else if (string.Equals(e.PropertyName, "FadingCurveType", StringComparison.Ordinal) ||
                 string.Equals(e.PropertyName, "FadingCurveCoefficient", StringComparison.Ordinal)) {
          modifyingPoint = (sender as Effect).FadingStart;
        }
        else {
          return;
        }

        modifiedWaveFormPoint = Math.Min(modifiedWaveFormPoint, (int)(modifyingPoint.TotalMilliseconds / WaveFormPartLengthInMilliseconds));

        var ev = this.Modified;

        if (ev != null)
          ev(this, EventArgs.Empty);
      }

      public void Refresh()
      {
        if (availableWaveFormLength < modifiedWaveFormPoint)
          return;

        AbortAndWait();

        availableWaveFormLength = Math.Max(0, modifiedWaveFormPoint);

        ReadWaveForm();
      }

      private void AbortAndWait()
      {
        if (readWaveFormTask == null)
          return;

        try {
          readWaveFormCancellation.Cancel();

          try {
            readWaveFormTask.Wait();
          }
          catch (AggregateException ex) {
            if (ex.InnerExceptions.Count == 1 && !(ex.InnerException is OperationCanceledException))
              throw;
          }
        }
        finally {
          readWaveFormTask.Dispose();
          readWaveFormTask = null;

          readWaveFormCancellation.Dispose();
          readWaveFormCancellation = null;
        }
      }

      private void ReadWaveForm()
      {
        if (stream == null)
          return; // disposed or stream is not available

        AbortAndWait();

        totalWaveFormLength = (int)((stream.Length / stream.Format.BlockAlign) / samplesPerWaveFormPart) + 1;

        if (totalWaveFormLength <= availableWaveFormLength)
          return; // all sample available

        if (waveFormLeft.Length < totalWaveFormLength)
          Array.Resize(ref waveFormLeft, totalWaveFormLength);
        if (waveFormRight.Length < totalWaveFormLength)
          Array.Resize(ref waveFormRight, totalWaveFormLength);

        stream.Position = availableWaveFormLength * samplesPerWaveFormPart * stream.Format.BlockAlign;

        Task.Factory.StartNew(() => {
          var ev = this.Refreshing;

          if (ev != null)
            ev(this, EventArgs.Empty);
        }, CancellationToken.None, TaskCreationOptions.None, TaskScheduler.FromCurrentSynchronizationContext());

        readWaveFormCancellation = new CancellationTokenSource();

        readWaveFormTask = Task.Factory.StartNew(() => ReadWaveFormCore(readWaveFormCancellation.Token), readWaveFormCancellation.Token);

        readWaveFormTask.ContinueWith((t) => {
          var ev = this.Refreshed;

          if (ev != null)
            ev(this, EventArgs.Empty);
        }, CancellationToken.None, TaskContinuationOptions.NotOnFaulted, TaskScheduler.FromCurrentSynchronizationContext());
      }

      private void ReadWaveFormCore(CancellationToken ct)
      {
        const float divider = 1.0f / Int16.MaxValue;

        var reader = new BinaryReader(stream);

        for (;;) {
          ct.ThrowIfCancellationRequested();

          int left = 0;
          int right = 0;

          for (int sample = 0; sample < samplesPerWaveFormPart; sample++) {
            int l, r;

            try {
              l = reader.ReadInt16();
              r = reader.ReadInt16();
            }
            catch (EndOfStreamException) {
              break;
            }

            if (0 < l && left < l) {
              left = l;
            }
            else {
              l = -l;

              if (left < l)
                left = l;
            }

            if (0 < r && right < r) {
              right = r;
            }
            else {
              r = -r;

              if (right < r)
                right = r;
            }

            const int cullingStride = (1 << 5) - 1;

            sample += 1 * cullingStride;

            reader.BaseStream.Position += /*stream.Format.BlockAlign*/4 * cullingStride;
          }

          waveFormLeft[availableWaveFormLength] = left * divider;
          waveFormRight[availableWaveFormLength] = right * divider;
          availableWaveFormLength++;

          modifiedWaveFormPoint = availableWaveFormLength;

          if (reader.BaseStream.Length <= reader.BaseStream.Position)
            break; // end of stream

          if (totalWaveFormLength <= availableWaveFormLength)
            break; // all sample available
        }
      }

      public void Fetch(TimeSpan start, TimeSpan length, Action<float[], int, int> fetchLeftWaveForm, Action<float[], int, int> fetchRightWaveForm)
      {
        var waveFormPartStart = (int)(start.TotalMilliseconds / WaveFormPartLengthInMilliseconds);
        var waveFormPartCount = (int)(length.TotalMilliseconds / WaveFormPartLengthInMilliseconds);

        if (availableWaveFormLength < waveFormPartStart) {
          fetchLeftWaveForm (EmptyArray<float>.Instance, 0, 0);
          fetchRightWaveForm(EmptyArray<float>.Instance, 0, 0);
        }
        else {
          waveFormPartCount = Math.Min(availableWaveFormLength, waveFormPartCount);

          fetchLeftWaveForm (waveFormLeft,  waveFormPartStart, waveFormPartCount);
          fetchRightWaveForm(waveFormRight, waveFormPartStart, waveFormPartCount);
        }
      }

      private EffectAppliedBgmStream stream;
      private Effect effect;
      private Task readWaveFormTask = null;
      private CancellationTokenSource readWaveFormCancellation = null;
      private TimeSpan effectPreviousFadingStartTime;
      private readonly int samplesPerWaveFormPart;
      private int totalWaveFormLength;
      private int availableWaveFormLength;
      private int modifiedWaveFormPoint;
      private float[] waveFormLeft;
      private float[] waveFormRight;
    }

    private class EffectAnchor : Control {
#if TRANSPARENT_ANCHOR
      protected override CreateParams CreateParams
      {
        get
        {
          const int WS_EX_TRANSPARENT = 0x20;
          var cp = base.CreateParams;
          cp.ExStyle |= WS_EX_TRANSPARENT;
          return cp;
        }
      }
#endif

      public EffectAnchor()
      {
#if TRANSPARENT_ANCHOR
        SetStyle(ControlStyles.Opaque, true);
        UpdateStyles();
#endif
        this.Size = new Size(9, 9);
        this.Cursor = Cursors.SizeWE;
      }

#if !TRANSPARENT_ANCHOR
      protected override void OnPaintBackground(System.Windows.Forms.PaintEventArgs pevent)
      {
        ControlPaint.DrawBorder3D(pevent.Graphics, this.Bounds);

        base.OnPaintBackground(pevent);
      }
#endif
    }

    public event EventHandler<SeekBarMovedEventArgs> SeekBarMoved;

    public string BgmSourcePath {
      get; set;
    }

    public Track Track {
      get { return track; }
      set
      {
        if (track == value)
          return;

        track = value;

        OnTrackChanged();
      }
    }

    public RectangleF XAxisRulerBounds {
      get; private set;
    }

    public RectangleF YAxisRulerBounds {
      get; private set;
    }

    public RectangleF WaveFormBounds {
      get; private set;
    }

    public TimeSpan SeekBarPosition {
      get { return seekBarPosition; }
      set
      {
        if (value < TimeSpan.Zero)
          throw new ArgumentOutOfRangeException("SeekBarPosition", "must be zero or positive value");
        var last = seekBarPosition;
        seekBarPosition = value;
        UpdateSeekBarPosition(last, SeekBarAutoTracking);
      }
    }

    public bool SeekBarAutoTracking {
      get; set;
    }

    private float PixelsPerSecondScale = 1.0f;

    private float OffsetInPixels {
      get { return -(float)(Offset.TotalSeconds * PixelsPerSecondScale); }
    }

    private const int OffsetScrollValuePerSecondScale = 16;

    private TimeSpan Offset {
      get { return TimeSpan.FromSeconds(hscrollOffset.Value / (float)OffsetScrollValuePerSecondScale); }
      set {
        var newValue = (int)(value.TotalSeconds * OffsetScrollValuePerSecondScale);

        hscrollOffset.Value = Math.Min(newValue, hscrollOffset.Maximum - hscrollOffset.LargeChange + 1);
      }
    }

    private float WaveFormBoundsMiddle {
      get; set;
    }

    public const int ZoomScaleMax = 5;
    public const int ZoomScaleMin = -1;
    public const int ZoomScaleDefault = 1;

    private int zoomScale = ZoomScaleDefault;

    public int ZoomScale {
      get { return zoomScale; }
      set { SetZoomScale(value); }
    }

    public WaveFormView()
    {
      this.SetStyle(ControlStyles.DoubleBuffer, true);
      this.SetStyle(ControlStyles.UserPaint, true);
      this.SetStyle(ControlStyles.AllPaintingInWmPaint, true);
      
      base.ResizeRedraw = true;

      this.hscrollOffset = new HScrollBar();
      this.hscrollOffset.Anchor = AnchorStyles.Left | AnchorStyles.Bottom;
      this.hscrollOffset.Minimum = 0;
      this.hscrollOffset.Maximum = 300 * OffsetScrollValuePerSecondScale; // must be greater than LargeChange
      this.hscrollOffset.SmallChange = 5 /*secs*/ * OffsetScrollValuePerSecondScale;
      this.hscrollOffset.LargeChange = 60/*secs*/ * OffsetScrollValuePerSecondScale; 
      this.hscrollOffset.ValueChanged += delegate {
        UpdateAnchorLocation();

        Refresh();
      };

      this.Controls.Add(hscrollOffset);

      this.buttonZoomOut = new Button();
      this.buttonZoomOut.Anchor = AnchorStyles.Right | AnchorStyles.Bottom;
      this.buttonZoomOut.Text = "-";
      this.buttonZoomOut.TextAlign = ContentAlignment.MiddleCenter;
      this.buttonZoomOut.FlatStyle = FlatStyle.Flat;
      this.buttonZoomOut.Click += delegate {
        SetZoomScale(Math.Max(zoomScale - 1, ZoomScaleMin));
      };

      this.Controls.Add(buttonZoomOut);

      this.buttonZoomIn = new Button();
      this.buttonZoomIn.Anchor = AnchorStyles.Right | AnchorStyles.Bottom;
      this.buttonZoomIn.Text = "+";
      this.buttonZoomIn.TextAlign = ContentAlignment.MiddleCenter;
      this.buttonZoomIn.FlatStyle = FlatStyle.Flat;
      this.buttonZoomIn.Click += delegate {
        SetZoomScale(Math.Min(zoomScale + 1, ZoomScaleMax));
      };

      this.Controls.Add(buttonZoomIn);

      this.fadingStartAnchor = new EffectAnchor();
      this.fadingEndAnchor = new EffectAnchor();
      this.trailingSilenceEndAnchor = new EffectAnchor();

      anchors = new[] {fadingStartAnchor, fadingEndAnchor, trailingSilenceEndAnchor};

      foreach (var anchor in anchors) {
        anchor.MouseDown += EffectAnchor_MouseDown;
        anchor.MouseMove += EffectAnchor_MouseMove;
        anchor.MouseUp   += EffectAnchor_MouseUp;
      }

      this.Controls.Add(fadingStartAnchor);
      this.Controls.Add(fadingEndAnchor);
      this.Controls.Add(trailingSilenceEndAnchor);

      Controls.SetChildIndex(fadingStartAnchor, 2);
      Controls.SetChildIndex(fadingEndAnchor, 1);
      Controls.SetChildIndex(trailingSilenceEndAnchor, 0);

      this.fadingStartAnchor.MouseHover += delegate(object sender, EventArgs e) {
        var a = sender as EffectAnchor;
        var text = Texts.GetText("FadingStartPoint");

        if (track != null)
          text = string.Concat(text,
                               Environment.NewLine,
                               Effect.TimeSpanToStringConverter(track.Effect.FadingStart));

        this.toolTipAnchor.Show(text, a, a.Width, a.Height, 1500);
      };
      this.fadingEndAnchor.MouseHover += delegate(object sender, EventArgs e) {
        var a = sender as EffectAnchor;
        var text = Texts.GetText("FadingEndPoint");

        if (track != null)
          text = string.Concat(text,
                               Environment.NewLine,
                               Effect.TimeSpanToStringConverter(track.Effect.FadingStart + track.Effect.FadingLength),
                               Environment.NewLine,
                               Texts.GetText("FadingLength"),
                               Environment.NewLine,
                               Effect.TimeSpanToStringConverter(track.Effect.FadingLength));

        this.toolTipAnchor.Show(text, a, a.Width, a.Height, 1500);
      };
      this.trailingSilenceEndAnchor.MouseHover += delegate(object sender, EventArgs e) {
        var a = sender as EffectAnchor;

        var text = Texts.GetText("TrackEndPoint");

        if (track != null)
          text = string.Concat(text,
                               Environment.NewLine,
                               Effect.TimeSpanToStringConverter(track.Effect.TotalLength),
                               Environment.NewLine,
                               Texts.GetText("TrailingSilenceLength"),
                               Environment.NewLine,
                               Effect.TimeSpanToStringConverter(track.Effect.TrailingSilenceLength));

        this.toolTipAnchor.Show(text, a, a.Width, a.Height, 1500);
      };

      this.toolTipAnchor = new ToolTip();
      this.toolTipAnchor.ShowAlways = false;

      this.toolTipPosition = new ToolTip();
      this.toolTipPosition.ShowAlways = false;

      components.Add(hscrollOffset);
      components.Add(buttonZoomIn);
      components.Add(buttonZoomOut);
      components.Add(fadingStartAnchor);
      components.Add(fadingEndAnchor);
      components.Add(trailingSilenceEndAnchor);
      components.Add(toolTipAnchor);
      components.Add(toolTipPosition);

      SetZoomScale(ZoomScaleDefault);
    }

    protected override void OnSizeChanged(EventArgs e)
    {
      const float rulerSize = 16.0f;
      const int toolSize = 14;

      XAxisRulerBounds = RectangleF.FromLTRB(rulerSize, 0.0f, (float)Width, rulerSize);
      YAxisRulerBounds = RectangleF.FromLTRB(0.0f, rulerSize, rulerSize, (float)(Height - toolSize));
      WaveFormBounds = RectangleF.FromLTRB(rulerSize, rulerSize, XAxisRulerBounds.Right, YAxisRulerBounds.Bottom);
      WaveFormBoundsMiddle = WaveFormBounds.Top + WaveFormBounds.Height * 0.5f;

      hscrollOffset.Width = ClientSize.Width - toolSize * 2; // space for zoom button
      hscrollOffset.Height = toolSize;
      hscrollOffset.Left = 0;
      hscrollOffset.Top = ClientSize.Height - toolSize;

      buttonZoomIn.Width = toolSize;
      buttonZoomIn.Height = toolSize;
      buttonZoomIn.Left = ClientSize.Width - toolSize;
      buttonZoomIn.Top = ClientSize.Height - toolSize;

      buttonZoomOut.Width = toolSize;
      buttonZoomOut.Height = toolSize;
      buttonZoomOut.Left = buttonZoomIn.Left - toolSize;
      buttonZoomOut.Top = buttonZoomIn.Top;

      UpdateAnchorLocation();

      DiscardCachedWaveFormImage();

      base.OnSizeChanged(e);
    }

    protected override void Dispose(bool disposing)
    {
      if (disposing) {
        if (components != null)
          components.Dispose();

        if (waveForm != null) {
          waveForm.Dispose();
          waveForm = null;
        }

        DiscardCachedWaveFormImage();
      }

      base.Dispose(disposing);
    }

    protected override void OnCreateControl()
    {
      UpdateScrollRange();

      UpdateAnchorLocation();

      base.OnCreateControl();
    }

    private void OnTrackChanged()
    {
      if (waveForm != null) {
        waveForm.Refreshing -= WaveForm_Refreshing;
        waveForm.Refreshed -= WaveForm_Refreshed;
        waveForm.Modified -= WaveForm_Modified;
        waveForm.Dispose();
        waveForm = null;
      }

      seekBarPosition = TimeSpan.Zero;

      DiscardCachedWaveFormImage();

      if (track == null) {
        waveForm = null;
      }
      else {
        waveForm = WaveForm.Create(BgmSourcePath, track);
        waveForm.Refreshing += WaveForm_Refreshing;
        waveForm.Refreshed += WaveForm_Refreshed;
        waveForm.Modified += WaveForm_Modified;
        waveForm.Refresh();
      }

      UpdateView();
    }

    private void SetZoomScale(int newZoomScale)
    {
      if (newZoomScale < ZoomScaleMin || ZoomScaleMax < newZoomScale)
        throw ExceptionUtils.CreateArgumentMustBeInRange(ZoomScaleMin, ZoomScaleMax, "newZoomScale", newZoomScale);

      if (zoomScale == newZoomScale)
        return;

      zoomScale = newZoomScale;

      PixelsPerSecondScale = (float)Math.Pow(2, newZoomScale);

      DiscardCachedWaveFormImage();

      UpdateView();
    }

    public void UpdateView()
    {
      UpdateScrollRange();

      UpdateAnchorLocation();

      Refresh();
    }

    private void UpdateScrollRange()
    {
      const int marginInSeconds = 30; // [secs]

      var maxValueInSeconds = (track == null) ? 0 : (int)Math.Ceiling(track.Effect.TotalLength.TotalSeconds - WaveFormBounds.Width / PixelsPerSecondScale) + marginInSeconds;
      var maxValue = maxValueInSeconds * OffsetScrollValuePerSecondScale;

      // (maximum value) = 1 + .Maximum - .LargeChange
      // .Maximum = (maximum value) - 1 + .LargeChange
      var maximum = maxValue - 1 + hscrollOffset.LargeChange;

      if (maximum < hscrollOffset.LargeChange) {
        hscrollOffset.Value = 0;
        hscrollOffset.Enabled = false;
      }
      else {
        hscrollOffset.Enabled = true;
        hscrollOffset.Maximum = maximum;
      }
    }

    private void UpdateAnchorLocation()
    {
      if (track == null) {
        foreach (var anchor in anchors) {
          anchor.Visible = false;
        }
        return;
      }

      foreach (var pair in new[] {
        new {Anchor = fadingStartAnchor,        Location = track.Effect.FadingStart.TotalSeconds},
        new {Anchor = fadingEndAnchor,          Location = (track.Effect.FadingStart + track.Effect.FadingLength).TotalSeconds},
        new {Anchor = trailingSilenceEndAnchor, Location = track.Effect.TotalLength.TotalSeconds},
      }) {
        pair.Anchor.Location = new Point((int)(pair.Location * PixelsPerSecondScale + OffsetInPixels + WaveFormBounds.Left - pair.Anchor.Width / 2), (int)WaveFormBoundsMiddle - pair.Anchor.Height / 2);
        pair.Anchor.Visible = YAxisRulerBounds.Right <= pair.Anchor.Right;
      }
    }

    private void UpdateSeekBarPosition(TimeSpan lastSeekBarPosition, bool tracking)
    {
      if (tracking)
        Offset = TimeSpan.FromSeconds(Math.Max(0.0, seekBarPosition.TotalSeconds - (WaveFormBounds.Width / 2) / PixelsPerSecondScale));

      var newX = (float)(seekBarPosition.TotalSeconds * PixelsPerSecondScale) + OffsetInPixels + WaveFormBounds.Left;
      var lastX = (float)(lastSeekBarPosition.TotalSeconds * PixelsPerSecondScale) + OffsetInPixels + WaveFormBounds.Left;

      Invalidate(Rectangle.Ceiling(RectangleF.FromLTRB(newX - 2.0f, WaveFormBounds.Top, newX + 2.0f, WaveFormBounds.Bottom)));
      Invalidate(Rectangle.Ceiling(RectangleF.FromLTRB(lastX - 2.0f, WaveFormBounds.Top, lastX + 2.0f, WaveFormBounds.Bottom)));
      Update();
    }

    protected override void OnPaint(PaintEventArgs e)
    {
      var g = e.Graphics;

      g.Clear(SystemColors.AppWorkspace);
      g.InterpolationMode = InterpolationMode.HighQualityBicubic;
      g.SmoothingMode = SmoothingMode.AntiAlias;

      // ruler
      var drawSecondsInRuler = 40.0f < (10.0f * PixelsPerSecondScale); // 40px < 10sec

      for (var second = 0; ; second += 10) {
        var x = second * PixelsPerSecondScale + OffsetInPixels;

        if (x < 0)
          continue;

        if (XAxisRulerBounds.Width < x || XAxisRulerBounds.Width <= 0)
          break;

        x += XAxisRulerBounds.Left;

        if (drawSecondsInRuler || second % 60 == 0) {
          g.DrawLine(SystemPens.ControlText, x, XAxisRulerBounds.Bottom - 4.0f, x, XAxisRulerBounds.Bottom);
          g.DrawString(TimeSpan.FromSeconds(second).ToString(@"mm\:ss"), Control.DefaultFont, SystemBrushes.ControlText, x, XAxisRulerBounds.Bottom - 16.0f);
        }
      }

      // wave form bound
      g.FillRectangle(Brushes.Black, WaveFormBounds);

      // effect and wave form
      if (track == null)
        goto DrawEffectDone;

      var effect = track.Effect;
      var fadeStartX  = (float)(effect.FadingStart.TotalSeconds * PixelsPerSecondScale) + OffsetInPixels;
      var fadeEndX    = (float)((effect.FadingStart + effect.FadingLength).TotalSeconds * PixelsPerSecondScale) + OffsetInPixels;
      var silenceEndX = (float)((effect.FadingStart + effect.FadingLength + effect.TrailingSilenceLength).TotalSeconds * PixelsPerSecondScale) + OffsetInPixels;

      using (var brushFadeEffect = new SolidBrush(Color.FromArgb(0x40, 0x00, 0x40, 0xff))) {
        if (WaveFormBounds.Width < fadeStartX)
          g.FillRectangle(brushFadeEffect, WaveFormBounds);
        else
          g.FillRectangle(brushFadeEffect, WaveFormBounds.Left, WaveFormBounds.Top - 1, fadeStartX, WaveFormBounds.Height);

        g.SmoothingMode = SmoothingMode.None;

        using (var b = new SolidBrush(Color.FromArgb(0x20, 0x20, 0x20))) {
          g.FillRectangle(b, RectangleF.FromLTRB(WaveFormBounds.Left + silenceEndX, WaveFormBounds.Top, WaveFormBounds.Right, WaveFormBounds.Bottom));
        }

        g.DrawLine(SystemPens.ControlText, WaveFormBounds.Left, WaveFormBounds.Top, WaveFormBounds.Right, WaveFormBounds.Top);
        g.DrawLine(SystemPens.ControlText, WaveFormBounds.Left, WaveFormBounds.Top, WaveFormBounds.Left, WaveFormBounds.Bottom);

        g.SmoothingMode = SmoothingMode.AntiAlias;

        if ((fadeStartX < 0.0f && fadeEndX < 0.0f) || (WaveFormBounds.Width < fadeStartX))
          goto DrawFadeEffectDone;

        // fading curve
        PointF[] fadingCurvePoints;

        switch (effect.FadingCurveType) {
          case FadingCurveType.None:
            fadingCurvePoints = new[] {
              new PointF(fadeStartX, 1.0f),
              new PointF(fadeEndX,   1.0f),
            };
            break;

          case FadingCurveType.Linear:
            fadingCurvePoints = new[] {
              new PointF(fadeStartX, 1.0f),
              new PointF(fadeEndX,   0.0f),
            };
            break;

          case FadingCurveType.Exponential:
          case FadingCurveType.Logarithmic:
          case FadingCurveType.Trigonometrical:
            fadingCurvePoints = new PointF[30];

            for (var i = 0; i < fadingCurvePoints.Length; i++) {
              var ratio = i / (double)(fadingCurvePoints.Length - 1);
              var vol = 1.0 - ratio;

              if (effect.FadingCurveType == FadingCurveType.Exponential)
                vol = Effect.CalculateVolumeExponential(vol, effect.FadingCurveCoefficient);
              else if (effect.FadingCurveType == FadingCurveType.Logarithmic)
                vol = Effect.CalculateVolumeLogarithmic(vol, effect.FadingCurveCoefficient);
              else if (effect.FadingCurveType == FadingCurveType.Trigonometrical)
                vol = Effect.CalculateVolumeTrigonometical(vol, effect.FadingCurveCoefficient);

              fadingCurvePoints[i] = new PointF((float)(fadeStartX + (fadeEndX - fadeStartX) * ratio), (float)vol);
            }
            break;

          default:
            fadingCurvePoints = EmptyArray<PointF>.Instance;
            break;
        }

        // draw outline
        var fadingUpperCurvePoints = Array.ConvertAll(fadingCurvePoints, delegate(PointF p) {
          return new PointF(p.X + WaveFormBounds.Left, (1.0f - p.Y) * WaveFormBounds.Height * 0.5f + WaveFormBounds.Top);
        });
        var fadingLowerCurvePoints = Array.ConvertAll(fadingCurvePoints, delegate(PointF p) {
          return new PointF(p.X + WaveFormBounds.Left, (0.5f * p.Y + 0.5f) * WaveFormBounds.Height + WaveFormBounds.Top);
        });

        using (var penFadeEffect = new Pen(Color.FromArgb(0x80, 0x00, 0x40, 0xff))) {
          g.DrawCurve(penFadeEffect, fadingUpperCurvePoints);
          g.DrawCurve(penFadeEffect, fadingLowerCurvePoints);
        }

        // merges upper and lower curve points to fill
        var countPoints = fadingUpperCurvePoints.Length;

        Array.Resize(ref fadingUpperCurvePoints, fadingUpperCurvePoints.Length + fadingLowerCurvePoints.Length);
        Array.Reverse(fadingLowerCurvePoints);
        Array.Copy(fadingLowerCurvePoints, 0, fadingUpperCurvePoints, countPoints, fadingLowerCurvePoints.Length);

        g.FillPolygon(brushFadeEffect, fadingUpperCurvePoints, FillMode.Winding);
      } // using Brush, Pen

DrawFadeEffectDone:
      // wave form
      var truncatedWaveFormBounds = Rectangle.Truncate(WaveFormBounds);

      lock (cachedWaveFormImageLockObject) {
        if (cachedWaveFormImage == null)
          RedrawCachedWaveFormImage();

        if (cachedWaveFormImage != null)
          g.DrawImage(cachedWaveFormImage, truncatedWaveFormBounds, -(int)OffsetInPixels, 0, truncatedWaveFormBounds.Width, truncatedWaveFormBounds.Height, GraphicsUnit.Pixel);
      }

      // repeated point
      if (0 < track.TrackInfo.RepeatLength && 0.0f < WaveFormBounds.Width) {
        var prevSmoothingMode = g.SmoothingMode;

        g.SmoothingMode = SmoothingMode.None;

        using (var p = new Pen(Color.LightBlue)) {
          p.DashStyle = DashStyle.Dash;

          for (var repetition = track.TrackInfo.CalculateRepeatedTimes(Offset); ;repetition++) {
            var x = (float)(track.TrackInfo.CalculateRepeatPosition(repetition).TotalSeconds * PixelsPerSecondScale) + OffsetInPixels;

            if (x < 0.0f)
              continue;
            if (WaveFormBounds.Width <= x)
              break;

            x += WaveFormBounds.Left;

            g.DrawLine(p, x, WaveFormBounds.Top, x, WaveFormBounds.Bottom);
            g.DrawString((repetition + 1).ToString(), Control.DefaultFont, Brushes.White, x, WaveFormBounds.Top);
          }
        }

        g.SmoothingMode = prevSmoothingMode;
      }

      // anchors
      {
        var prevSmoothingMode = g.SmoothingMode;

        g.SmoothingMode = SmoothingMode.None;

        // base line
        g.DrawLine(Pens.White, WaveFormBounds.Left, WaveFormBoundsMiddle, WaveFormBounds.Left + silenceEndX, WaveFormBoundsMiddle);

        // anchor point line
        using (var p = new Pen(Color.FromArgb(0xc0, 0x00, 0x40, 0xff))) {
          p.DashStyle = DashStyle.Dash;

          g.DrawLine(p, WaveFormBounds.Left + fadeStartX, WaveFormBounds.Top, WaveFormBounds.Left + fadeStartX, WaveFormBounds.Bottom);
          g.DrawLine(p, WaveFormBounds.Left + fadeEndX,   WaveFormBounds.Top, WaveFormBounds.Left + fadeEndX,   WaveFormBounds.Bottom);
        }

        g.SmoothingMode = SmoothingMode.AntiAlias;

        // anchors
        foreach (var anchorPoint in new[] {fadeStartX, fadeEndX, silenceEndX}) {
          var ax = WaveFormBounds.Left + anchorPoint;
          var ay = WaveFormBoundsMiddle;

          // diamond shape
          const float anchorSize = 4.5f;

          var points = new PointF[] {
            new PointF(ax - anchorSize, ay),
            new PointF(ax             , ay + anchorSize),
            new PointF(ax + anchorSize, ay),
            new PointF(ax             , ay - anchorSize),
          };

          g.FillPolygon(Brushes.White, points);
        }

        g.SmoothingMode = prevSmoothingMode;
      }

DrawEffectDone:;
      // seek bar
      var seekBarX = (float)(seekBarPosition.TotalSeconds * PixelsPerSecondScale) + OffsetInPixels;

      if (0.0 <= seekBarX) {
        var prevSmoothingMode = g.SmoothingMode;
        g.SmoothingMode = SmoothingMode.None;
        using (var p = new Pen(Color.Red, 1)) {
          g.DrawLine(p, seekBarX + WaveFormBounds.Left, WaveFormBounds.Top, seekBarX + WaveFormBounds.Left, WaveFormBounds.Bottom);
        }
        g.SmoothingMode = prevSmoothingMode;
      }

      // track title
      if (track != null) {
        g.TextRenderingHint = TextRenderingHint.ClearTypeGridFit;

        using (var format = StringFormat.GenericDefault.Clone() as StringFormat) {
          format.Alignment = StringAlignment.Near;
          format.LineAlignment = StringAlignment.Far;
          format.FormatFlags |= StringFormatFlags.NoWrap;
          format.Trimming = StringTrimming.EllipsisCharacter;

          g.DrawString(track.TrackInfo.Title, Control.DefaultFont, Brushes.White, WaveFormBounds, format);
        }
      }

      base.OnPaint(e);
    }

    private void DiscardCachedWaveFormImage()
    {
      lock (cachedWaveFormImageLockObject) {
        if (cachedWaveFormImage != null) {
          cachedWaveFormImage.Dispose();
          cachedWaveFormImage = null;
        }
      }
    }

    private void RedrawCachedWaveFormImage()
    {
      try {
        Cursor = Cursors.WaitCursor;

        var totalLength = track.Effect.TotalLength;
        var width = (int)(totalLength.TotalSeconds * PixelsPerSecondScale) + 1;
        var height = (int)WaveFormBounds.Height + 1;

        if (width <= 0 || height <= 0)
          return;

        cachedWaveFormImage = new Bitmap(width, height, PixelFormat.Format32bppArgb);

        using (var g = Graphics.FromImage(cachedWaveFormImage)) {
          g.Clear(Color.Transparent);

          if (waveForm == null)
            return;

          var waveFormBoundsHalfHeight = WaveFormBounds.Height * 0.5f;
          var waveFormLeft  = new PointF[width];
          var waveFormRight = new PointF[width];
          var scale = WaveForm.WaveFormPartLengthInMilliseconds * PixelsPerSecondScale / 1000.0f;

          for (var x = 0; x < width; x++) {
            waveFormLeft [x] = new PointF((float)x, waveFormBoundsHalfHeight);
            waveFormRight[x] = new PointF((float)x, waveFormBoundsHalfHeight);
          }

          waveForm.Fetch(TimeSpan.Zero,
                         totalLength,
                         delegate (float[] samplesLeft, int start, int count) {
                           int xp = 0;

                           for (int offset = 0, index = start; offset < count; offset++, index++) {
                             int x = (int)(offset * scale);

                             if (waveFormLeft.Length <= x)
                               break; // may length changed while fetching

                           if (xp < x)
                               waveFormLeft[x] = new PointF((float)x, Math.Min(waveFormLeft[x].Y, waveFormBoundsHalfHeight - samplesLeft[index] * waveFormBoundsHalfHeight));

                             xp = x;
                           }
                         },
                         delegate (float[] samplesRight, int start, int count) {
                           int xp = 0;

                           for (int offset = 0, index = start; offset < count; offset++, index++) {
                             int x = (int)(offset * scale);

                             if (waveFormRight.Length <= x)
                               break; // may length changed while fetching

                           if (xp < x)
                               waveFormRight[x] = new PointF((float)x, Math.Max(waveFormLeft[x].Y, waveFormBoundsHalfHeight + samplesRight[index] * waveFormBoundsHalfHeight));

                             xp = x;
                           }
                         });

          if (1 < waveFormLeft.Length && 1 < waveFormRight.Length) {
            g.SmoothingMode = SmoothingMode.AntiAlias;

            using (var p = new Pen(Color.FromArgb(0xc0, 0x60, 0xa0, 0xff), 1.0f)) {
              g.DrawLines(p, waveFormLeft);
              g.DrawLines(p, waveFormRight);
            }

            using (var b = new SolidBrush(Color.FromArgb(0x60, 0x60, 0xa0, 0xff))) {
              g.FillPolygon(b, waveFormLeft, FillMode.Winding);
              g.FillPolygon(b, waveFormRight, FillMode.Winding);
            }
          }
        } // using
      }
      finally {
        Cursor = Cursors.Default;
      }
    }

    private TimeSpan ConvertPointToPosition(Point point)
    {
      return TimeSpan.FromSeconds(((point.X - WaveFormBounds.Left) - OffsetInPixels) / (double)PixelsPerSecondScale);
    }

    protected override void OnClick(EventArgs e)
    {
      var point = PointToClient(Cursor.Position);

      if (WaveFormBounds.Contains(point)) {
        var newPosition = ConvertPointToPosition(point);

        if (track != null && newPosition <= track.Effect.TotalLength)
          OnSeekBarMoved(new SeekBarMovedEventArgs(newPosition));
      }

      base.OnClick(e);
    }

    private Point pointOfLastToolTipPosition = Point.Empty;

    protected override void OnMouseMove(MouseEventArgs e)
    {
      var point = new Point(e.X, e.Y);

      if (pointOfLastToolTipPosition != point && WaveFormBounds.Contains(point)) {
        var position = ConvertPointToPosition(point);

        this.toolTipPosition.Show(position.ToString(@"mm\:ss\.fff"), this, point.X + 10, point.Y + 3, 2000);

        pointOfLastToolTipPosition = point;
      }

      base.OnMouseMove(e);
    }

    protected virtual void OnSeekBarMoved(SeekBarMovedEventArgs e)
    {
      var ev = this.SeekBarMoved;

      if (ev != null)
        ev(this, e);

      if (!e.Cancel) {
        var lastSeekBarPosition = seekBarPosition;
        seekBarPosition = e.SeekBarPosition;
        UpdateSeekBarPosition(lastSeekBarPosition, false);
      }
    }

#region "EffectAnchor drag and drop operations"
    private Point effectAnchor_InitialCursorPosition;
    private Point effectAnchor_InitialLocation;
    private Effect effectAnchor_InitialEffect;

    private bool IsEffectAnchorDragging {
      get { return effectAnchor_InitialEffect != null; }
    }

    private void EffectAnchor_MouseDown(object sender, MouseEventArgs e)
    {
      if (track == null)
        return;

      effectAnchor_InitialLocation = (sender as EffectAnchor).Location;
      effectAnchor_InitialEffect = track.Effect.Clone();
      effectAnchor_InitialCursorPosition = Cursor.Position;
    }

    private void EffectAnchor_MouseMove(object sender, MouseEventArgs e)
    {
      if (track == null)
        return;
      if (!IsEffectAnchorDragging)
        return;

      var difference = TimeSpan.FromSeconds((Cursor.Position.X - effectAnchor_InitialCursorPosition.X) / (double)PixelsPerSecondScale);

      try {
        updatingEffectProperty = true;

        if (sender == fadingStartAnchor) {
          var newValue = effectAnchor_InitialEffect.FadingStart + difference;

          track.Effect.FadingStart = (newValue < TimeSpan.Zero) ? TimeSpan.Zero : newValue;
        }
        else if (sender == fadingEndAnchor) {
          var newValue = effectAnchor_InitialEffect.FadingLength + difference;

          track.Effect.FadingLength = (newValue < TimeSpan.Zero) ? TimeSpan.Zero : newValue;
        }
        else if (sender == trailingSilenceEndAnchor) {
          var newValue = effectAnchor_InitialEffect.TrailingSilenceLength + difference;

          track.Effect.TrailingSilenceLength = (newValue < TimeSpan.Zero) ? TimeSpan.Zero : newValue;
        }
      }
      finally {
        updatingEffectProperty = false;
      }

      UpdateAnchorLocation();

      Invalidate(Rectangle.FromLTRB(Math.Min(fadingStartAnchor.Left, effectAnchor_InitialLocation.X) - 5,
                                    (int)WaveFormBounds.Top,
                                    (int)WaveFormBounds.Right,
                                    (int)WaveFormBounds.Bottom));
      Update();
    }

    private void EffectAnchor_MouseUp(object sender, MouseEventArgs e)
    {
      effectAnchor_InitialEffect = null;

      if (waveForm != null)
        waveForm.Refresh();

      UpdateScrollRange();

      Refresh();
    }
#endregion

    private void WaveForm_Refreshing(object sender, EventArgs e)
    {
      Cursor = Cursors.AppStarting;
    }

    private void WaveForm_Refreshed(object sender, EventArgs e)
    {
      DiscardCachedWaveFormImage();

      InvalidateWaveFormBounds();

      Cursor = Cursors.Default;
    }

    private void WaveForm_Modified(object sender, EventArgs e)
    {
      if (IsEffectAnchorDragging)
        return;

      waveForm.Refresh();

      DiscardCachedWaveFormImage();

      if (!updatingEffectProperty)
        UpdateView();
    }

    private void InvalidateWaveFormBounds()
    {
      if (InvokeRequired) {
        Invoke((MethodInvoker)InvalidateWaveFormBounds);
      }
      else {
        Invalidate(Rectangle.Ceiling(WaveFormBounds));
        Update();
      }
    }

    private Track track = null;
    private WaveForm waveForm = null;
    private System.ComponentModel.Container components = new System.ComponentModel.Container();
    private HScrollBar hscrollOffset = null;
    private Button buttonZoomIn;
    private Button buttonZoomOut;
    private EffectAnchor fadingStartAnchor;
    private EffectAnchor fadingEndAnchor;
    private EffectAnchor trailingSilenceEndAnchor;
    private EffectAnchor[] anchors;
    private ToolTip toolTipAnchor;
    private ToolTip toolTipPosition;
    private TimeSpan seekBarPosition = TimeSpan.Zero;
    private bool updatingEffectProperty = false;
    private Bitmap cachedWaveFormImage;
    private object cachedWaveFormImageLockObject = new object();
  }
}
