//  
//  Author:
//       smdn <smdn@smdn.jp>
// 
//  Copyright (c) 2004-2014 smdn
// 
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
// 
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
//  GNU General Public License for more details.
//  
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

using System;
using System.Collections.Generic;
using System.Windows.Forms;

using Smdn.Formats.Thbgm;

namespace Smdn.Applications.ThbgmExtractor {
  public partial class CollectiveTrackTagEditDialog : Form, ITrackFormatProvider {
    private class TrackInfoFormatToolStripMenuItem : ToolStripMenuItem {
      public string Format {
        get; private set;
      }

      public TrackInfoFormatToolStripMenuItem(string format, string text, EventHandler onClick)
        : base(text, null, onClick)
      {
        this.Format = format;
      }
    }

    private class TrackInfoFormatPresetToolStripMenuItem : TrackInfoFormatToolStripMenuItem {
      public TrackInfoFormatPresetToolStripMenuItem(string format, string text, EventHandler onClick)
        : base(format, text, onClick)
      {
      }
    }

    public string OutputFileName {
      get { return textBoxOutputFileName.Text; }
      set { textBoxOutputFileName.Text = value; }
    }

    public TrackFormatInfo TrackFormatInfo {
      get { return trackFormatInfo; }
      set
      {
        if (trackFormatInfo == value)
          return;

        trackFormatInfo = value ?? new TrackFormatInfo();

        if (lockChanged)
          return;

        try {
          lockChanged = true;

          OnTrackFormatInfoChanged();
        }
        finally {
          lockChanged = false;
        }
      }
    }

    TrackFormatInfo ITrackFormatProvider.FormatPrivider {
      get { return trackFormatInfo; }
    }

    public string Album {
      get { return textBoxAlbum.Text; }
    }

    public string[] Performers {
      get { return TrackTag.GetSplitted(textBoxPerformers.Text); }
    }

    public string[] AlbumArtists {
      get { return TrackTag.GetSplitted(textBoxAlbumArtists.Text); }
    }

    public string[] Composers {
      get { return TrackTag.GetSplitted(textBoxComposers.Text); }
    }

    public string[] Genres {
      get { return TrackTag.GetSplitted(textBoxGenres.Text); }
    }

    public string Comment {
      get { return textBoxComment.Text; }
    }

    public TrackInfo TrackForDefaults {
      get { return trackForDefaults; }
      set
      {
        if (trackForDefaults == value)
          return;

        trackForDefaults = value;

        OnTrackForDefaultsChanged();
      }
    }

    public CollectiveTrackTagEditDialog()
    {
      InitializeComponent();
      Initialize();
    }

    private void Initialize()
    {
      var presetsMenuItem = new ToolStripMenuItem(ThbgmExtractor.GetText("CollectiveTrackTagEditDialog_Presets"));

      foreach (var pair in new[] {
        new {Format = "$title",                       Text = ThbgmExtractor.GetText("FileNameFormat_Preset_TitleOnly")},
        new {Format = "$trackno. $title",             Text = ThbgmExtractor.GetText("FileNameFormat_Preset_MusicRoomStyle1")},
        new {Format = "No$trackno. $title",           Text = ThbgmExtractor.GetText("FileNameFormat_Preset_MusicRoomStyle2")},
        new {Format = "$prefix_$trackno",             Text = ThbgmExtractor.GetText("FileNameFormat_Preset_Th06BgmFileStyle")},
        new {Format = "$pshortest$trackno：$title",   Text = ThbgmExtractor.GetText("FileNameFormat_Preset_THxxBGMStyle")},
      }) {
        presetsMenuItem.DropDownItems.Add(new TrackInfoFormatPresetToolStripMenuItem(pair.Format, pair.Text, contextMenuStripFormats_ItemsClick));
      }

      contextMenuStripFormats.Items.AddRange(new ToolStripItem[] {
        presetsMenuItem,
        new ToolStripSeparator(),
      });

      foreach (var pair in new[] {
        new {Format = "$title",     Text = ThbgmExtractor.GetText("FileNameFormat_title")},
        new {Format = "$trackno",   Text = ThbgmExtractor.GetText("FileNameFormat_trackno")},
        new {Format = "$tracks",    Text = ThbgmExtractor.GetText("FileNameFormat_tracks")},
        new {Format = "$creator",   Text = ThbgmExtractor.GetText("FileNameFormat_creator")},
        new {Format = "$product",   Text = ThbgmExtractor.GetText("FileNameFormat_product")},
        new {Format = "$pshort1",   Text = ThbgmExtractor.GetText("FileNameFormat_pshort1")},
        new {Format = "$pshort2",   Text = ThbgmExtractor.GetText("FileNameFormat_pshort2")},
        new {Format = "$pshortest", Text = ThbgmExtractor.GetText("FileNameFormat_pshortest")},
        new {Format = "$prefix",    Text = ThbgmExtractor.GetText("FileNameFormat_prefix")},
        new {Format = "$ryear",     Text = ThbgmExtractor.GetText("FileNameFormat_ryear")},
        new {Format = "$rmonth",    Text = ThbgmExtractor.GetText("FileNameFormat_rmonth")},
        new {Format = "$rday",      Text = ThbgmExtractor.GetText("FileNameFormat_rday")},
      }) {
        contextMenuStripFormats.Items.Add(new TrackInfoFormatToolStripMenuItem(pair.Format, pair.Text, contextMenuStripFormats_ItemsClick));
      }

      contextMenuStripFormats.Items.AddRange(new ToolStripItem[] {
        new ToolStripSeparator(),
        new ToolStripMenuItem(ThbgmExtractor.GetText("CollectiveTrackTagEditDialog_Clear"), null, delegate {
          textBoxOutputFileName.Clear();
        }),
      });
    }

    protected override void OnShown(EventArgs e)
    {
      OnTrackFormatInfoChanged();

      base.OnShown(e);
    }

    private void buttonSetFormat_Click(object sender, EventArgs e)
    {
      var control = sender as Control;

      contextMenuStripFormats.Show(control, control.PointToClient(Cursor.Position));
    }

    private void textBoxOutputFileName_TextChanged(object sender, EventArgs e)
    {
      var formattedFilename = (trackForDefaults == null)
        ? textBoxOutputFileName.Text
        : trackForDefaults.ToString(textBoxOutputFileName.Text, trackFormatInfo);

      labelExample.Text = ThbgmExtractor.GetText("CollectiveTrackTagEditDialog_ExampleFormatString", formattedFilename);
    }

    private void contextMenuStripFormats_ItemsClick(object sender, EventArgs e)
    {
      if (sender is TrackInfoFormatPresetToolStripMenuItem) {
        var item = sender as TrackInfoFormatPresetToolStripMenuItem;

        textBoxOutputFileName.Text = item.Format;
      }
      else if (sender is TrackInfoFormatToolStripMenuItem) {
        var item = sender as TrackInfoFormatToolStripMenuItem;

        if (0 < textBoxOutputFileName.TextLength)
          textBoxOutputFileName.Text = string.Format("{0} {1}", textBoxOutputFileName.Text, item.Format);
        else
          textBoxOutputFileName.Text = item.Format;
      }
    }

    private bool lockChanged = false;

    private void checkBoxZeroPadding_CheckedChanged(object sender, EventArgs e)
    {
      if (lockChanged)
        return;

      try {
        lockChanged = true;

        trackFormatInfo.ZeroPadding = checkBoxZeroPadding.Checked;

        OnTrackFormatInfoChanged();
      }
      finally {
        lockChanged = false;
      }
    }

    private void checkBoxUseWaveDashInsteadOfFullWidthTilde_CheckedChanged(object sender, EventArgs e)
    {
      if (lockChanged)
        return;

      try {
        lockChanged = true;

        if (checkBoxUseWaveDashInsteadOfFullWidthTilde.Checked)
          trackFormatInfo.TitleUnificationOptions = (trackFormatInfo.TitleUnificationOptions & ~TitleUnificationOptions.TitleDelimiterOptionMask) | TitleUnificationOptions.WaveDash;
        else
          trackFormatInfo.TitleUnificationOptions = (trackFormatInfo.TitleUnificationOptions & ~TitleUnificationOptions.TitleDelimiterOptionMask) | TitleUnificationOptions.FullWidthTilde;

        OnTrackFormatInfoChanged();
      }
      finally {
        lockChanged = false;
      }
    }

    private void checkBoxUseTwoSpacesInsteadOfIdeographicSpace_CheckedChanged(object sender, EventArgs e)
    {
      if (lockChanged)
        return;

      try {
        lockChanged = true;

        if (checkBoxUseTwoSpacesInsteadOfIdeographicSpace.Checked)
          trackFormatInfo.TitleUnificationOptions = (trackFormatInfo.TitleUnificationOptions & ~TitleUnificationOptions.SpaceOptionMask) | TitleUnificationOptions.DoubleSpace;
        else
          trackFormatInfo.TitleUnificationOptions = (trackFormatInfo.TitleUnificationOptions & ~TitleUnificationOptions.SpaceOptionMask) | TitleUnificationOptions.SingleIdeographicSpace;

        OnTrackFormatInfoChanged();
      }
      finally {
        lockChanged = false;
      }
    }

    private void OnTrackFormatInfoChanged()
    {
      checkBoxZeroPadding.Checked = trackFormatInfo.ZeroPadding;

      checkBoxUseWaveDashInsteadOfFullWidthTilde.Checked = ((trackFormatInfo.TitleUnificationOptions & TitleUnificationOptions.TitleDelimiterOptionMask) == TitleUnificationOptions.WaveDash);
      checkBoxUseTwoSpacesInsteadOfIdeographicSpace.Checked = ((trackFormatInfo.TitleUnificationOptions & TitleUnificationOptions.SpaceOptionMask) == TitleUnificationOptions.DoubleSpace);

      textBoxOutputFileName_TextChanged(null, EventArgs.Empty);
    }

    private void OnTrackForDefaultsChanged()
    {
      if (trackForDefaults == null) {
        textBoxAlbum.Text         = string.Empty;
        textBoxPerformers.Text    = string.Empty;
        textBoxAlbumArtists.Text  = string.Empty;
        textBoxComposers.Text     = string.Empty;
        textBoxGenres.Text        = string.Empty;
        textBoxComment.Text       = string.Empty;
      }
      else {
        var trackTag = new TrackTag(trackForDefaults, this);

        textBoxAlbum.Text         = trackTag.Album;
        textBoxPerformers.Text    = trackTag.JoinedPerformers;
        textBoxAlbumArtists.Text  = trackTag.JoinedAlbumArtists;
        textBoxComposers.Text     = trackTag.JoinedComposers;
        textBoxGenres.Text        = trackTag.JoinedGenres;
        textBoxComment.Text       = trackTag.Comment;
      }
    }

    private void buttonOK_Click(object sender, EventArgs e)
    {
      DialogResult = DialogResult.OK;

      Close();
    }

    private TrackInfo trackForDefaults = null;
    private TrackFormatInfo trackFormatInfo = new TrackFormatInfo();
  }
}