//  
//  Author:
//       smdn <smdn@smdn.jp>
// 
//  Copyright (c) 2004-2014 smdn
// 
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
// 
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
//  GNU General Public License for more details.
//  
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Diagnostics;
using System.IO;
using System.Linq;

using Smdn.Collections;
using Smdn.Formats.Ini;

namespace Smdn.Applications.ThbgmExtractor {
  public class EncoderConfig {
    /*
     * class members
     */
    public static readonly string ConfigFileName = @"Encoders.conf";
    public static readonly EncoderConfig Empty = new EncoderConfig(ThbgmExtractor.GetText("EncoderConfig_Empty_Name"), ".wav");

    public static ReadOnlyCollection<EncoderConfig> LoadFrom(string configFile)
    {
      var configs = new List<EncoderConfig>();

      configs.Add(Empty);

      if (!File.Exists(configFile))
        return configs.AsReadOnly();

      var document = IniDocument.Load(configFile);

      foreach (var section in document.Sections) {
        if (section.IsDefaultSection)
          continue;

        try {
          configs.Add(new EncoderConfig(section));
        }
        catch (Exception ex) {
          // ignore invalid section
          Console.WriteLine("invalid encoder config '{0}': {1}", section.Name, ex.Message);
        }
      }

      return configs.AsReadOnly();
    }

    public static bool IsNullOrEmpty(EncoderConfig encoderConfig)
    {
      if (encoderConfig == null)
        return true;
      else if (encoderConfig == Empty)
        return true;
      else
        return false;
    }

    /*
     * instance members
     */
    public string Name {
      get; private set;
    }

    public string ExecutableFile {
      get; private set;
    }

    public string Arguments {
      get; private set;
    }

    public string OutputExtension {
      get; private set;
    }

    public bool UseStandardInputAsOutput {
      get; private set;
    }

    private Dictionary<string, string> additionalEnvironmentVariables;

    public IDictionary<string, string> AdditionalEnvironmentVariables {
      get { return additionalEnvironmentVariables.AsReadOnly(); }
    }

    private EncoderConfig(string name, string outputExtension)
    {
      this.Name = name;
      this.OutputExtension = outputExtension;
      this.UseStandardInputAsOutput = false;
    }

    private EncoderConfig(IniSection section)
    {
      this.Name = section.Name;

      /*
       * required keys
       */
      this.ExecutableFile     = section["executable"];
      this.Arguments          = section["arguments"];
      this.OutputExtension    = section["extension"];

      if (string.IsNullOrEmpty(this.ExecutableFile))
        throw new ArgumentException("'executable' is null or empty");
      else
        this.ExecutableFile = Environment.ExpandEnvironmentVariables(this.ExecutableFile);

      if (string.IsNullOrEmpty(this.Arguments))
        throw new ArgumentException("'arguments' is null or empty");

      if (string.IsNullOrEmpty(this.OutputExtension))
        throw new ArgumentException("'extension' is null or empty");

      /*
       * optional keys
       */
      this.UseStandardInputAsOutput = section.Get("stdin", false, delegate(string entry) {
        if (entry == null)
          return false;
        else if (string.Equals(entry, "yes", StringComparison.OrdinalIgnoreCase))
          return true;
        else
          return false;
      });

      if (!this.OutputExtension.StartsWith('.'))
        this.OutputExtension = "." + this.OutputExtension;

      additionalEnvironmentVariables = new Dictionary<string, string>(StringComparer.OrdinalIgnoreCase);

      foreach (var entry in section) {
        if (entry.Key.StartsWith("env[", StringComparison.OrdinalIgnoreCase) && entry.Key.EndsWith(']')) {
          var name = entry.Key.Substring(4, entry.Key.Length - 5); // remove 'env[' and  ']'

          additionalEnvironmentVariables[name] = entry.Value;
        }
      }
    }

    private string FindExecutableFullPath()
    {
      var psi = Shell.CreateProcessStartInfo(Runtime.IsRunningOnWindows ? "where" : "which", ExecutableFile);

      psi.CreateNoWindow = true;

      foreach (var envvar in AdditionalEnvironmentVariables) {
        psi.EnvironmentVariables[envvar.Key] = Environment.ExpandEnvironmentVariables(envvar.Value);
      }

      psi.RedirectStandardOutput = true;

      using (var process = Process.Start(psi)) {
        try {
          var fullPath = process.StandardOutput.ReadToEnd().Trim();

          if (!process.WaitForExit(1000))
            process.Kill();

          if (File.Exists(fullPath))
            return fullPath;
          else
            return null;
        }
        catch (Exception ex) {
          Console.Error.WriteLine(ex);
          return null;
        }
      }
    }

    public bool GetExecutableFullPathAndVersion(out string executableFullPath, out FileVersionInfo versionInfo)
    {
      executableFullPath = FindExecutableFullPath();
      versionInfo = null;

      if (string.IsNullOrEmpty(executableFullPath)) {
        return false;
      }
      else {
        versionInfo = FileVersionInfo.GetVersionInfo(executableFullPath);
        return true;
      }
    }

    public override string ToString()
    {
      return Name;
    }
  }
}
