//  
//  Author:
//       smdn <smdn@smdn.jp>
// 
//  Copyright (c) 2004-2017 smdn
// 
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
// 
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
//  GNU General Public License for more details.
//  
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Diagnostics;
using System.Globalization;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading;
using System.Windows.Forms;

using Smdn.Collections;
using Smdn.Formats.Thbgm;
using Smdn.IO;
using Smdn.Threading;
using Smdn.Windows.Forms.ThbgmPlayer;

namespace Smdn.Applications.ThbgmExtractor {
  public class ExtractBackgroundWorker : BackgroundWorker {
    public static IEnumerable<string> GetOutputFilePaths(ExtractionConfig config)
    {
      var doEncode = !EncoderConfig.IsNullOrEmpty(config.EncoderConfig);
      var generatesEncodeSource = !(config.EncoderConfig == null || config.EncoderConfig.UseStandardInputAsOutput);

      foreach (var track in config.Tracks) {
        if (doEncode)
          yield return ExtractTrackTask.GetEncodedFilePath(config, track);

        if (generatesEncodeSource)
          yield return ExtractTrackTask.GetExtractedFilePath(config, track);
      }
    }

    public ExtractBackgroundWorker()
    {
      WorkerReportsProgress = true;
      WorkerSupportsCancellation = true;
    }

    protected override void OnDoWork(DoWorkEventArgs e)
    {
      base.OnDoWork(e);

      var config = e.Argument as ExtractionConfig;
      var progress = new ExtractProgress(config.Tracks.Count());

      using (QueueingBackgroundWorker
             workerExtract = new QueueingBackgroundWorker(),
             workerEncode = new QueueingBackgroundWorker(),
             workerSetTag = new QueueingBackgroundWorker()) {
        var workerExtractCancelled = false;
        var workerEncodeCancelled = false;
        var workerSetTagCancelled = false;
        var trackAndOutputPaths = new Dictionary<Track, string>();

        workerExtract.WorkerReportsProgress = true;
        workerExtract.Cancelled += delegate { workerExtractCancelled = true; };
        workerExtract.ProgressChanged += delegate(object sender, ProgressChangedEventArgs progChangedArgs) {
          ReportProgress(progChangedArgs.UserState as IExtractTrackProgress);
        };

        workerEncode.WorkerReportsProgress = true;
        workerEncode.Cancelled += delegate { workerEncodeCancelled = true; };
        workerEncode.ProgressChanged += delegate(object sender, ProgressChangedEventArgs progChangedArgs) {
          ReportProgress(progChangedArgs.UserState as IExtractTrackProgress);
        };

        workerSetTag.WorkerReportsProgress = true;
        workerSetTag.Cancelled += delegate { workerSetTagCancelled = true; };
        workerSetTag.ProgressChanged += delegate(object sender, ProgressChangedEventArgs progChangedArgs) {
          ReportProgress(progChangedArgs.UserState as IExtractTrackProgress);
        };

        /*
         * extraction
         */
        workerExtract.DoWork += delegate(object sender, DoWorkEventArgs doWorkArgs) {
          var task = doWorkArgs.Argument as ExtractTrackTask;
          var worker = sender as BackgroundWorker;

          task.Extract(worker);

          if (worker.CancellationPending)
            doWorkArgs.Cancel = true;
          else
            doWorkArgs.Result = task;
        };
        workerExtract.RunWorkerCompleted += delegate(object sender, RunWorkerCompletedEventArgs completedArgs) {
          if (completedArgs.Error == null) {
            if (!completedArgs.Cancelled)
              workerEncode.QueueWorkerAsync(completedArgs.Result as ExtractTrackTask);
          }
          else {
            Trace.ErrorWriteLine("An exception occurred while extracting: {0}", completedArgs.Error);

            progress.FailedCount++;
          }
        };

        /*
         * encoding
         */
        workerEncode.DoWork += delegate(object sender, DoWorkEventArgs doWorkArgs) {
          var task = doWorkArgs.Argument as ExtractTrackTask;
          var worker = sender as BackgroundWorker;

          task.Encode(worker);

          if (worker.CancellationPending)
            doWorkArgs.Cancel = true;
          else
            doWorkArgs.Result = task;
        };
        workerEncode.RunWorkerCompleted += delegate(object sender, RunWorkerCompletedEventArgs completedArgs) {
          if (completedArgs.Error == null) {
            if (!completedArgs.Cancelled)
              workerSetTag.QueueWorkerAsync(completedArgs.Result as ExtractTrackTask);
          }
          else {
            Trace.ErrorWriteLine("An exception occurred while encoding: {0}", completedArgs.Error);

            progress.FailedCount++;
          }
        };

        /*
         * setting tags
         */
        workerSetTag.DoWork += delegate(object sender, DoWorkEventArgs doWorkArgs) {
          var task = doWorkArgs.Argument as ExtractTrackTask;
          var worker = sender as BackgroundWorker;

          task.SetTag(worker);

          if (worker.CancellationPending)
            doWorkArgs.Cancel = true;
          else
            doWorkArgs.Result = task;
        };
        workerSetTag.RunWorkerCompleted += delegate(object sender, RunWorkerCompletedEventArgs completedArgs) {
          if (completedArgs.Error == null) {
            if (!completedArgs.Cancelled) {
              progress.SucceededCount++;

              var task = completedArgs.Result as ExtractTrackTask;

              trackAndOutputPaths.Add(task.Track, task.OutputFile);

              task.SetCompleted();

              ReportProgress(0, task);
            }
          }
          else {
            Trace.ErrorWriteLine("An exception occurred while tagging: {0}", completedArgs.Error);

            progress.FailedCount++;
          }
        };

        /*
         * invoke process
         */
        progress.StatusText = ThbgmExtractor.GetText("Extractor_Extracting",
                                                     config.Product.Title);

        ReportProgress(progress);

        foreach (var track in config.Tracks) {
          var t = new ExtractTrackTask(config, track);

          workerExtract.QueueWorkerAsync(t);
          workerExtract.ReportProgress(0, t);
        }

        /*
         * wait for completion or abortion
         */
        for (;;) {
          if (!workerExtract.IsBusy && !workerEncode.IsBusy && !workerSetTag.IsBusy) {
            // all done
            break;
          }
          else {
            // sleep 500ms
            for (var i = 0; i < 10; i++) {
              Thread.Sleep(50);

              if (CancellationPending)
                break;
            }

            progress.StatusText = ThbgmExtractor.GetText("Extractor_Progress",
                                                         progress.TotalCount - (progress.SucceededCount + progress.FailedCount),
                                                         progress.SucceededCount,
                                                         progress.FailedCount);

            ReportProgress(progress);

            if (CancellationPending)
              break;
            else
              continue;
          }
        }

        if (CancellationPending) {
          workerExtract.CancelPendingAndRunningWorkerAsync();
          workerEncode.CancelPendingAndRunningWorkerAsync();
          workerSetTag.CancelPendingAndRunningWorkerAsync();

          e.Cancel = true;

          progress.StatusText = ThbgmExtractor.GetText("Extractor_Aborting");

          ReportProgress(progress);

          // wait for cancelled
          for (var i = 0; i < 20; i++) {
            if (workerExtractCancelled && workerEncodeCancelled && workerSetTagCancelled)
              // cancelled
              break;

            Thread.Sleep(50);
          }

          progress.StatusText = ThbgmExtractor.GetText("Extractor_Aborted");
        }
        else {
          /*
           * generate playlist
           */
          if (config.PlaylistFormat != null) {
            progress.StatusText =  ThbgmExtractor.GetText("Extractor_GeneratingPlaylist");

            ReportProgress(progress);

            try {
              TrackPlaylist.Save(config.PlaylistPath, trackAndOutputPaths, config.PlaylistFormat);
            }
            catch (Exception ex) {
              Trace.ErrorWriteLine("An exception occurred while generating playlist: {0}", ex);
            }
          }

          progress.StatusText = ThbgmExtractor.GetText("Extractor_Completed");

          e.Result = config;
        }

        ReportProgress(progress);
      } // using
    }

    private void ReportProgress(IProgress progress)
    {
      ReportProgress((int)(100.0 * progress.ProgressPercent), progress);
    }

    private string lastExtractStatusText = string.Empty;
    private string lastExtractTrackStatusText = string.Empty;

    protected override void OnProgressChanged(ProgressChangedEventArgs e)
    {
      if (e.UserState is IExtractProgress) {
        var progress = e.UserState as IExtractProgress;

        if (progress.StatusText != lastExtractStatusText) {
          lastExtractStatusText = progress.StatusText;

          Trace.OutputWriteLine(lastExtractStatusText);
        }
      }
      else if (e.UserState is IExtractTrackProgress) {
        var progress = e.UserState as IExtractTrackProgress;

        if (progress.StatusText != lastExtractTrackStatusText) {
          lastExtractTrackStatusText = progress.StatusText;

          Trace.OutputWriteLine("[{0}] {1}", progress.Track.TrackInfo.Title, lastExtractTrackStatusText);
        }
      }

      base.OnProgressChanged(e);
    }

    private static class Trace {
      public static void OutputWriteLine(string format, params object[] arg)
      {
        Console.Out.Write("{0} ", DateTime.Now.ToString("o"));
        Console.Out.WriteLine(format, arg);
      }

      public static void ErrorWriteLine(string format, params object[] arg)
      {
        Console.Error.Write("{0} ", DateTime.Now.ToString("o"));
        Console.Error.WriteLine(format, arg);
      }
    }

    private class EscapedFileProcessor {
      public string InputFilePath {
        get; private set;
      }

      public string OutputFilePath {
        get; private set;
      }

      public EscapedFileProcessor(string inputFilePath, string outputFilePath)
      {
        this.InputFilePath = inputFilePath;
        this.OutputFilePath = outputFilePath;
      }

      public void Process(bool deleteInputIfSuccess, Action<string, string> action)
      {
        var escapedInputFilePath = InputFilePath;
        var escapedOutputFilePath = OutputFilePath;
        var escaped =
          (InputFilePath != null && PathUtils.ContainsShellEscapeChar(escapedInputFilePath)) ||
          PathUtils.ContainsShellEscapeChar(escapedOutputFilePath);

        if (escaped) {
          if (InputFilePath == null) {
            // XXX: check existence
            escapedOutputFilePath = PathUtils.ChangeFileName(escapedOutputFilePath, Path.GetRandomFileName());
          }
          else {
            escapedInputFilePath = PathUtils.RenameUnique(escapedInputFilePath);
            escapedOutputFilePath = Path.ChangeExtension(escapedInputFilePath, Path.GetExtension(OutputFilePath));
          }
        }

        if (File.Exists(escapedOutputFilePath))
          // overwrite
          File.Delete(escapedOutputFilePath);

        action(escapedInputFilePath, escapedOutputFilePath);

        if (escaped && File.Exists(escapedOutputFilePath)) {
          if (File.Exists(OutputFilePath))
            // overwrite
            File.Delete(OutputFilePath);

          File.Move(escapedOutputFilePath, OutputFilePath);
        }

        if (InputFilePath != null && deleteInputIfSuccess && File.Exists(OutputFilePath)) {
          if (escaped)
            File.Delete(escapedInputFilePath);
          else
            File.Delete(InputFilePath);
        }
      }
    }

    public interface IProgress {
      double ProgressPercent { get; }
      string StatusText { get; }
    }

    public interface IExtractProgress : IProgress {
      int TotalCount { get; }
      int FailedCount { get; }
      int SucceededCount { get; }
    }

    private class ExtractProgress : IExtractProgress {
      public string StatusText {
        get; set;
      }

      public int TotalCount {
        get; private set;
      }

      public int SucceededCount {
        get; set;
      }

      public int FailedCount {
        get; set;
      }

      public double ProgressPercent {
        get { return (SucceededCount + FailedCount) / (double)TotalCount; }
      }

      public ExtractProgress(int totalCount)
      {
        StatusText = string.Empty;
        TotalCount = totalCount;
        SucceededCount = 0;
        FailedCount = 0;
      }
    }

    public interface IExtractTrackProgress : IProgress {
      Track Track { get; }
    }

    private class ExtractTrackTask : IExtractTrackProgress {
      public ExtractionConfig Config {
        get; private set;
      }

      public Track Track {
        get; private set;
      }

      public string OutputFile {
        get; private set;
      }

      public string StatusText {
        get; private set;
      }

      public double ProgressPercent {
        get; private set;
      }

      public ExtractTrackTask(ExtractionConfig config, Track track)
      {
        this.Config = config;
        this.Track = track;
        this.StatusText = ThbgmExtractor.GetText("Extractor_ExtractionPending");
      }

      public void SetCompleted()
      {
        StatusText = ThbgmExtractor.GetText("Extractor_TrackCompleted", Track.TrackInfo.Title);
        ProgressPercent = 1.0;
      }

      public static string GetExtractedFilePath(ExtractionConfig config, Track track)
      {
        return GetFilePath(config, track) + ".wav";
      }

      public static string GetEncodedFilePath(ExtractionConfig config, Track track)
      {
        return GetFilePath(config, track) + (config.EncoderConfig ?? EncoderConfig.Empty).OutputExtension;
      }

      private static string GetFilePath(ExtractionConfig config, Track track)
      {
        return Path.Combine(config.OutputDirectory, config.TrackTags[track.TrackInfo].FormattedOutputFileName);
      }

      private static ProcessStartInfo CreateEncoderProcessStartInfo(EncoderConfig encoderConfig, string inputFilePath, string outputFilePath)
      {
        var psi = Shell.CreateProcessStartInfo(encoderConfig.ExecutableFile, encoderConfig.Arguments);

        psi.WorkingDirectory = Path.GetDirectoryName(Application.ExecutablePath);
        psi.CreateNoWindow = true;

        psi.EnvironmentVariables.Add("THBGM_INPUTFILEPATH", inputFilePath);
        psi.EnvironmentVariables.Add("THBGM_OUTPUTFILEPATH", outputFilePath);

        foreach (var envvar in encoderConfig.AdditionalEnvironmentVariables) {
          psi.EnvironmentVariables[envvar.Key] = Environment.ExpandEnvironmentVariables(envvar.Value);
        }

        Trace.OutputWriteLine("invoking encoder '{0}'", encoderConfig.Name);
        Trace.OutputWriteLine("  command line: {0} {1}", psi.FileName, psi.Arguments);
#if DEBUG
        foreach (System.Collections.DictionaryEntry entry in psi.EnvironmentVariables) {
          Trace.OutputWriteLine("  {0}={1}", entry.Key, entry.Value);
        }
#else
        Trace.OutputWriteLine("  THBGM_INPUTFILEPATH ={0}", inputFilePath);
        Trace.OutputWriteLine("  THBGM_OUTPUTFILEPATH={0}", outputFilePath);
#endif

        return psi;
      }

      private static void EncoderOutputDataReceived(object sender, DataReceivedEventArgs e)
      {
        if (e.Data != null)
          Trace.OutputWriteLine("[Encoder] {0}", e.Data);
      }

      private static void EncoderErrorDataReceived(object sender, DataReceivedEventArgs e)
      {
        if (e.Data != null)
          Trace.ErrorWriteLine("[Encoder] {0}", e.Data);
      }

      private static byte[] CreateRiffInfoListChunkFromTrackTag(TrackTag trackTag)
      {
        return WaveStream.CreateRiffInfoListChunk(new Dictionary<FourCC, string>() {
          {new FourCC("IPRD"), trackTag.Album},
          {new FourCC("IAAR"), trackTag.JoinedAlbumArtists}, // XXX: not displayed on explorer
          {new FourCC("INAM"), trackTag.Title},
          {new FourCC("IART"), trackTag.JoinedPerformers},
          {new FourCC("ICOP"), trackTag.Copyright},
          {new FourCC("ITRK"), trackTag.Track.ToString("D", CultureInfo.InvariantCulture)},
          {new FourCC("IFRM"), trackTag.TrackCount.ToString("D", CultureInfo.InvariantCulture)},
          {new FourCC("IRIP"), ThbgmExtractor.VersionString},
          {new FourCC("ISFT"), ThbgmExtractor.VersionString},
          {new FourCC("ICRD"), trackTag.Year.ToString("D", CultureInfo.InvariantCulture)},
          {new FourCC("IGNR"), trackTag.JoinedGenres},
          {new FourCC("ICMT"), trackTag.Comment},
        });
      }
      
      public void Extract(BackgroundWorker worker)
      {
        var pipedEncode = (Config.EncoderConfig ?? EncoderConfig.Empty).UseStandardInputAsOutput;

        StatusText = ThbgmExtractor.GetText("Extractor_StartExtraction");
        ProgressPercent = 0.0;

        worker.ReportProgress(0, this);

        using (var inputStream = Track.GetStream(Config.BgmSourcePath)) {
          var lastPart = EffectAppliedBgmStreamPart.EndOfStream;
          var lastRepeatedTimes = inputStream.RepeatedTimes;
          var length = (double)inputStream.Length;

          var fileProcessor = new EscapedFileProcessor(null,
                                                       pipedEncode
                                                         ? GetEncodedFilePath(Config, Track)
                                                         : GetExtractedFilePath(Config, Track));

          fileProcessor.Process(false, delegate(string notused, string output) {
            StatusText = ThbgmExtractor.GetText("Extractor_ExtractingTrack", Track.TrackInfo.Title, output);
            ProgressPercent = 0.0;

            worker.ReportProgress(0, this);

            using (var outputStream = pipedEncode
                   ? (Stream)new PipeOutStream(CreateEncoderProcessStartInfo(Config.EncoderConfig, null, output), EncoderOutputDataReceived, EncoderErrorDataReceived)
                   : (Stream)File.OpenWrite(output)) {
              // preset RIFF INFO tags if needed
              var riffInfoList = (Config.EnableTagging && Config.TagFormat == TagFormat.WavRiffInfoOnly)
                ? CreateRiffInfoListChunkFromTrackTag(Config.TrackTags[Track.TrackInfo])
                : EmptyArray<byte>.Instance;

              if (pipedEncode) {
                var header = WaveStream.CreateRiffWaveHeaderFrom(inputStream, false, 0, riffInfoList.Length);

                outputStream.Write(header, 0, header.Length);
              }
              else {
                var header = WaveStream.CreateRiffWaveHeaderFrom(inputStream, false, 2048, riffInfoList.Length);

                outputStream.SetLength(header.Length + inputStream.Length + riffInfoList.Length);
                outputStream.Write(header, 0, header.Length);
              }
              
              var data = new byte[inputStream.Format.BytesPerSecond];
              var written = 0L;
              var reportCountUp = 0;
              var reportCountMax = pipedEncode ? 5 : 15;

              for (;;) {
                if (worker.CancellationPending)
                  return;

                if (inputStream.Part != lastPart || inputStream.RepeatedTimes != lastRepeatedTimes) {
                  lastPart          = inputStream.Part;
                  lastRepeatedTimes = inputStream.RepeatedTimes;

                  switch (lastPart) {
                    case EffectAppliedBgmStreamPart.Intro:            StatusText = ThbgmExtractor.GetText("Extractor_ExtractingIntroPart"); break;
                    case EffectAppliedBgmStreamPart.Loop:             StatusText = ThbgmExtractor.GetText("Extractor_ExtractingLoopPart", lastRepeatedTimes); break;
                    case EffectAppliedBgmStreamPart.FadeOut:          StatusText = ThbgmExtractor.GetText("Extractor_ExtractingFadeOutPart", lastRepeatedTimes); break;
                    case EffectAppliedBgmStreamPart.TrailingSilence:  StatusText = ThbgmExtractor.GetText("Extractor_ExtractingTrailingSilence"); break;
                    default: break;
                  }

                  reportCountUp = reportCountMax;
                }

                var read = inputStream.Read(data, 0, data.Length);

                if (read <= 0)
                  break;

                outputStream.Write(data, 0, read);

                written += read;

                if (reportCountMax <= reportCountUp++) {
                  ProgressPercent = written / length;

                  worker.ReportProgress(0, this);

                  reportCountUp = 0;
                }
              }

              outputStream.Write(riffInfoList, 0, riffInfoList.Length);

              outputStream.Flush();
            }
          });

          if (worker.CancellationPending)
            StatusText = ThbgmExtractor.GetText("Extractor_ExtractTrackAborted", Track.TrackInfo.Title);
          else
            StatusText = ThbgmExtractor.GetText("Extractor_ExtractTrackFinished", Track.TrackInfo.Title);

          ProgressPercent = 1.0;

          worker.ReportProgress(0, this);

          OutputFile = fileProcessor.OutputFilePath;
        }
      }

      public void Encode(BackgroundWorker worker)
      {
        if (OutputFile == null)
          throw new InvalidOperationException();

        if (EncoderConfig.IsNullOrEmpty(Config.EncoderConfig))
          return;
        else if (Config.EncoderConfig.UseStandardInputAsOutput)
          return;

        var fileProcessor = new EscapedFileProcessor(OutputFile, GetEncodedFilePath(Config, Track));

        fileProcessor.Process(!Config.SaveSourceAfterEncode, delegate(string input, string output) {
          var psi = CreateEncoderProcessStartInfo(Config.EncoderConfig, input, output);

          psi.RedirectStandardOutput = true;
          psi.RedirectStandardError = true;
          psi.UseShellExecute = false; // redirecting stdout and stderr

          using (var encoderProcess = new Process()) {
            // start encoder process
            encoderProcess.StartInfo = psi;
            encoderProcess.OutputDataReceived += EncoderOutputDataReceived;
            encoderProcess.ErrorDataReceived += EncoderErrorDataReceived;

            encoderProcess.Start();
            encoderProcess.BeginOutputReadLine();
            encoderProcess.BeginErrorReadLine();

            StatusText = ThbgmExtractor.GetText("Extractor_EncodingTrack", Track.TrackInfo.Title);
            ProgressPercent = double.NaN;

            while (!encoderProcess.HasExited) {
              if (worker.CancellationPending) {
                encoderProcess.Kill();

                return;
              }

              worker.ReportProgress(0, this);

              encoderProcess.WaitForExit(250);
            }
          }
        });

        if (worker.CancellationPending)
          StatusText = ThbgmExtractor.GetText("Extractor_AbortedEncodingTrack", Track.TrackInfo.Title);
        else
          StatusText = ThbgmExtractor.GetText("Extractor_EncodeTrackFinished", Track.TrackInfo.Title);

        ProgressPercent = 1.0;

        worker.ReportProgress(0, this);

        OutputFile = fileProcessor.OutputFilePath;
      }

      public void SetTag(BackgroundWorker worker)
      {
        if (OutputFile == null)
          throw new InvalidOperationException();

        if (!Config.EnableTagging)
          return;

        if (worker.CancellationPending)
          return;

        StatusText = ThbgmExtractor.GetText("Extractor_SetTagReading", Track.TrackInfo.Title);
        ProgressPercent = 0.0;

        worker.ReportProgress(0, this);

        using (var file = TagLib.File.Create(OutputFile)) {
          var editTrackTag = true;

          // specify tag format for WAV files
          if (file.MimeType.EndsWith("/wav", StringComparison.OrdinalIgnoreCase) ||
              file.MimeType.EndsWith("/wave", StringComparison.OrdinalIgnoreCase)) {
            editTrackTag = (Config.TagFormat == TagFormat.WavRiffInfoWithID3);
          }

          if (editTrackTag) {
            var trackTag = Config.TrackTags[Track.TrackInfo];

            StatusText = ThbgmExtractor.GetText("Extractor_SetTagWriting", Track.TrackInfo.Title);
            ProgressPercent = 0.25;

            worker.ReportProgress(0, this);

            trackTag.CopyTo(file.Tag, true);

            // modify tags
            var tags = new List<TagLib.Tag>();

            if (file.Tag is TagLib.CombinedTag)
              tags.AddRange((file.Tag as TagLib.CombinedTag).Tags);
            else
              tags.Add(file.Tag);

            foreach (var tag in tags) {
              if (tag is TagLib.Mpeg4.AppleTag) {
                var isomMeta = tag as TagLib.Mpeg4.AppleTag;
                var tooBoxType = (TagLib.ByteVector)ByteString.ToByteArray("\x00a9too");

                var originalEncoder = isomMeta.GetText(tooBoxType);
                var encoder = (originalEncoder == null || originalEncoder.Length == 0)
                  ? ThbgmExtractor.VersionString
                  : TrackTag.GetJoined(originalEncoder.Append(ThbgmExtractor.VersionString));

                // encoder
                isomMeta.SetData(tooBoxType,
                                 TagLib.ByteVector.FromString(encoder, TagLib.StringType.UTF8),
                                 (uint)TagLib.Mpeg4.AppleDataBox.FlagType.ContainsText);

                // copyright (TagLib uses ISO 'cprt' box)
                isomMeta.SetData((TagLib.ByteVector)ByteString.ToByteArray("\x00a9cpy"),
                                 TagLib.ByteVector.FromString(trackTag.Copyright, TagLib.StringType.UTF8),
                                 (uint)TagLib.Mpeg4.AppleDataBox.FlagType.ContainsText);
              }
              else if (tag is TagLib.Id3v2.Tag) {
                var id3v2 = tag as TagLib.Id3v2.Tag;
                var tencIdent = (TagLib.ByteVector)ByteString.ToByteArray("TENC");

                // encoded by
                id3v2.SetTextFrame(tencIdent, ThbgmExtractor.VersionString);
              }
              else if (tag is TagLib.Riff.InfoTag) {
                var riffInfo = tag as TagLib.Riff.InfoTag;

                // software
                riffInfo.SetValue((TagLib.ByteVector)ByteString.ToByteArray("ISFT"),
                                  ThbgmExtractor.VersionString);
              }
            }

            file.Save();
          } // if file is wav
        } // using file

        StatusText = ThbgmExtractor.GetText("Extractor_SetTagFinished", Track.TrackInfo.Title);
        ProgressPercent = 1.0;

        worker.ReportProgress(0, this);
      }
    }
  } // class ExtractBackgroundWorker
}
