//  
//  Author:
//       smdn <smdn@smdn.jp>
// 
//  Copyright (c) 2004-2017 smdn
// 
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
// 
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
//  GNU General Public License for more details.
//  
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

using System;
using System.IO;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.ComponentModel;
using System.Linq;
using System.Text;
using System.Xml;

using Smdn.Collections;
using Smdn.Formats.Ini;
using Smdn.Formats.Playlists;
using Smdn.Formats.Thbgm;
using Smdn.IO;
using Smdn.Windows.Forms.ThbgmPlayer;
using Smdn.Xml;

namespace Smdn.Applications.ThbgmExtractor {
  public class ExtractionConfig : ICloneable, INotifyPropertyChanging, INotifyPropertyChanged, ITrackFormatProvider {
    public static readonly string FileExtension = ".ini";
    public static readonly string THxxBGMFadeoutSettingsFileExtension = ".fsx";

    public event PropertyChangingEventHandler PropertyChanging;
    public event PropertyChangedEventHandler PropertyChanged;

    private ProductInfo product;

    public ProductInfo Product {
      get { return product; }
      set
      {
        if (product == value)
          return;

        if (OnPropertyChanging("Product")) {
          product = value;

          OnPropertyChanged("Product");
        }
      }
    }

    private string bgmSourcePath;

    public string BgmSourcePath {
      get { return bgmSourcePath; }
      set
      {
        if (bgmSourcePath == value)
          return;

        if (OnPropertyChanging("BgmSourcePath")) {
          bgmSourcePath = value;

          OnPropertyChanged("BgmSourcePath");
        }
      }
    }

    private string outputDirectory;

    public string OutputDirectory {
      get { return outputDirectory; }
      set
      {
        if (outputDirectory == value)
          return;

        if (OnPropertyChanging("OutputDirectory")) {
          outputDirectory = value;

          OnPropertyChanged("OutputDirectory");
        }
      }
    }

    private TrackFormatInfo trackFormatInfo = Config.TrackFormatInfoDefaults.TrackFormatInfo.Clone();

    public TrackFormatInfo TrackFormatInfo {
      get { return trackFormatInfo; }
      set
      {
        if (trackFormatInfo == value)
          return;

        if (OnPropertyChanging("TrackFormatInfo")) {
          trackFormatInfo = value;

          OnPropertyChanged("TrackFormatInfo");
        }
      }
    }

    TrackFormatInfo ITrackFormatProvider.FormatPrivider {
      get { return trackFormatInfo; }
    }

    private string defaultOutputFileName = Config.TrackFormatInfoDefaults.DefaultOutputFileName;

    public string DefaultOutputFileName {
      get { return defaultOutputFileName; }
      set
      {
        if (defaultOutputFileName == value)
          return;

        if (OnPropertyChanging("DefaultOutputFileName")) {
          defaultOutputFileName = value;

          OnPropertyChanged("DefaultOutputFileName");
        }
      }
    }

    private EncoderConfig encoderConfig = Config.ExtractionConfigDefaults.EncoderConfig;

    public EncoderConfig EncoderConfig {
      get { return encoderConfig; }
      set
      {
        if (encoderConfig == value)
          return;

        if (OnPropertyChanging("EncoderConfig")) {
          encoderConfig = value;

          OnPropertyChanged("EncoderConfig");
        }
      }
    }

    private bool saveSourceAfterEncode = Config.ExtractionConfigDefaults.SaveSourceAfterEncode;

    public bool SaveSourceAfterEncode {
      get { return saveSourceAfterEncode; }
      set
      {
        if (saveSourceAfterEncode == value)
          return;

        if (OnPropertyChanging("SaveSourceAfterEncode")) {
          saveSourceAfterEncode = value;

          OnPropertyChanged("SaveSourceAfterEncode");
        }
      }
    }

    private PlaylistFormat playlistFormat = Config.ExtractionConfigDefaults.PlaylistFormat;

    public PlaylistFormat PlaylistFormat {
      get { return playlistFormat; }
      set
      {
        if (playlistFormat == value)
          return;

        if (OnPropertyChanging("PlaylistFormat")) {
          playlistFormat = value;

          OnPropertyChanged("PlaylistFormat");
        }
      }
    }

    private string playlistPath;

    public string PlaylistPath {
      get { return playlistPath; }
      set
      {
        if (playlistPath == value)
          return;

        if (OnPropertyChanging("PlaylistPath")) {
          playlistPath = value;

          OnPropertyChanged("PlaylistPath");
        }
      }
    }

    private ReadOnlyCollection<Track> tracks = Array.AsReadOnly(EmptyArray<Track>.Instance);

    public ReadOnlyCollection<Track> Tracks {
      get { return tracks; }
      set
      {
        if (tracks == null)
          throw new ArgumentNullException("Tracks");

        if (tracks == value)
          return;

        if (OnPropertyChanging("Tracks")) {
          tracks = value;
          trackTags = tracks.Select(delegate(Track t) {
            return new TrackTag(t.TrackInfo, this);
          }).ToDictionary(delegate(TrackTag tt) {
            return tt.TrackInfo;
          });

          OnPropertyChanged("Tracks");
        }
      }
    }

    private bool enableTagging = Config.ExtractionConfigDefaults.EnableTagging;

    public bool EnableTagging {
      get { return enableTagging; }
      set
      {
        if (enableTagging == value)
          return;

        if (OnPropertyChanging("EnableTagging")) {
          enableTagging = value;

          OnPropertyChanged("EnableTagging");
        }
      }
    }

    private TagFormat tagFormat = Config.ExtractionConfigDefaults.TagFormat;

    public TagFormat TagFormat {
      get { return tagFormat; }
      set {
        if (tagFormat == value)
          return;

        if (OnPropertyChanging("TagFormat")) {
          tagFormat = value;

          OnPropertyChanged("TagFormat");
        }
      }
    }

    private IDictionary<TrackInfo, TrackTag> trackTags = (new Dictionary<TrackInfo, TrackTag>()).AsReadOnly();

    public IDictionary<TrackInfo, TrackTag> TrackTags {
      get { return trackTags; }
    }

    public bool AreAnyTracksDirty {
      get
      {
        if (tracks == null)
          return false;

        foreach (var track in tracks) {
          if (track.IsDirty)
            return true;
        }

        foreach (var trackTag in trackTags.Values) {
          if (trackTag.IsDirty)
            return true;
        }

        return false;
      }
    }

    object ICloneable.Clone()
    {
      return Clone();
    }

    public virtual ExtractionConfig Clone()
    {
      var cloned = (ExtractionConfig)MemberwiseClone();

      cloned.trackFormatInfo = trackFormatInfo.Clone();
      cloned.tracks = tracks.Select(delegate(Track t) {
        return t.Clone();
      }).ToList().AsReadOnly();
      cloned.trackTags = cloned.tracks.Select(delegate(Track t) {
        return trackTags[t.TrackInfo].Clone(cloned);
      }).ToDictionary(delegate(TrackTag tt) {
        return tt.TrackInfo;
      });
      cloned.PropertyChanged = null;
      cloned.PropertyChanging = null;

      return cloned;
    }

    public void ClearTracks()
    {
      if (tracks == null)
        return;

      foreach (var track in tracks) {
        track.ClearDirty();
      }

      foreach (var trackTag in trackTags.Values) {
        trackTag.ClearDirty();
      }
    }

    public void SetTracksAndTags(Dictionary<Track, TrackTag> tracksAndTags, bool clone)
    {
      var tracks = new List<Track>();
      var trackTags = new Dictionary<TrackInfo, TrackTag>();

      if (clone) {
        foreach (var pair in tracksAndTags) {
          var cloned = pair.Key.Clone();

          tracks.Add(cloned);
          trackTags.Add(cloned.TrackInfo, pair.Value.Clone(this));
        }
      }
      else {
        foreach (var pair in tracksAndTags) {
          tracks.Add(pair.Key);
          trackTags.Add(pair.Key.TrackInfo, pair.Value);
        }
      }

      this.tracks = tracks.AsReadOnly();
      this.trackTags = trackTags.AsReadOnly();
    }

    public void Save(string file)
    {
      var document = new IniDocument(StringComparer.CurrentCulture);

      document.DefaultSection["Version"] = System.Windows.Forms.Application.ProductVersion;

      var extractionConfigSection = document["ExtractionConfig"];

      extractionConfigSection["Product"]                = product.Title;
      extractionConfigSection["BgmSourcePath"]          = bgmSourcePath;
      extractionConfigSection["OutputDirectory"]        = outputDirectory;
      extractionConfigSection["DefaultOutputFileName"]  = defaultOutputFileName;
      extractionConfigSection["TrackFormatInfo.ZeroPadding"]              = trackFormatInfo.ZeroPadding.ToString();
      extractionConfigSection["TrackFormatInfo.TitleUnificationOptions"]  = trackFormatInfo.TitleUnificationOptions.ToString();
      extractionConfigSection["SaveSourceAfterEncode"]  = saveSourceAfterEncode.ToString();
      extractionConfigSection["EncoderConfig"]          = EncoderConfig.IsNullOrEmpty(encoderConfig) ? string.Empty : encoderConfig.Name;
      extractionConfigSection["PlaylistFormat"]         = (playlistFormat == null) ? string.Empty : playlistFormat.Name;
      extractionConfigSection["PlaylistPath"]           = playlistPath;
      extractionConfigSection["EnableTagging"]          = enableTagging.ToString();
      extractionConfigSection["TagFormat"]              = tagFormat.ToString();

      foreach (var track in tracks) {
        var trackSection = document[string.Format("Track{0}", track.TrackInfo.TrackNumber)];

        //trackSection["TrackNumber"] = track.TrackInfo.TrackNumber.ToString();
        trackSection["Checked"]     = track.Checked.ToString();

        trackSection["FadingStart"]             = track.Effect.FadingStart.ToString();
        trackSection["FadingLength"]            = track.Effect.FadingLength.ToString();
        trackSection["FadingCurveType"]         = track.Effect.FadingCurveType.ToString();
        trackSection["FadingCurveCoefficient"]  = track.Effect.FadingCurveCoefficient.ToString();
        trackSection["TrailingSilenceLength"]   = track.Effect.TrailingSilenceLength.ToString();

        var trackTag = trackTags[track.TrackInfo];

        trackSection["TrackTag.OutputFileName"]   = trackTag.OutputFileName;
        trackSection["TrackTag.Title"]            = trackTag.Title;
        trackSection["TrackTag.Album"]            = trackTag.Album;
        trackSection["TrackTag.Performers"]       = TrackTag.GetIniFormatJoined(trackTag.Performers);
        trackSection["TrackTag.AlbumArtists"]     = TrackTag.GetIniFormatJoined(trackTag.AlbumArtists);
        trackSection["TrackTag.Composers"]        = TrackTag.GetIniFormatJoined(trackTag.Composers);
        trackSection["TrackTag.Genres"]           = TrackTag.GetIniFormatJoined(trackTag.Genres);
        trackSection["TrackTag.Comment"]          = trackTag.Comment;
      }

      document.Save(file);
    }

    public static ExtractionConfig LoadFrom(string file)
    {
      try {
        if (PathUtils.AreExtensionEqual(file, THxxBGMFadeoutSettingsFileExtension)) {
          try {
            return LoadFromTHxxBGMFadeoutSettings(file);
          }
          catch (NotSupportedException) {
            throw;
          }
          catch (Exception ex) {
            // ignore, try load as extraction config file
#if DEBUG
            Console.Error.WriteLine(ex);
#endif
          }
        }

        return LoadFromCore(file);
      }
      catch (NotSupportedException) {
        throw;
      }
      catch (Exception ex) {
        throw new System.IO.InvalidDataException("invalid extraction config", ex);
      }
    }

    private static ExtractionConfig LoadFromCore(string file)
    {
      var document = IniDocument.Load(file, StringComparer.CurrentCulture);
      var config = new ExtractionConfig();

      var version = ParseVersion(document.DefaultSection["Version"]);

      var extractionConfigSection = document["ExtractionConfig"];

      config.Product = ThbgmExtractor.Products.FirstOrDefault(delegate(ProductInfo product) {
        return product.Title == extractionConfigSection["Product"];
      });

      if (config.product == null)
        throw new NotSupportedException("unsupported product");

      config.bgmSourcePath   = extractionConfigSection["BgmSourcePath"];
      config.outputDirectory = extractionConfigSection["OutputDirectory"];

      config.trackFormatInfo = new TrackFormatInfo();
      config.trackFormatInfo.ZeroPadding              = extractionConfigSection.Get<bool>("TrackFormatInfo.ZeroPadding", config.trackFormatInfo.ZeroPadding, Convert.ToBoolean);
      config.trackFormatInfo.TitleUnificationOptions  = extractionConfigSection.Get<TitleUnificationOptions>("TrackFormatInfo.TitleUnificationOptions",  config.trackFormatInfo.TitleUnificationOptions, EnumUtils.Parse<TitleUnificationOptions>);
      config.defaultOutputFileName = extractionConfigSection.Get("DefaultOutputFileName", config.defaultOutputFileName);

      config.saveSourceAfterEncode = extractionConfigSection.Get<bool>("SaveSourceAfterEncode", config.saveSourceAfterEncode, Convert.ToBoolean);

      config.encoderConfig = ThbgmExtractor.EncoderConfigs.FirstOrDefault(delegate(EncoderConfig encoderConfig) {
        return encoderConfig.Name == extractionConfigSection["EncoderConfig"];
      }) ?? EncoderConfig.Empty;

      var playlistFormatName = extractionConfigSection["PlaylistFormat"];

      config.playlistFormat = PlaylistFormat.Formats.FirstOrDefault(delegate(PlaylistFormat format) {
        return format.Name == playlistFormatName;
      }) ?? config.playlistFormat;

      config.playlistPath = extractionConfigSection["PlaylistPath"];

      config.enableTagging = extractionConfigSection.Get<bool>("EnableTagging", config.enableTagging, Convert.ToBoolean);

      config.tagFormat = extractionConfigSection.Get<TagFormat>("TagFormat", config.tagFormat, EnumUtils.Parse<TagFormat>);

      var tracks = new List<Track>();
      var trackTags = new Dictionary<TrackInfo, TrackTag>();

      foreach (var trackInfo in config.product.Tracks) {
        var trackSection = document.Find(string.Format("Track{0}", trackInfo.TrackNumber));

        Track track;
        TrackTag trackTag;

        if (trackSection == null) {
          track = new Track(trackInfo);
          trackTag = new TrackTag(trackInfo, config);
        }
        else {
          var effect = Effect.CreateDefault(trackInfo);

          effect.FadingStart            = TimeSpanUtils.ParsePositive(trackSection["FadingStart"], effect.FadingStart);
          effect.FadingLength           = TimeSpanUtils.ParsePositive(trackSection["FadingLength"], effect.FadingLength);
          effect.TrailingSilenceLength  = TimeSpanUtils.ParsePositive(trackSection["TrailingSilenceLength"], effect.TrailingSilenceLength);

          var fadingCurveType = trackSection.Get<FadingCurveType>("FadingCurveType", effect.FadingCurveType, EnumUtils.Parse<FadingCurveType>);
          var fadingCurveCoefficient = trackSection.Get<double>("FadingCurveCoefficient", double.NaN/*as defaut*/, Convert.ToDouble);

          effect.SetFadingCurve(fadingCurveType, fadingCurveCoefficient, true);

          track = new Track(trackInfo, effect, bool.Parse(trackSection["Checked"]));

          trackTag = new TrackTag(trackInfo, config);

          trackTag.OutputFileName   = trackSection.Get("TrackTag.OutputFileName", trackTag.OutputFileName);
          trackTag.Title            = trackSection.Get("TrackTag.Title", trackTag.Title);
          trackTag.Album            = trackSection.Get("TrackTag.Album", trackTag.Album);
          trackTag.Performers       = trackSection.Get("TrackTag.Performers", trackTag.Performers, TrackTag.GetIniFormatSplitted);
          trackTag.AlbumArtists     = trackSection.Get("TrackTag.AlbumArtists", trackTag.AlbumArtists, TrackTag.GetIniFormatSplitted);
          trackTag.Composers        = trackSection.Get("TrackTag.Composers", trackTag.Composers, TrackTag.GetIniFormatSplitted);
          trackTag.Genres           = trackSection.Get("TrackTag.Genres", trackTag.Genres, TrackTag.GetIniFormatSplitted);
          trackTag.Comment          = trackSection.Get("TrackTag.Comment", trackTag.Comment);

          trackTag.ClearDirty();
        }

        tracks.Add(track);
        trackTags.Add(trackInfo, trackTag);
      }

      config.tracks = tracks.AsReadOnly();
      config.trackTags = trackTags.AsReadOnly();

      if (version.Major == 1 && version.Minor == 4) {
        // version 1.4.x extraction config entries
        if (extractionConfigSection.Entries.ContainsKey("PrependTrackNumber")) {
          if (extractionConfigSection.Get<bool>("PrependTrackNumber", true, Convert.ToBoolean))
            config.defaultOutputFileName = "$trackno. $title";
          else
            config.defaultOutputFileName = "$title";

          config.trackFormatInfo.TitleUnificationOptions = TitleUnificationOptions.NoUnify;
        }

        if (extractionConfigSection.Entries.ContainsKey("UseWaveDashInsteadOfFullWidthTilde") || extractionConfigSection.Entries.ContainsKey("UseTwoSpacesInsteadOfIdeographicSpace")) {
          var useWaveDashInsteadOfFullWidthTilde    = extractionConfigSection.Get<bool>("UseWaveDashInsteadOfFullWidthTilde", false, Convert.ToBoolean);
          var useTwoSpacesInsteadOfIdeographicSpace = extractionConfigSection.Get<bool>("UseTwoSpacesInsteadOfIdeographicSpace", false, Convert.ToBoolean);

          if (useWaveDashInsteadOfFullWidthTilde)
            config.trackFormatInfo.TitleUnificationOptions |= TitleUnificationOptions.WaveDash;
          else
            config.trackFormatInfo.TitleUnificationOptions |= TitleUnificationOptions.FullWidthTilde;

          if (useTwoSpacesInsteadOfIdeographicSpace)
            config.trackFormatInfo.TitleUnificationOptions |= TitleUnificationOptions.DoubleSpace;
          else
            config.trackFormatInfo.TitleUnificationOptions |= TitleUnificationOptions.SingleIdeographicSpace;
        }
      }

      return config;
    }

    private static Version ParseVersion(string version)
    {
      try {
        return new Version(version);
      }
      catch {
        return null;
      }
    }

    private bool OnPropertyChanging(string propertyName)
    {
      var e = new CancellablePropertyChangingEventArgs(propertyName);

      OnPropertyChanging(e);

      return !e.Cancel;
    }

    protected virtual void OnPropertyChanging(PropertyChangingEventArgs e)
    {
      var ev = this.PropertyChanging;

      if (ev != null)
        ev(this, e);
    }

    private void OnPropertyChanged(string propertyName)
    {
      OnPropertyChanged(new PropertyChangedEventArgs(propertyName));
    }

    protected virtual void OnPropertyChanged(PropertyChangedEventArgs e)
    {
      var ev = this.PropertyChanged;

      if (ev != null)
        ev(this, e);
    }

#region "loading/saving THxxBGMFadeoutSettings"
    private static ExtractionConfig LoadFromTHxxBGMFadeoutSettings(string fsxFile)
    {
      var xml = new XmlDocument();

      xml.Load(fsxFile);

      var config = new ExtractionConfig();

      var productGuid = xml.GetSingleNodeValueOf<Guid>("/THxxBGMFadeoutSettings/Guid/text()", Guid.Empty, XmlConvert.ToGuid);

      if (productGuid == Guid.Empty)
        config.product = null;
      else
        config.product = ThbgmExtractor.Products.FirstOrDefault(delegate(ProductInfo product) {
          return product.THxxBGMProductGuid == productGuid;
        });

      if (config.product == null)
        throw new NotSupportedException("unsupported product");

      config.bgmSourcePath = config.product.InstalledOrDefaultBgmSourcePath;

      // translate '/THxxBGMFadeoutSettings/Data/Item'
      var tracks = new List<Track>();

      for (var i = 0; i < config.product.Tracks.Count; i++) {
        var trackInfo = config.product.Tracks[i];
        var itemNode = xml.SelectSingleNode(string.Format("/THxxBGMFadeoutSettings/Data/Item[{0}]", i + 1));
        Track track;

        if (itemNode == null) {
          track = new Track(trackInfo);
        }
        else {
          var timesToRepeat = itemNode.GetSingleNodeValueOf<int>("LoopCount/text()", Effect.DefaultTimesToRepeat, Convert.ToInt32);

          // translate FadeoutTime and BgmEnd
          var fadingLength = trackInfo.StreamFormat.ToTimeSpan(itemNode.GetSingleNodeValueOf<int>("FadeoutTime/text()", 0, Convert.ToInt32));
          var fadingOffset = trackInfo.StreamFormat.ToTimeSpan(itemNode.GetSingleNodeValueOf<int>("BgmEnd/text()", 0, Convert.ToInt32));

          var effect = new Effect();

          effect.SetLength(trackInfo, timesToRepeat, fadingLength, -fadingLength + fadingOffset, TimeSpan.Zero);

          // translate FadeoutPattern
          switch (itemNode.GetSingleNodeValueOf<int>("FadeoutPattern/text()", 1, Convert.ToInt32)) {
            case 2: // linear
              effect.SetFadingCurve(FadingCurveType.Linear, true);
              break;

            case 3:
              // (linear + exponential) / 2
              effect.SetFadingCurve(FadingCurveType.Exponential, 10.0/*approx*/, true);
              break;

            case 4:
              effect.SetFadingCurve(FadingCurveType.Trigonometrical, 1.0/*approx*/, true);
              break;

            case 1: // exponential
            default:
              effect.SetFadingCurve(FadingCurveType.Exponential, 100.0/*approx*/, true);
              break;
          }

          // TODO: translate BgmBegin and BgmEnd

          track = new Track(trackInfo, effect, true);
        }

        tracks.Add(track);
      }

      config.tracks = tracks.AsReadOnly();

      return config;
    }

    public void SaveAsTHxxBGMFadeoutSettings(string file)
    {
      if (Guid.Empty.Equals(product.THxxBGMProductGuid))
        throw new InvalidOperationException(string.Format("THxxBMG does not support product '{0}'", product.Title));

      var xml = new XmlDocument();

      xml.AppendChild(xml.CreateXmlDeclaration("1.0", "UTF-8", "no")); // XXX: why standalone='no'?

      // /THxxBGMFadeoutSettings
      var root = xml.AppendChild(xml.CreateElement("THxxBGMFadeoutSettings"));

      // /THxxBGMFadeoutSettings/GameTitle
      root.AppendChild(xml.CreateElement("GameTitle")).AppendChild(xml.CreateTextNode(product.ShortName));

      // /THxxBGMFadeoutSettings/Guid (GUID with braces)
      root.AppendChild(xml.CreateElement("Guid")).AppendChild(xml.CreateTextNode(product.THxxBGMProductGuid.ToString("B").ToUpperInvariant()));

      // /THxxBGMFadeoutSettings/Data
      var data = root.AppendChild(xml.CreateElement("Data"));

      // /THxxBGMFadeoutSettings/Data/Item
      foreach (var track in tracks) {
        int fadeoutPattern;

        switch (track.Effect.FadingCurveType) {
          case FadingCurveType.Linear: fadeoutPattern = 2; break;
          case FadingCurveType.Trigonometrical: fadeoutPattern = 4; break;

          case FadingCurveType.Exponential:
          default:
            // THxxBGM default
            fadeoutPattern = 1;
            break;
        }

        var loopCount = track.TrackInfo.CalculateRepeatedTimes(track.Effect.TotalLength);
        var fadeoutTime = track.TrackInfo.StreamFormat.ToAlignedByteCount(track.Effect.FadingLength);
        var repeatedTimes = track.TrackInfo.CalculateRepeatedTimes(track.Effect.TotalLength);

        // THxxBGM does not support LoopCount=0?
        if (loopCount == 0) {
          loopCount += 1;
          repeatedTimes += 1;
        }

        var fadingOffset = track.Effect.TotalLength - track.TrackInfo.CalculateRepeatPosition(repeatedTimes);
        var bgmEnd = track.TrackInfo.StreamFormat.ToAlignedByteCount(fadingOffset);

        var item = data.AppendChild(xml.CreateElement("Item"));

        item.AppendChild(xml.CreateElement("LoopCount")).AppendChild(xml.CreateTextNode(XmlConvert.ToString(loopCount)));
        item.AppendChild(xml.CreateElement("FadeoutPattern")).AppendChild(xml.CreateTextNode(XmlConvert.ToString(fadeoutPattern)));
        item.AppendChild(xml.CreateElement("BgmBegin")).AppendChild(xml.CreateTextNode("0")); // ThbgmExtractor unsupported feature
        item.AppendChild(xml.CreateElement("BgmEnd")).AppendChild(xml.CreateTextNode(XmlConvert.ToString(bgmEnd)));
        item.AppendChild(xml.CreateElement("FadeoutTime")).AppendChild(xml.CreateTextNode(XmlConvert.ToString(fadeoutTime)));
      }

      using (var stream = File.OpenWrite(file)) {
        var settings = new XmlWriterSettings();

        settings.Encoding = new UTF8Encoding(false); // output no preamble
        settings.Indent = true;
        settings.IndentChars = "\t";
        settings.NewLineChars = Environment.NewLine;

        var writer = XmlWriter.Create(stream, settings);

        xml.WriteTo(writer);

        writer.Flush();
      }
    }
#endregion
  }
}
