//  
//  Author:
//       smdn <smdn@smdn.jp>
// 
//  Copyright (c) 2004-2017 smdn
// 
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
// 
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
//  GNU General Public License for more details.
//  
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using System.Windows.Forms;

using Smdn.IO;
using Smdn.Formats.Playlists;
using Smdn.Windows.Forms;
using Smdn.Windows.Forms.ThbgmPlayer;

namespace Smdn.Applications.ThbgmExtractor {
  public partial class ExtractionConfigDialog : Form {
    private class NullPlaylistFormat : PlaylistFormat {
      public static readonly NullPlaylistFormat Instance = new NullPlaylistFormat();

      public override string Name {
        get { return "null"; }
      }

      public override string DefaultExtension {
        get { return string.Empty; }
      }

      public override Encoding DefaultEncoding {
        get { return Encoding.Default; }
      }

      public override MimeType MimeType {
        get { return MimeType.TextPlain; }
      }

      public override Formatter Format {
        get { return FormatImpl; }
      }

      public override Parser Parse {
        get { return null; }
      }

      internal NullPlaylistFormat()
        : base()
      {
      }

      private static void FormatImpl(Playlist playlist, Stream stream, Uri baseUri, Encoding encoding)
      {
        // do nothing
      }
    }

    // for designer code
    private class PlaylistFormatDropDownList : FormattingDropDownList<PlaylistFormat> {
    }

    // for designer code
    private class EncoderConfigDropDownList : FormattingDropDownList<EncoderConfig> {
    }

    // for designer code
    private class TagFormatDropDownList : FormattingDropDownList<TagFormat> {
      public TagFormatDropDownList()
        : base()
      {
        Items.Add(TagFormat.WavRiffInfoOnly);
        Items.Add(TagFormat.WavRiffInfoWithID3);

        StringFormatter = delegate(TagFormat format) {
          switch (format) {
            case TagFormat.WavRiffInfoOnly:     return ThbgmExtractor.GetText("TagFormat_WavRiffInfoOnly");
            case TagFormat.WavRiffInfoWithID3:  return ThbgmExtractor.GetText("TagFormat_WavRiffInfoWithID3");
            default: return format.ToString();
          }
        };

        SelectedIndex = 0;
      }
    }

    public ExtractionConfig Config
    {
      get { return config; }
    }

    public ExtractionConfigDialog(ExtractionConfig baseConfig)
      : this()
    {
      if (baseConfig == null)
        throw new ArgumentNullException("baseConfig");

      config = baseConfig.Clone();
    }

    public ExtractionConfigDialog()
    {
      InitializeComponent();

      availablePlaylistFormats = PlaylistFormat.Formats.Where(delegate(PlaylistFormat f) {
        return f.Format != null;
      }).ToArray();

      dropDownListEncoderConfig.StringFormatter += delegate(EncoderConfig e) {
        return e.Name;
      };
      dropDownListPlaylistFormat.StringFormatter += delegate(PlaylistFormat f) {
        if (f == NullPlaylistFormat.Instance)
          return ThbgmExtractor.GetText("PlaylistFormat_Null_Name");
        else
          return string.Format("{0} ({1})", f.Name, f.DefaultExtension);
      };

      foreach (var e in ThbgmExtractor.EncoderConfigs) {
        dropDownListEncoderConfig.Items.Add(e);
      }

      dropDownListPlaylistFormat.Items.Add(NullPlaylistFormat.Instance);

      foreach (var f in availablePlaylistFormats) {
        dropDownListPlaylistFormat.Items.Add(f);
      }
    }

    public bool ConfigureExtraction(IWin32Window owner, bool extractSingleTrack)
    {
      checkBoxEnableTagging.Checked = config.EnableTagging;
      dropDownListTagFormat.SelectedItem = config.TagFormat;
      checkBoxSaveSourceAfterEncode.Checked = config.SaveSourceAfterEncode;
      dropDownListPlaylistFormat.SelectedItem = config.PlaylistFormat ?? NullPlaylistFormat.Instance;
      dropDownListEncoderConfig.SelectedItem = config.EncoderConfig ?? EncoderConfig.Empty;

      if (dropDownListPlaylistFormat.SelectedIndex == -1)
        dropDownListPlaylistFormat.SelectedIndex = 0;
      if (dropDownListEncoderConfig.SelectedIndex == -1)
        dropDownListEncoderConfig.SelectedIndex = 0;
      if (dropDownListTagFormat.SelectedIndex == -1)
        dropDownListTagFormat.SelectedIndex = 0;

      if (extractSingleTrack)
        return ConfigureExtractSingleTrack(owner);
      else
        return ConfigureExtractMultipleTracks(owner);
    }

#region "configure single track"
    private bool ConfigureExtractSingleTrack(IWin32Window owner)
    {
      /*
       * initialize control states
       */
      var track = config.Tracks.First();

      groupBoxEncoder.Enabled = true;
      groupBoxPlaylist.Enabled = false;
      textBoxPlaylistPath.Text = string.Empty;

      buttonChangeOutputPath.Click += delegate {
        SelectOutputFile(track);
      };
      dropDownListEncoderConfig.SelectedIndexChanged += delegate {
        ChangeOutputFileExtension();
        UpdateEncoderConfigControlState();
      };

      UpdateEncoderConfigControlState();

      Text = ThbgmExtractor.GetText("ExtractionConfigDialog_ConfigureSingleTrack_Text", track.TrackInfo.Title);

      var encoderConfig = config.EncoderConfig ?? EncoderConfig.Empty;
      var outputDirectory = config.OutputDirectory;

      if (string.IsNullOrEmpty(outputDirectory))
        outputDirectory = Environment.GetFolderPath(Environment.SpecialFolder.Desktop);

      var outputFile = PathUtils.ReplaceInvalidFileNameCharsWithBlanks(track.TrackInfo.ToString(config.DefaultOutputFileName,
                                                                                                config.TrackFormatInfo));

      textBoxOutputPath.Text = Path.Combine(outputDirectory, outputFile) + encoderConfig.OutputExtension;

      /*
       * configure
       */
      if (ShowDialog(owner) != DialogResult.OK)
        return false;

      /*
       * set configuration
       */
      config.PlaylistFormat = null; // not to generate playlist
      config.OutputDirectory = Path.GetDirectoryName(textBoxOutputPath.Text);
      config.EncoderConfig = dropDownListEncoderConfig.SelectedItem;
      config.SaveSourceAfterEncode = checkBoxSaveSourceAfterEncode.Checked;
      config.EnableTagging = checkBoxEnableTagging.Checked;
      config.TagFormat = dropDownListTagFormat.SelectedItem;

      /*
       * set tracks and tags
       */
      var trackTag = new TrackTag(track.TrackInfo, config);

      trackTag.OutputFileName = Path.GetFileNameWithoutExtension(textBoxOutputPath.Text);

      var extractTrackAndTag = new Dictionary<Track, TrackTag>(1) {
        {track, trackTag}
      };

      config.SetTracksAndTags(extractTrackAndTag, true);

      return true;
    }

    private void SelectOutputFile(Track track)
    {
      using (var dialog = new SaveFileDialog()) {
        dialog.Filter = FileDialogFilter.CreateFilterString(ThbgmExtractor.EncoderConfigs.Select(delegate(EncoderConfig e) {
          return new FileDialogFilter.Filter(EncoderConfig.IsNullOrEmpty(e)
                                             ? ThbgmExtractor.GetText("EncoderConfig_NonEncoded")
                                             : e.Name,
                                             "*" + e.OutputExtension);
        }));

        var selectedEncoderConfig = dropDownListEncoderConfig.SelectedItem;

        dialog.Title = ThbgmExtractor.GetText("ExtractionConfigDialog_SelectOutputFile_DialogTitle", track.TrackInfo.Title);
        dialog.FilterIndex = ThbgmExtractor.EncoderConfigs.IndexOf(selectedEncoderConfig) + 1; // 1-based
        dialog.FileName = textBoxOutputPath.Text;
        dialog.InitialDirectory = Path.GetDirectoryName(textBoxOutputPath.Text);
        dialog.RestoreDirectory = true;
        dialog.AddExtension = true;

        if (dialog.ShowDialog(this) == DialogResult.OK) {
          selectedEncoderConfig = ThbgmExtractor.EncoderConfigs[dialog.FilterIndex - 1]; // 1-based

          dropDownListEncoderConfig.SelectedItem = selectedEncoderConfig;
          textBoxOutputPath.Text = dialog.FileName;
        }
      }
    }

    private void ChangeOutputFileExtension()
    {
      var selectedEncoderConfig = dropDownListEncoderConfig.SelectedItem;

      textBoxOutputPath.Text = Path.ChangeExtension(textBoxOutputPath.Text,
                                                    selectedEncoderConfig.OutputExtension);
    }
#endregion

#region "configure multiple tracks"
    private bool ConfigureExtractMultipleTracks(IWin32Window owner)
    {
      /*
       * initialize control states
       */
      groupBoxEncoder.Enabled = true;
      groupBoxPlaylist.Enabled = true;

      buttonChangeOutputPath.Click += delegate {
        SelectOutputDirectory();
      };
      buttonChangePlaylistPath.Click += delegate {
        SelectPlaylistPath();
      };
      dropDownListEncoderConfig.SelectedIndexChanged += delegate {
        UpdateEncoderConfigControlState();
        UpdateTagFormatControlState();
      };
      checkBoxEnableTagging.CheckedChanged += delegate {
        UpdateTagFormatControlState();
      };
      dropDownListPlaylistFormat.SelectedIndexChanged += delegate {
        ChangePlaylistPathExtension();
        UpdatePlaylistFormatControlState();
      };

      UpdateEncoderConfigControlState();
      UpdatePlaylistFormatControlState();
      UpdateTagFormatControlState();

      Text = ThbgmExtractor.GetText("ExtractionConfigDialog_ConfigureSingleTrack_Text", config.Product.Title);

      var outputDirectory = config.OutputDirectory;

      if (string.IsNullOrEmpty(outputDirectory)) {
        var dirname = config.Product.GetUnifiedTitle(config.TrackFormatInfo.TitleUnificationOptions);

        outputDirectory = Path.Combine(Environment.GetFolderPath(Environment.SpecialFolder.Desktop),
                                       PathUtils.ReplaceInvalidPathCharsWithBlanks(dirname));
      }

      textBoxOutputPath.Text = outputDirectory;

      if (config.PlaylistFormat == null) {
        dropDownListPlaylistFormat.SelectedItem = NullPlaylistFormat.Instance;
        textBoxPlaylistPath.Text = string.Empty;
      }
      else {
        dropDownListPlaylistFormat.SelectedItem = config.PlaylistFormat;

        var playlistPath = config.PlaylistPath;

        if (string.IsNullOrEmpty(playlistPath))
          playlistPath = GetDefaultPlaylistPath(config.PlaylistFormat);

        textBoxPlaylistPath.Text = playlistPath;
      }

      /*
       * configure
       */
      if (ShowDialog(owner) != DialogResult.OK)
        return false;

      /*
       * set configuration
       */
      config.OutputDirectory = textBoxOutputPath.Text;
      config.EncoderConfig = dropDownListEncoderConfig.SelectedItem;
      config.SaveSourceAfterEncode = checkBoxSaveSourceAfterEncode.Checked;
      config.EnableTagging = checkBoxEnableTagging.Checked;
      config.TagFormat = dropDownListTagFormat.SelectedItem;

      if (NullPlaylistFormat.Instance == dropDownListPlaylistFormat.SelectedItem)
        config.PlaylistFormat = null;
      else
        config.PlaylistFormat = dropDownListPlaylistFormat.SelectedItem;

      config.PlaylistPath = textBoxPlaylistPath.Text;

      /*
       * set tracks and tags
       */
      var extractTracksAndTags = new Dictionary<Track, TrackTag>();

      foreach (var track in config.Tracks) {
        if (!track.Checked)
          continue;

        extractTracksAndTags.Add(track, config.TrackTags[track.TrackInfo]);
      }

      config.SetTracksAndTags(extractTracksAndTags, true);

      return true;
    }

    private void SelectOutputDirectory()
    {
      using (var dialog = new FolderBrowserDialog()) {
        dialog.Description = ThbgmExtractor.GetText("FolderBrowserDialog_SelectOutputDirectory_Description");
        dialog.RootFolder = Environment.SpecialFolder.Desktop;
        dialog.SelectedPath = textBoxOutputPath.Text;
        dialog.ShowNewFolderButton = true;

        if (dialog.ShowDialog(this) == DialogResult.OK) {
          textBoxOutputPath.Text = dialog.SelectedPath;
        }
      }
    }

    private string GetDefaultPlaylistPath(PlaylistFormat playlistFormat)
    {
      var playlistPath = Path.Combine(textBoxOutputPath.Text,
                                      config.Product.GetUnifiedTitle(config.TrackFormatInfo.TitleUnificationOptions));

      playlistPath = Path.ChangeExtension(playlistPath, playlistFormat.DefaultExtension);

      return playlistPath;
    }

    private void SelectPlaylistPath()
    {
      using (var dialog = new SaveFileDialog()) {
        dialog.Filter = FileDialogFilter.CreateFilterString(availablePlaylistFormats.Select(delegate(PlaylistFormat f) {
          return new FileDialogFilter.Filter(f.Name, "*" + f.DefaultExtension);
        }));

        var selectedPlaylistFormat = dropDownListPlaylistFormat.SelectedItem;

        dialog.Title = ThbgmExtractor.GetText("ExtractionConfigDialog_SelectPlaylistPath_DialogTitle");
        dialog.FilterIndex = Array.IndexOf(availablePlaylistFormats, selectedPlaylistFormat) + 1; // 1-based
        dialog.FileName = Path.GetFileNameWithoutExtension(textBoxPlaylistPath.Text);
        dialog.InitialDirectory = Path.GetDirectoryName(textBoxPlaylistPath.Text);
        dialog.RestoreDirectory = true;
        dialog.AddExtension = true;

        if (dialog.ShowDialog(this) == DialogResult.OK) {
          selectedPlaylistFormat = availablePlaylistFormats[dialog.FilterIndex - 1]; // 1-based

          dropDownListPlaylistFormat.SelectedItem = selectedPlaylistFormat;
          textBoxPlaylistPath.Text = dialog.FileName;
        }
      }
    }

    private void ChangePlaylistPathExtension()
    {
      var selectedPlaylistFormat = dropDownListPlaylistFormat.SelectedItem;

      if (selectedPlaylistFormat == NullPlaylistFormat.Instance)
        return;

      if (string.IsNullOrEmpty(textBoxPlaylistPath.Text))
        textBoxPlaylistPath.Text = GetDefaultPlaylistPath(selectedPlaylistFormat);
      else
        textBoxPlaylistPath.Text = Path.ChangeExtension(textBoxPlaylistPath.Text,
                                                        selectedPlaylistFormat.DefaultExtension);
    }

    private void UpdatePlaylistFormatControlState()
    {
      var enabled = (dropDownListPlaylistFormat.SelectedItem != NullPlaylistFormat.Instance);

      buttonChangePlaylistPath.Enabled = enabled;
      textBoxPlaylistPath.Enabled = enabled;
    }
#endregion

    private void UpdateEncoderConfigControlState()
    {
      checkBoxSaveSourceAfterEncode.Enabled = (dropDownListEncoderConfig.SelectedItem != EncoderConfig.Empty && !dropDownListEncoderConfig.SelectedItem.UseStandardInputAsOutput);
    }

    private void UpdateTagFormatControlState()
    {
      dropDownListTagFormat.Enabled = (dropDownListEncoderConfig.SelectedItem == EncoderConfig.Empty) && checkBoxEnableTagging.Checked;
    }

    private readonly PlaylistFormat[] availablePlaylistFormats;
    private readonly ExtractionConfig config;
  }
}
