﻿//  
//  Author:
//       smdn <smdn@smdn.jp>
// 
//  Copyright (c) 2004-2017 smdn
// 
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
// 
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
//  GNU General Public License for more details.
//  
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

using System;
using System.IO;
using System.Collections;
using System.Collections.Generic;
using System.ComponentModel;
using System.Diagnostics;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading;
using System.Windows.Forms;

using Smdn.Collections;
using Smdn.Formats.Playlists;
using Smdn.Formats.Thbgm;
using Smdn.IO;
using Smdn.Windows.Forms.ThbgmPlayer;

namespace Smdn.Applications.ThbgmExtractor {
  public partial class MainDialog : Form {
    private class FadingCuverTypeStripMenuItem : ToolStripMenuItem {
      public FadingCurveType FadingCurveType {
        get; private set;
      }

      public FadingCuverTypeStripMenuItem(FadingCurveType fadingCurveType, EventHandler onClick)
        : base(Texts.GetText(fadingCurveType), null, onClick)
      {
        this.FadingCurveType = fadingCurveType;
      }
    }

    private class TimesToRepeatStripMenuItem : ToolStripMenuItem {
      public int TimesToRepeat {
        get; private set;
      }

      public TimesToRepeatStripMenuItem(int timesToRepeat, EventHandler onClick)
        : base(ThbgmExtractor.GetText("TimesToRepeatStripMenuItem_FormatString", timesToRepeat), null, onClick)
      {
        this.TimesToRepeat = timesToRepeat;
      }
    }

    private class ExtractionConfigFileStripMenuItem : ToolStripMenuItem {
      public string File {
        get; private set;
      }

      public ExtractionConfigFileStripMenuItem(string file, EventHandler onClick)
        : base(Path.GetFileName(file), null, onClick)
      {
        
        this.File = file;
      }
    }

    public string CurrentConfigFilePath {
      get { return currentConfigFilePath; }
      set
      {
        if (currentConfigFilePath == value)
          return;

        currentConfigFilePath = value;

        UpdateText();
      }
    }

    public bool IsEdittingExtractionConfigDirty {
      get { return isEdittingExtractionConfigDirty; }
      set
      {
        isEdittingExtractionConfigDirty = value;

        UpdateText();
      }
    }

    internal Player Player {
      get { return player; }
    }

    public MainDialog()
    {
      InitializeComponent();
      Initialize();

      if (Runtime.IsRunningOnMono) {
        // XXX
        this.ClientSize = new Size(784, 563);
        productSelector.Width = this.ClientSize.Width;
        splitContainerMain.Size = new Size(784, 486);
        splitContainerMain.SplitterDistance = 300;
      }

      UpdateText();
    }

    private void Initialize()
    {
      textBoxLog.MaxLength = int.MaxValue;
      textBoxLog.Font = new Font(FontFamily.GenericMonospace, Control.DefaultFont.SizeInPoints, GraphicsUnit.Point);

      timerLog.Start();

      productSelector.SelectedProductChanged += delegate {
        edittingExtractionConfig.Product = productSelector.SelectedProduct;

        trackListView.Product = edittingExtractionConfig.Product;
      };

      productSelector.BgmSourcePathChanged += delegate {
        edittingExtractionConfig.BgmSourcePath = productSelector.BgmSourcePath;

        player.BgmSourcePath = edittingExtractionConfig.BgmSourcePath;
      };

      toolStripMenuItemExtract.DropDownOpening += delegate {
        toolStripMenuItemTrackList_ExtractTrack.Enabled =
        toolStripMenuItemExtract_ExtractTrack.Enabled = (trackListView.SelectedTrack != null) && !backgroundWorkerExtraction.IsBusy;
      };

      trackListView.ProductChanged += delegate {
        edittingExtractionConfig.Tracks = trackListView.Tracks;
        trackTagListView.SetTrackTags(edittingExtractionConfig.TrackTags.Values);
        toolStripMenuItemTrackList_ExtractTrack.Enabled =
        toolStripMenuItemExtract_ExtractTrack.Enabled = false;
        player.Track = null;
        UpdateText();
      };
      trackListView.SelectedIndexChanged += delegate {
        toolStripMenuItemTrackList_ExtractTrack.Enabled =
        toolStripMenuItemExtract_ExtractTrack.Enabled = (trackListView.SelectedTrack != null) && !backgroundWorkerExtraction.IsBusy;
        player.Track = trackListView.SelectedTrack;
      };
      trackListView.TrackEffectPropertyChanged += delegate {
        UpdateText();
      };

      trackListView.Product = null;
      trackListView.Product = productSelector.SelectedProduct;

      foreach (var fadingCurveType in new[] {
        FadingCurveType.Linear,
        FadingCurveType.Exponential,
        FadingCurveType.Logarithmic,
        FadingCurveType.Trigonometrical,
        FadingCurveType.None,
      }) {
        toolStripMenuItemEdit_FadingCurveType.DropDownItems.Add(new FadingCuverTypeStripMenuItem(fadingCurveType, delegate(object sender, EventArgs e) {
          if (trackListView.SelectedTrack != null)
            trackListView.SelectedTrack.Effect.SetFadingCurve((sender as FadingCuverTypeStripMenuItem).FadingCurveType, true);
        }));
      }

      toolStripMenuItemEdit_FadingCurveType.DropDownOpening += delegate {
        var selectedFadingCurveType = (trackListView.SelectedTrack == null)
          ? (FadingCurveType?)null
          : (FadingCurveType?)trackListView.SelectedTrack.Effect.FadingCurveType;

        foreach (FadingCuverTypeStripMenuItem item in toolStripMenuItemEdit_FadingCurveType.DropDownItems) {
          item.Checked = (item.FadingCurveType == selectedFadingCurveType);
        }
      };

      for (var i = 0; i < 10; i++) {
        toolStripMenuItemEdit_TimesToRepeat.DropDownItems.Add(new TimesToRepeatStripMenuItem(i, delegate(object sender, EventArgs e) {
          if (trackListView.SelectedTrack == null)
            return;

          var track = trackListView.SelectedTrack;

          track.Effect.SetTimesToRepeat(track.TrackInfo, (sender as TimesToRepeatStripMenuItem).TimesToRepeat);
        }));
      }

      toolStripMenuItemEdit_TimesToRepeat.DropDownOpening += delegate {
        var selectedTimesToRepeat = (trackListView.SelectedTrack == null)
          ? -1
          : trackListView.SelectedTrack.TrackInfo.CalculateRepeatedTimes(trackListView.SelectedTrack.Effect.TotalLength);

        foreach (TimesToRepeatStripMenuItem item in toolStripMenuItemEdit_TimesToRepeat.DropDownItems) {
          item.Checked = (item.TimesToRepeat == selectedTimesToRepeat);
        }
      };

      toolStripMenuItemEdit_CheckAll.Click += delegate {
        trackListView.CheckAll();
      };

      toolStripMenuItemEdit_UncheckAll.Click += delegate {
        trackListView.UncheckAll();
      };

      editToolStripItems = toolStripMenuItemEdit.DropDownItems.Cast<ToolStripItem>().ToArray();
      trackListToolStripItems = contextMenuStripTrackList.Items.Cast<ToolStripItem>().ToArray();

      contextMenuStripTrackList.Items.Add(contextMenuDummyItem);

      contextMenuStripTrackList.Opening += delegate {
        var enabled = (trackListView.SelectedTrack != null);

        toolStripMenuItemEdit_FadingCurveType.Enabled = enabled;
        toolStripMenuItemEdit_TimesToRepeat.Enabled = enabled;

        toolStripMenuItemEdit.DropDownItems.Clear();
        contextMenuStripTrackList.Items.Clear();
        contextMenuStripTrackList.Items.AddRange(trackListToolStripItems);
        contextMenuStripTrackList.Items.AddRange(editToolStripItems);
      };

      contextMenuStripTrackList.Closed += delegate {
        contextMenuStripTrackList.Items.Clear();
        contextMenuStripTrackList.Items.Add(contextMenuDummyItem);
        toolStripMenuItemEdit.DropDownItems.Clear();
        toolStripMenuItemEdit.DropDownItems.AddRange(editToolStripItems);
      };

      toolStripMenuItemEdit.DropDownOpening += delegate {
        var enabled = (trackListView.SelectedTrack != null);

        toolStripMenuItemEdit_FadingCurveType.Enabled = enabled;
        toolStripMenuItemEdit_TimesToRepeat.Enabled = enabled;

        contextMenuStripTrackList.Items.Clear();
        contextMenuStripTrackList.Items.Add(contextMenuDummyItem);
        toolStripMenuItemEdit.DropDownItems.Clear();
        toolStripMenuItemEdit.DropDownItems.AddRange(editToolStripItems);
      };

      toolStripMenuItemEdit.DropDownClosed += delegate {
        toolStripMenuItemEdit.DropDownItems.Clear();
        contextMenuStripTrackList.Items.Clear();
        contextMenuStripTrackList.Items.AddRange(trackListToolStripItems);
        contextMenuStripTrackList.Items.AddRange(editToolStripItems);
      };

      // drop down dummy item
      toolStripMenuItemFile_Export_TitleFiles.DropDownItems.Add(new ToolStripSeparator());
      toolStripMenuItemFile_RecentFiles.DropDownItems.Add(new ToolStripSeparator());

      player.PlayDone += player_PlayDone;
      player.PlayerStatusChanged += player_PlayerStatusChanged;
      player.PrevButtonClick += player_PrevButtonClick;
      player.NextButtonClick += player_NextButtonClick;

      SetControlStateToNotExtracting();

      InitializeProducts();

      initialized = true;
    }

    private void InitializeProducts()
    {
      var products = ThbgmExtractor.Products.ToArray();

      productSelector.SetProducts(products);

      var initialProduct = products[0];
      var initialExtractionConfig = new ExtractionConfig();

      initialExtractionConfig.Product = initialProduct;
      initialExtractionConfig.BgmSourcePath = initialProduct.InstalledOrDefaultBgmSourcePath;

      SetExtractionConfig(initialExtractionConfig, false);
    }

    protected override void OnLoad(EventArgs e)
    {
      base.OnLoad(e);

      Config.MainDialogState.Restore(this);
    }

    public void ExtractOnShown(ExtractionConfig extractionConfig)
    {
      SetExtractionConfig(extractionConfig, true);

      doExtractOnShown = true;
    }

    protected override void OnShown(EventArgs e)
    {
      base.OnShown(e);

      if (doExtractOnShown) {
        IsEdittingExtractionConfigDirty = true;
        ExtractWithEdittingConfig();
      }
      else {
        ClearEdittingExtractionConfigDirtiness();
      }
    }

    private void toolStripMenuItemFile_Quit_Click(object sender, EventArgs e)
    {
      Close();
    }

    protected override void OnFormClosing(FormClosingEventArgs e)
    {
      Config.MainDialogState.Store(this);

      try {
        if (e.CloseReason != CloseReason.UserClosing)
          // no confirmation
          return;

        if (backgroundWorkerExtraction.IsBusy) {
          var result = MessageBox.Show(this, ThbgmExtractor.GetText("MessageBox_OnClosingAbortExtraction"), Application.ProductName, MessageBoxButtons.YesNo, MessageBoxIcon.Exclamation, MessageBoxDefaultButton.Button2);

          if (result != DialogResult.Yes) {
            // continue extraction
            e.Cancel = true;
            return;
          }

          // abort
          backgroundWorkerExtraction.CancelAsync();

          // wait for RunWorkerCompleted
          while (backgroundWorkerExtraction.IsBusy) {
            Application.DoEvents();
            Thread.Sleep(250);
          }
        }

        e.Cancel = !CheckAndConfirmSaveExtractionConfig(true, false);
      }
      finally {
        base.OnFormClosing(e);
      }
    }

    private void UpdateText()
    {
      UpdateText(null);
    }

    private void UpdateText(string extractionProgress)
    {
      var dirty = (IsEdittingExtractionConfigDirty || edittingExtractionConfig.AreAnyTracksDirty)
        ? "*"
        : string.Empty;

      string context = null;

      if (string.IsNullOrEmpty(extractionProgress)) {
        var playingTrack = (player != null && player.IsPlaying)
          ? string.Format("{0}{1}.{2} - ",
                          player.IsPausing ? string.Empty : "♫ ",
                          player.Track.TrackInfo.TrackNumber,
                          player.Track.TrackInfo.Title)
          : string.Empty;

        context = playingTrack + (CurrentConfigFilePath ?? productSelector.SelectedProduct.Title);
      }
      else {
        context = extractionProgress;
      }

      Text = string.Format("{0} - {1}{2}", context, ThbgmExtractor.GetText("ThbgmExtractor"), dirty);
    }

    private void toolStripMenuItemEdit_EditEffectsCollective_Click(object sender, EventArgs e)
    {
      Config.CollectiveEffectEditDefaults.OpenConfigAppliedDialog(delegate(CollectiveEffectEditDialog dialog) {
        if (dialog.ShowDialog(this) == DialogResult.OK) {
          // cancel editing before apply effects
          trackListView.CancelEdit();
        }
        else {
          trackListView.EnterEdit();
          return;
        }

        foreach (var track in trackListView.Tracks) {
          track.Effect.SetFadingCurve(dialog.FadingCurveType, dialog.FadingCurveCoefficient, true);

          switch (dialog.EffectLengthBasis) {
            case EffectLengthBasis.FadingStart:
              track.Effect.SetLength(dialog.FadingStart, dialog.FadingLength, dialog.TrailingSilenceLength);
              break;
            case EffectLengthBasis.TotalLength:
              track.Effect.SetTotalLength(dialog.TotalLength, dialog.FadingLength, dialog.TrailingSilenceLength);
              break;
            case EffectLengthBasis.TimesToRepeat:
              track.Effect.SetLength(track.TrackInfo, dialog.TimesToRepeat, dialog.FadingLength, dialog.FadingOffset, dialog.TrailingSilenceLength);
              break;
          }
        }

        tabControlTracks.SelectedTab = tabPageEffects;
      });
    }

    private void toolStripMenuItemEdit_EditTagsCollective_Click(object sender, EventArgs e)
    {
      Config.TrackFormatInfoDefaults.OpenConfigAppliedCollectiveTrackTagEditDialog(delegate(CollectiveTrackTagEditDialog dialog) {
        dialog.TrackFormatInfo = edittingExtractionConfig.TrackFormatInfo.Clone();
        dialog.OutputFileName = edittingExtractionConfig.DefaultOutputFileName;

        if (trackListView.SelectedTrack == null) {
          if (edittingExtractionConfig.Product == null)
            dialog.TrackForDefaults = null;
          else
            dialog.TrackForDefaults = edittingExtractionConfig.Product.Tracks[0];
        }
        else {
          dialog.TrackForDefaults = trackListView.SelectedTrack.TrackInfo;
        }

        if (dialog.ShowDialog(this) != DialogResult.OK)
          return;

        edittingExtractionConfig.TrackFormatInfo = dialog.TrackFormatInfo.Clone();
        edittingExtractionConfig.DefaultOutputFileName = dialog.OutputFileName;

        foreach (var trackTag in edittingExtractionConfig.TrackTags.Values) {
          trackTag.OutputFileName = dialog.OutputFileName;
          trackTag.Album          = dialog.Album;
          trackTag.Performers     = dialog.Performers;
          trackTag.AlbumArtists   = dialog.AlbumArtists;
          trackTag.Composers      = dialog.Composers;
          trackTag.Genres         = dialog.Genres;
          trackTag.Comment        = dialog.Comment;
        }

        tabControlTracks.SelectedTab = tabPageTags;

        IsEdittingExtractionConfigDirty = true;
      });
    }

    private void toolStripMenuItemFile_RecentFiles_Clear_Click(object sender, EventArgs e)
    {
      Config.RecentFiles.ClearExtractionConfigFiles();
    }


    private void toolStripMenuItemFile_ReloadTitleFiles_Click(object sender, EventArgs e)
    {
      // check and save current config
      if (!CheckAndConfirmSaveExtractionConfig(true, true))
        return; // cancel

      ClearEdittingExtractionConfigDirtiness(); // not to confirm again

      ThbgmExtractor.ReloadProductInfoes();

      var text = new StringBuilder();

      foreach (var p in ThbgmExtractor.Products) {
        if (!p.IsEmbedded)
          text.AppendLine(p.Title);
      }

      if (text.Length == 0)
        text.Append(ThbgmExtractor.GetText("MessageBox_ReloadTitleFiles_NoFilesAvailable"));
      else
        text.Insert(0, ThbgmExtractor.GetText("MessageBox_ReloadTitleFiles_Available") + Environment.NewLine + Environment.NewLine);

      MessageBox.Show(this,
                      text.ToString(),
                      ThbgmExtractor.GetText("MessageBox_ReloadTitleFiles_Caption"),
                      MessageBoxButtons.OK,
                      MessageBoxIcon.Information);

      InitializeProducts();
    }

    private void toolStripMenuItemFile_ReloadEncoderConfigFile_Click(object sender, EventArgs e)
    {
      ThbgmExtractor.ReloadEncoderConfig();

      var text = new StringBuilder();

      foreach (var encoder in ThbgmExtractor.EncoderConfigs) {
        if (encoder != EncoderConfig.Empty)
          text.AppendLine(encoder.Name);
      }

      if (text.Length == 0)
        text.Append(ThbgmExtractor.GetText("MessageBox_ReloadEncoderConfigFile_NoEncodersAvailable"));
      else
        text.Insert(0, ThbgmExtractor.GetText("MessageBox_ReloadEncoderConfigFile_Available") + Environment.NewLine + Environment.NewLine);

      MessageBox.Show(this,
                      text.ToString(),
                      ThbgmExtractor.GetText("MessageBox_ReloadEncoderConfigFile_Caption", EncoderConfig.ConfigFileName),
                      MessageBoxButtons.OK,
                      MessageBoxIcon.Information);
    }

    private void toolStripMenuItemHelp_About_Click(object sender, EventArgs e)
    {
      using (var dialog = new AboutDialog()) {
        dialog.ShowDialog(this);
      }
    }
#region "open/save"
    private void toolStripMenuItemFile_Open_BmgSource_Click(object sender, EventArgs e)
    {
      productSelector.ShowChangeBmgSourcePathDialog();
    }

    private void toolStripMenuItemFile_Open_ExtractionConfig_Click(object sender, EventArgs e)
    {
      using (var dialog = new OpenFileDialog()) {
        dialog.Filter = FileDialogFilter.CreateFilterString(new[] {
          new[] {ThbgmExtractor.GetText("ExtractionConfigFile"), "*" + ExtractionConfig.FileExtension},
          new[] {Texts.GetText("AllFiles"), "*.*"},
        });

        dialog.FilterIndex = 0;
        dialog.FileName = CurrentConfigFilePath;
        dialog.InitialDirectory = string.IsNullOrEmpty(CurrentConfigFilePath) ? Environment.CurrentDirectory : Path.GetDirectoryName(CurrentConfigFilePath);
        dialog.RestoreDirectory = true;

        if (dialog.ShowDialog(this) == DialogResult.OK)
          LoadExtractionConfig(dialog.FileName, true);
      }
    }

    private void toolStripMenuItemFile_SaveExtractionConfigAs_Click(object sender, EventArgs e)
    {
      SaveExtractionConfigAs();
    }

    private void toolStripMenuItemFile_SaveExtractionConfig_Click(object sender, EventArgs e)
    {
      SaveExtractionConfig();
    }

    private void toolStripMenuItemFile_RecentFiles_DropDownOpening(object sender, EventArgs e)
    {
      toolStripMenuItemFile_RecentFiles.DropDownItems.Clear();

      var entryExists = false;

      foreach (var file in Config.RecentFiles.GetExtractionConfigFiles()) {
        entryExists = true;

        toolStripMenuItemFile_RecentFiles.DropDownItems.Add(new ExtractionConfigFileStripMenuItem(file, delegate(object se, EventArgs eargs) {
          LoadExtractionConfig((se as ExtractionConfigFileStripMenuItem).File, true);
        }));
      }

      if (entryExists) {
        toolStripMenuItemFile_RecentFiles.DropDownItems.Add(toolStripMenuItemFile_RecentFiles_Separator);
        toolStripMenuItemFile_RecentFiles_Clear.Enabled = true;
      }
      else {
        toolStripMenuItemFile_RecentFiles_Clear.Enabled = false;
      }

      toolStripMenuItemFile_RecentFiles.DropDownItems.Add(toolStripMenuItemFile_RecentFiles_Clear);
    }
#endregion

#region "save/load extraction config"
    private bool CheckAndConfirmSaveExtractionConfig(bool checkAllConfig, bool checkExtracting)
    {
      if (!initialized)
        return true;

      if (checkExtracting && backgroundWorkerExtraction.IsBusy) {
        MessageBox.Show(this, ThbgmExtractor.GetText("MessageBox_ExtractionProceeding"), Application.ProductName, MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
        return false; // cancel
      }

      if (!(edittingExtractionConfig.AreAnyTracksDirty || (checkAllConfig && IsEdittingExtractionConfigDirty)))
        return true;

      var result = MessageBox.Show(this, ThbgmExtractor.GetText("MessageBox_CheckAndConfirmSaveExtractionConfig"), Application.ProductName, MessageBoxButtons.YesNoCancel, MessageBoxIcon.Question, MessageBoxDefaultButton.Button1);

      switch (result) {
        case DialogResult.Yes:
          return SaveExtractionConfig() == DialogResult.OK;

        case DialogResult.No:
          return true;

        default:
          return false; // cancel
      }
    }

    public void LoadExtractionConfig(string path, bool persistent)
    {
      try {
        if (!CheckAndConfirmSaveExtractionConfig(true, true))
          return;

        SetExtractionConfig(ExtractionConfig.LoadFrom(path), true);

        if (persistent)
          CurrentConfigFilePath = path;

        IsEdittingExtractionConfigDirty = false;

        if (persistent)
          Config.RecentFiles.AddExtractionConfigFile(path);
      }
      catch (InvalidDataException ex) {
        MessageBox.Show(this, ThbgmExtractor.GetText("MessageBox_LoadExtractionConfig_InvalidDataException", ex.InnerException.Message), Application.ProductName, MessageBoxButtons.OK, MessageBoxIcon.Error);
        Console.Error.WriteLine(ex);
      }
      catch (NotSupportedException) {
        MessageBox.Show(this, ThbgmExtractor.GetText("MessageBox_LoadExtractionConfig_NotSupportedException"), Application.ProductName, MessageBoxButtons.OK, MessageBoxIcon.Error);
      }
    }

    private void SetExtractionConfig(ExtractionConfig newExtractionConfig, bool setTracks)
    {
      var propertyChanging = new PropertyChangingEventHandler(delegate(object sender, PropertyChangingEventArgs e) {
        var config = sender as ExtractionConfig;

        if (e.PropertyName == "BgmSourcePath") {
          lastBgmSourcePath = config.BgmSourcePath;
        }
        else if (e.PropertyName == "Product") {
          if (string.IsNullOrEmpty(productSelector.BgmSourcePath) || productSelector.BgmSourcePath == config.Product.InstalledOrDefaultBgmSourcePath)
            productSelector.BgmSourcePath = string.Empty;

          config.BgmSourcePath = lastBgmSourcePath;

          if (CheckAndConfirmSaveExtractionConfig(false, true)) {
            config.BgmSourcePath = productSelector.BgmSourcePath;

            // force save-as in next time
            CurrentConfigFilePath = null;

            return; // extraction config is not dirty or saved
          }
          else /* canceled */ {
            // revert bgm source and product
            productSelector.SelectedProduct = config.Product;
            productSelector.BgmSourcePath = config.BgmSourcePath;
            player.BgmSourcePath = config.BgmSourcePath;

            (e as CancellablePropertyChangingEventArgs).Cancel = true;
          }
        }
      });

      var propertyChanged = new PropertyChangedEventHandler(delegate(object sender, PropertyChangedEventArgs e) {
        var config = sender as ExtractionConfig;

        if (e.PropertyName == "Product" && string.IsNullOrEmpty(productSelector.BgmSourcePath))
          productSelector.BgmSourcePath = config.Product.InstalledOrDefaultBgmSourcePath;

        UpdateText();
      });

      edittingExtractionConfig.PropertyChanging -= propertyChanging;
      edittingExtractionConfig.PropertyChanged -= propertyChanged;

      productSelector.SelectedProduct = newExtractionConfig.Product;
      productSelector.BgmSourcePath = newExtractionConfig.BgmSourcePath;

      if (setTracks) {
        trackListView.SetProductAndTracks(newExtractionConfig.Product, newExtractionConfig.Tracks);
      }
      else {
        trackListView.Product = newExtractionConfig.Product;

        newExtractionConfig.Tracks = trackListView.Tracks;
      }

      trackTagListView.SetTrackTags(newExtractionConfig.TrackTags.Values);

      edittingExtractionConfig = newExtractionConfig;

      edittingExtractionConfig.PropertyChanging += propertyChanging;
      edittingExtractionConfig.PropertyChanged += propertyChanged;
    }

    private void ClearEdittingExtractionConfigDirtiness()
    {
      edittingExtractionConfig.ClearTracks();
      IsEdittingExtractionConfigDirty = false;
    }

    private DialogResult SaveExtractionConfig()
    {
      if (string.IsNullOrEmpty(CurrentConfigFilePath)) {
        return SaveExtractionConfigAs();
      }
      else {
        edittingExtractionConfig.Save(CurrentConfigFilePath);

        ClearEdittingExtractionConfigDirtiness();

        return DialogResult.OK;
      }
    }

    private DialogResult SaveExtractionConfigAs()
    {
      using (var dialog = new SaveFileDialog()) {
        dialog.Filter = FileDialogFilter.CreateFilterString(new[] {
          new[] {ThbgmExtractor.GetText("ExtractionConfigFile"), "*" + ExtractionConfig.FileExtension},
          new[] {Texts.GetText("AllFiles"), "*.*"},
        });

        var initialFilePath = string.IsNullOrEmpty(CurrentConfigFilePath)
          ? Path.Combine(Environment.CurrentDirectory, Smdn.IO.PathUtils.RemoveInvalidFileNameChars(productSelector.SelectedProduct.Title + ExtractionConfig.FileExtension))
          : CurrentConfigFilePath;

        dialog.FilterIndex = 0;
        dialog.FileName = initialFilePath;
        dialog.InitialDirectory = Path.GetDirectoryName(initialFilePath);
        dialog.RestoreDirectory = true;

        var result = dialog.ShowDialog(this);

        if (result == DialogResult.OK) {
          CurrentConfigFilePath = dialog.FileName;
          edittingExtractionConfig.Save(dialog.FileName);

          Config.RecentFiles.AddExtractionConfigFile(dialog.FileName);

          ClearEdittingExtractionConfigDirtiness();
        }

        return result;
      }
    }
#endregion

#region "import/export"
    private class EmbeddedProductInfoToolStripMenuItem : ToolStripMenuItem {
      public ProductInfo Product {
        get; private set;
      }

      public EmbeddedProductInfoToolStripMenuItem(ProductInfo product, EventHandler onClick)
        : base(product.Title, null, onClick)
      {
        this.Product = product;
      }
    }

    private void toolStripMenuItemFile_ExportTitleFile_DropDownOpening(object sender, EventArgs e)
    {
      toolStripMenuItemFile_Export_TitleFiles.DropDownItems.Clear();

      foreach (var product in ThbgmExtractor.Products) {
        if (product.IsEmbedded)
          toolStripMenuItemFile_Export_TitleFiles.DropDownItems.Add(new EmbeddedProductInfoToolStripMenuItem(product, toolStripMenuItemFile_Export_TitleFile_EmbeddedProductInfo_Click));
      }
    }

    private void toolStripMenuItemFile_Export_TitleFile_EmbeddedProductInfo_Click(object sender, EventArgs e)
    {
      var menuItem = sender as EmbeddedProductInfoToolStripMenuItem;

      if (menuItem == null)
        return;

      ExportTitleFile(menuItem.Product);
    }

    private void toolStripMenuItemFile_Export_THxxBGMFadeoutSettings_Click(object sender, EventArgs e)
    {
      ExportTHxxBGMFadeoutSettings();
    }
    
    private void toolStripMenuItemFile_Import_ThplayApplicationConfig_Click(object sender, EventArgs e)
    {
      ImportThplayApplicationConfig();
    }

    private void toolStripMenuItemFile_Import_THxxBGMFadeoutSettings_Click(object sender, EventArgs e)
    {
      ImportTHxxBGMFadeoutSettings();
    }

    private void toolStripMenuItemFile_Import_THxxBGMSettings_Click(object sender, EventArgs e)
    {
      ImportTHxxBGMSettings();
    }

    private void ExportTitleFile(ProductInfo product)
    {
      using (var dialog = new SaveFileDialog()) {
        dialog.Filter = FileDialogFilter.CreateFilterString(new[] {
          new[] {Texts.GetText("ProductInfoFile"), "*.txt"},
          new[] {Texts.GetText("AllFiles"), "*.*"},
        });

        dialog.FilterIndex = 0;
        dialog.FileName = Path.Combine(Environment.CurrentDirectory, string.Format("titles_{0}.txt", product.Prefix));
        dialog.InitialDirectory = Environment.CurrentDirectory;
        dialog.RestoreDirectory = true;

        if (dialog.ShowDialog(this) != DialogResult.OK)
          return;

        product.Save(dialog.FileName);
      }
    }

    private void ExportTHxxBGMFadeoutSettings()
    {
      if (Guid.Empty.Equals(edittingExtractionConfig.Product.THxxBGMProductGuid)) {
        MessageBox.Show(this,
                        ThbgmExtractor.GetText("MessageBox_ExportTHxxBGMFadeoutSettings_THxxBGMUnsupportedProduct", edittingExtractionConfig.Product.Title),
                        Application.ProductName,
                        MessageBoxButtons.OK,
                        MessageBoxIcon.Error);
        return;
      }

      using (var dialog = new SaveFileDialog()) {
        dialog.Filter = FileDialogFilter.CreateFilterString(new[] {
          new[] {ThbgmExtractor.GetText("THxxBGMFadeoutSettingsFile"), "*" + ExtractionConfig.THxxBGMFadeoutSettingsFileExtension},
          new[] {Texts.GetText("AllFiles"), "*.*"},
        });

        dialog.FilterIndex = 0;
        dialog.FileName = Path.Combine(Environment.CurrentDirectory, string.Format("{0}{1}", edittingExtractionConfig.Product.THxxBGMProductGuid.ToString("D").ToUpperInvariant(), ExtractionConfig.THxxBGMFadeoutSettingsFileExtension));
        dialog.InitialDirectory = Environment.CurrentDirectory;
        dialog.RestoreDirectory = true;

        if (dialog.ShowDialog(this) != DialogResult.OK)
          return;

        try {
          edittingExtractionConfig.SaveAsTHxxBGMFadeoutSettings(dialog.FileName);

          MessageBox.Show(this,
                          ThbgmExtractor.GetText("MessageBox_ExportTHxxBGMFadeoutSettings_ExportDone", dialog.FileName),
                          Application.ProductName,
                          MessageBoxButtons.OK,
                          MessageBoxIcon.Information);
        }
        catch (Exception ex) {
          Console.Error.WriteLine(ex);

          MessageBox.Show(this,
                          ThbgmExtractor.GetText("MessageBox_ExportTHxxBGMFadeoutSettings_ExportFailed"),
                          Application.ProductName,
                          MessageBoxButtons.OK,
                          MessageBoxIcon.Error);
        }
      }
    }

    private void ImportThplayApplicationConfig()
    {
      using (var dialog = new OpenFileDialog()) {
        dialog.Filter = FileDialogFilter.CreateFilterString(new[] {
          new[] {ThbgmExtractor.GetText("ThplayApplicationConfigFile"), Thplay.ApplicationConfigFileName},
          new[] {Texts.GetText("AllFiles"), "*.*"},
        });

        dialog.FilterIndex = 0;
        dialog.FileName = Thplay.ApplicationConfigFileName;
        dialog.InitialDirectory = Environment.GetFolderPath(Environment.SpecialFolder.Desktop);
        dialog.RestoreDirectory = true;

        if (dialog.ShowDialog(this) != DialogResult.OK)
          return;

        try {
          Thplay.ImportApplicationConfig(dialog.FileName);

          MessageBox.Show(this,
                          ThbgmExtractor.GetText("MessageBox_ImportThplayApplicationConfig_ImportDone", dialog.FileName),
                          Application.ProductName,
                          MessageBoxButtons.OK,
                          MessageBoxIcon.Information);
        }
        catch (Exception ex) {
          Console.Error.WriteLine(ex);

          MessageBox.Show(this,
                          ThbgmExtractor.GetText("MessageBox_ImportThplayApplicationConfig_ImportFailed"),
                          Application.ProductName,
                          MessageBoxButtons.OK,
                          MessageBoxIcon.Error);
        }
      }
    }

    private void ImportTHxxBGMFadeoutSettings()
    {
      using (var dialog = new OpenFileDialog()) {
        dialog.Filter = FileDialogFilter.CreateFilterString(new[] {
          new[] {ThbgmExtractor.GetText("THxxBGMFadeoutSettingsFile"), "*" + ExtractionConfig.THxxBGMFadeoutSettingsFileExtension},
          new[] {Texts.GetText("AllFiles"), "*.*"},
        });

        dialog.FilterIndex = 0;
        dialog.InitialDirectory = Environment.GetFolderPath(Environment.SpecialFolder.Desktop);
        dialog.RestoreDirectory = true;

        if (dialog.ShowDialog(this) != DialogResult.OK)
          return;

        try {
          LoadExtractionConfig(dialog.FileName, false);

          MessageBox.Show(this,
                          ThbgmExtractor.GetText("MessageBox_ImportTHxxBGMFadeoutSettings_ImportDone", dialog.FileName),
                          Application.ProductName,
                          MessageBoxButtons.OK,
                          MessageBoxIcon.Information);
        }
        catch (Exception ex) {
          Console.Error.WriteLine(ex);

          MessageBox.Show(this,
                          ThbgmExtractor.GetText("MessageBox_ImportTHxxBGMFadeoutSettings_ImportFailed"),
                          Application.ProductName,
                          MessageBoxButtons.OK,
                          MessageBoxIcon.Error);
        }
      }
    }

    private void ImportTHxxBGMSettings()
    {
      using (var dialog = new OpenFileDialog()) {
        dialog.Filter = FileDialogFilter.CreateFilterString(new[] {
          new[] {ThbgmExtractor.GetText("THxxBGMSettingsFile"), THxxBGM.THxxBGMSettingsFileName},
          new[] {Texts.GetText("AllFiles"), "*.*"},
        });

        dialog.FilterIndex = 0;
        dialog.FileName = THxxBGM.THxxBGMSettingsFileName;
        dialog.InitialDirectory = Environment.GetFolderPath(Environment.SpecialFolder.Desktop);
        dialog.RestoreDirectory = true;

        if (dialog.ShowDialog(this) != DialogResult.OK)
          return;

        try {
          THxxBGM.ImportTHxxBGMSettings(dialog.FileName);

          MessageBox.Show(this,
                          ThbgmExtractor.GetText("MessageBox_ImportTHxxBGMSettings_ImportDone", dialog.FileName),
                          Application.ProductName,
                          MessageBoxButtons.OK,
                          MessageBoxIcon.Information);
        }
        catch (Exception ex) {
          Console.Error.WriteLine(ex);

          MessageBox.Show(this,
                          ThbgmExtractor.GetText("MessageBox_ImportTHxxBGMSettings_ImportFailed"),
                          Application.ProductName,
                          MessageBoxButtons.OK,
                          MessageBoxIcon.Error);
        }
      }
    }
#endregion

#region "drag'n drop"
    protected override void OnDragEnter(DragEventArgs drgevent)
    {
      base.OnDragEnter(drgevent);

      if (drgevent.Data.GetDataPresent(DataFormats.FileDrop))
        drgevent.Effect = DragDropEffects.All;
      else
        drgevent.Effect = DragDropEffects.None;
    }

    protected override void OnDragDrop(DragEventArgs drgevent)
    {
      try {
        var files = (string[])drgevent.Data.GetData(DataFormats.FileDrop, false);

        if (0 < files.Length) {
          if (Smdn.IO.PathUtils.AreExtensionEqual(files[0], ExtractionConfig.FileExtension))
            LoadExtractionConfig(files[0], true);
          else
            productSelector.SetBgmSourcePath(files[0], true);
        }
      }
      catch {
        // ignore exceptions
      }
      finally {
        base.OnDragDrop(drgevent);
      }
    }
#endregion

#region "begin extraction"
    private void toolStripMenuItemExtract_ExtractTrack_Click(object sender, EventArgs e)
    {
      ExtractSelectedTrack();
    }

    private void toolStripMenuItemTrackList_ExtractTrack_Click(object sender, EventArgs e)
    {
      ExtractSelectedTrack();
    }

    private void ExtractSelectedTrack()
    {
      var track = trackListView.SelectedTrack;

      if (track == null)
        return;

      var extractTrackConfig = edittingExtractionConfig.Clone();

      extractTrackConfig.Tracks = Array.AsReadOnly(new Track[] {track});

      ConfigureAndExtract(extractTrackConfig, true);
    }

    private void toolStripMenuItemExtract_Extract_Click(object sender, EventArgs e)
    {
      ExtractWithEdittingConfig();
    }

    private void ExtractWithEdittingConfig()
    {
      ConfigureAndExtract(edittingExtractionConfig, false);
    }

    private void ConfigureAndExtract(ExtractionConfig baseConfig, bool extractSingleTrack)
    {
      if (backgroundWorkerExtraction.IsBusy) {
        MessageBox.Show(this, ThbgmExtractor.GetText("MessageBox_ExtractionProceeding"), Application.ProductName, MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
        return;
      }

      /*
       * check bmg source existence
       */
      if (!File.Exists(baseConfig.BgmSourcePath)) {
        var message = string.IsNullOrEmpty(baseConfig.BgmSourcePath)
          ? ThbgmExtractor.GetText("MessageBox_CheckAndAlertBmgSourceExistence_BgmSourceNotSet")
          : ThbgmExtractor.GetText("MessageBox_CheckAndAlertBmgSourceExistence_BgmSourceNotFound", baseConfig.BgmSourcePath);

        MessageBox.Show(this, message, Application.ProductName, MessageBoxButtons.OK, MessageBoxIcon.Exclamation);

        productSelector.Focus();
        productSelector.FocusBgmSourcePath();

        return;
      }

      using (var dialog = new ExtractionConfigDialog(baseConfig)) {
        /*
         * configure
         */
        dialog.Config.PropertyChanged += delegate(object sender, PropertyChangedEventArgs e) {
          var edittedConfig = sender as ExtractionConfig;

          if (e.PropertyName == "OutputDirectory") {
            edittingExtractionConfig.OutputDirectory = edittedConfig.OutputDirectory;
            IsEdittingExtractionConfigDirty = true;
          }
          else if (e.PropertyName == "EncoderConfig") {
            edittingExtractionConfig.EncoderConfig = edittedConfig.EncoderConfig;
            IsEdittingExtractionConfigDirty = true;
          }
          else if (e.PropertyName == "SaveSourceAfterEncode") {
            edittingExtractionConfig.SaveSourceAfterEncode = edittedConfig.SaveSourceAfterEncode;
            IsEdittingExtractionConfigDirty = true;
          }
          else if (e.PropertyName == "EnableTagging") {
            edittingExtractionConfig.EnableTagging = edittedConfig.EnableTagging;
            IsEdittingExtractionConfigDirty = true;
          }
          else if (e.PropertyName == "TagFormat") {
            edittingExtractionConfig.TagFormat = edittedConfig.TagFormat;
            IsEdittingExtractionConfigDirty = true;
          }
          else if (e.PropertyName == "PlaylistFormat") {
            edittingExtractionConfig.PlaylistFormat = edittedConfig.PlaylistFormat;
            IsEdittingExtractionConfigDirty = true;
          }
          else if (e.PropertyName == "PlaylistPath") {
            edittingExtractionConfig.PlaylistPath = edittedConfig.PlaylistPath;
            IsEdittingExtractionConfigDirty = true;
          }
        };

        if (!dialog.ConfigureExtraction(this, extractSingleTrack))
          return; // canceled

        var extractionConfig = dialog.Config;

        Config.ExtractionConfigDefaults.Store(extractionConfig);

        /*
         * check overwriting
         */
        if (!extractSingleTrack &&
            !CheckAndConfirmOverwriteFiles(ExtractBackgroundWorker.GetOutputFilePaths(extractionConfig)))
          return; // canceled

        /*
         * create output directory
         */
        Directory.CreateDirectory(extractionConfig.OutputDirectory);

        if (extractionConfig.PlaylistFormat != null)
          Directory.CreateDirectory(Path.GetDirectoryName(extractionConfig.PlaylistPath));

        /*
         * run extraction worker
         */
        trackListView.ShowExtractProgressColumn();

        backgroundWorkerExtraction.RunWorkerAsync(extractionConfig);
      }
    }

    private bool CheckAndConfirmOverwriteFiles(IEnumerable<string> files)
    {
      var fileExists = false;

      foreach (var file in files) {
        if (File.Exists(file)) {
          fileExists = true;
          break;
        }
      }

      if (!fileExists)
        return true;

      var result = MessageBox.Show(this,
                                   ThbgmExtractor.GetText("MessageBox_CheckAndConfirmOverwriteFiles"),
                                   Application.ProductName,
                                   MessageBoxButtons.YesNo,
                                   MessageBoxIcon.Question,
                                   MessageBoxDefaultButton.Button2);

      if (result == DialogResult.Yes)
        return true; // overwrite if exists
      else
        return false;
    }

    private void toolStripMenuItemTrackList_AbortExtraction_Click(object sender, EventArgs e)
    {
      AbortExtraction();
    }

    private void toolStripMenuItemExtract_Abort_Click(object sender, EventArgs e)
    {
      AbortExtraction();
    }

    private void AbortExtraction()
    {
      // abort
      if (!backgroundWorkerExtraction.IsBusy)
        return;

      backgroundWorkerExtraction.CancelAsync();
    }
#endregion

#region "backgroundWorkerExtraction"
    private void SetControlStateToNotExtracting()
    {
      if (InvokeRequired) {
        Invoke((MethodInvoker)SetControlStateToNotExtracting);
        return;
      }

      SetExtractionControlState(false);
    }

    private void SetControlStateToExtracting()
    {
      if (InvokeRequired) {
        Invoke((MethodInvoker)SetControlStateToExtracting);
        return;
      }

      SetExtractionControlState(true);
    }

    private void SetExtractionControlState(bool extracting)
    {
      productSelector.Enabled = !extracting;

      toolStripMenuItemExtract_Extract.Enabled =
      toolStripMenuItemExtract_ExtractTrack.Enabled =
      toolStripMenuItemTrackList_ExtractTrack.Enabled = !extracting;

      toolStripMenuItemExtract_Abort.Visible =
      toolStripMenuItemTrackList_AbortExtraction.Visible = extracting;
    }

    private void backgroundWorkerExtraction_DoWork(object sender, DoWorkEventArgs e)
    {
      SetControlStateToExtracting();
    }

    private void backgroundWorkerExtraction_ProgressChanged(object sender, ProgressChangedEventArgs e)
    {
      if (e.UserState is ExtractBackgroundWorker.IExtractTrackProgress) {
        var progress = e.UserState as ExtractBackgroundWorker.IExtractTrackProgress;

        trackListView.UpdateExtractProgress(progress);
      }
      else if (e.UserState is ExtractBackgroundWorker.IExtractProgress) {
        var progress = e.UserState as ExtractBackgroundWorker.IExtractProgress;

        trackListView.UpdateExtractProgress(progress);

        UpdateText(string.Format("[{0}/{1}] {2}",
                                 progress.SucceededCount + progress.FailedCount,
                                 progress.TotalCount,
                                 progress.StatusText));
      }
    }

    private void backgroundWorkerExtraction_RunWorkerCompleted(object sender, RunWorkerCompletedEventArgs e)
    {
      try {
        if (e.Error != null) {
          Console.Error.WriteLine(e.Error);

          MessageBox.Show(this, ThbgmExtractor.GetText("MessageBox_ExtractionFinished_AbortedByError"), Application.ProductName, MessageBoxButtons.OK, MessageBoxIcon.Information, MessageBoxDefaultButton.Button1);
        }
        else if (e.Cancelled) {
          MessageBox.Show(this, ThbgmExtractor.GetText("MessageBox_ExtractionFinished_Aborted"), Application.ProductName, MessageBoxButtons.OK, MessageBoxIcon.Information, MessageBoxDefaultButton.Button1);
        }
        else {
          var config = e.Result as ExtractionConfig;
          var result = MessageBox.Show(this, ThbgmExtractor.GetText("MessageBox_ExtractionFinished_ConfirmOpenOutputDirectory"), Application.ProductName, MessageBoxButtons.YesNo, MessageBoxIcon.Question, MessageBoxDefaultButton.Button1);

          if (result == DialogResult.Yes) {
            try {
              FileManager.Browse(config.OutputDirectory);
            }
            catch {
              // ignore exceptions
            }
          }
        }

        trackListView.HideExtractProgressColumn();

        UpdateText();
      }
      finally {
        SetControlStateToNotExtracting();
      }
    }
    #endregion

#region "player"
    private void player_PlayerStatusChanged(object sender, EventArgs e)
    {
      if (InvokeRequired)
        Invoke((Action)UpdateText);
      else
        UpdateText();
    }

    private void player_PlayDone(object sender, EventArgs e)
    {
      if (InvokeRequired)
        Invoke((Action<bool, bool>)PlayNextTrack, true, true);
      else
        PlayNextTrack(true, true);
    }

    private void PlayNextTrack(bool playOnSetTrack, bool forward)
    {
      // TODO: 連続再生させない場合
      var tracks = forward ? trackListView.TrackItems : trackListView.TrackItems.Reverse();
      var nextTrackItem = tracks
        .SkipWhile(t => t.Track != player.Track)
        .Skip(1)
        .Where(t => t.Track.Checked) // t.Checked throws InvalidOperationException
        .FirstOrDefault();

      if (nextTrackItem == null)
        return;

      // チェックされている次のトラックを設定して再生する
      player.Track = nextTrackItem.Track;

      if (playOnSetTrack)
        player.Play();
    }

    private static readonly TimeSpan playerCueingThreshold = TimeSpan.FromSeconds(3.0);

    private void player_PrevButtonClick(object sender, EventArgs e)
    {
      if (!player.IsPlaying || player.Position <= playerCueingThreshold) {
        // 前のトラックを設定・再生
        PlayNextTrack(player.IsPlaying, false);
      }
      else {
        // 先頭から再生
        var isPlaying = player.IsPlaying;

        player.Stop();

        if (isPlaying)
          player.Play();
      }
    }

    private void player_NextButtonClick(object sender, EventArgs e)
    {
      // 次のトラックを設定・再生
      PlayNextTrack(player.IsPlaying, true);
    }
#endregion

    private void timerLog_Tick(object sender, EventArgs e)
    {
      lock (ThbgmExtractor.LogLockObject) { // StringBuilder.ToString may throw ArgumentOutOfRangeException
        if (textBoxLog.TextLength == ThbgmExtractor.Log.Length)
          return;

        textBoxLog.Text = ThbgmExtractor.Log.ToString();
      }

      textBoxLog.Select(textBoxLog.TextLength - 1, 0);
      textBoxLog.ScrollToCaret();
    }

    private bool initialized = false;
    private ExtractionConfig edittingExtractionConfig = new ExtractionConfig();
    private bool isEdittingExtractionConfigDirty = false;
    private string currentConfigFilePath = null;
    private string lastBgmSourcePath = string.Empty;
    private bool doExtractOnShown = false;
    private /*readonly*/ ToolStripItem[] editToolStripItems;
    private /*readonly*/ ToolStripItem[] trackListToolStripItems;
    private ToolStripSeparator contextMenuDummyItem = new ToolStripSeparator();
  }
}
