//  
//  Author:
//       smdn <smdn@smdn.jp>
// 
//  Copyright (c) 2004-2014 smdn
// 
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
// 
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
//  GNU General Public License for more details.
//  
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

using System;
using System.ComponentModel;
using System.Linq;

using TagLib;

using Smdn.IO;
using Smdn.Formats.Thbgm;

namespace Smdn.Applications.ThbgmExtractor {
  public class TrackTag : Tag, INotifyPropertyChanged/*, ICloneable */ {
    private static bool SequenceEqual(string[] a, string[] b)
    {
      if (a == null && b == null)
        return true;
      else if (a == null || b == null)
        return false;
      else
        return a.SequenceEqual(b);
    }

    public event PropertyChangedEventHandler PropertyChanged;

    private ITrackFormatProvider trackFormatProvider;

    public TrackInfo TrackInfo {
      get; private set;
    }

    public bool IsDirty {
      get; private set;
    }

    public override TagTypes TagTypes {
      get { return TagTypes.AllTags; } // XXX
    }

    private string outputFileName;

    public string OutputFileName {
      get { return outputFileName; }
      set
      {
        if (outputFileName == value)
          return;

        outputFileName = value;

        OnPropertyChanged(new PropertyChangedEventArgs("OutputFileName"));
      }
    }

    public string FormattedOutputFileName {
      get
      {
        var formatted = string.IsNullOrEmpty(outputFileName)
          ? TrackInfo.Title
          : TrackInfo.ToString(outputFileName, trackFormatProvider.FormatPrivider);

        return PathUtils.ReplaceInvalidFileNameCharsWithBlanks(formatted);
      }
    }

    private string title;

    public override string Title {
      get { return title; }
      set
      {
        if (title == value)
          return;

        title = value;

        OnPropertyChanged(new PropertyChangedEventArgs("Title"));
      }
    }

    private string album;

    public override string Album {
      get { return album; }
      set
      {
        if (album == value)
          return;

        album = value;

        OnPropertyChanged(new PropertyChangedEventArgs("Album"));
      }
    }

    private string[] performers;

    public override string[] Performers {
      get { return performers; }
      set
      {
        if (SequenceEqual(performers, value))
          return;

        performers = value;

        OnPropertyChanged(new PropertyChangedEventArgs("Performers"));
      }
    }

    private string[] albumArtists;

    public override string[] AlbumArtists {
      get { return albumArtists; }
      set
      {
        if (SequenceEqual(albumArtists, value))
          return;

        albumArtists = value;

        OnPropertyChanged(new PropertyChangedEventArgs("AlbumArtists"));
      }
    }

    private string[] composers;

    public override string[] Composers {
      get { return composers; }
      set
      {
        if (SequenceEqual(composers, value))
          return;

        composers = value;

        OnPropertyChanged(new PropertyChangedEventArgs("Composers"));
      }
    }

    private string[] genres;

    public override string[] Genres {
      get { return genres; }
      set
      {
        if (SequenceEqual(genres, value))
          return;

        genres = value;

        OnPropertyChanged(new PropertyChangedEventArgs("Genres"));
      }
    }

    private string comment;

    public override string Comment {
      get { return comment; }
      set
      {
        if (comment == value)
          return;

        comment = value;

        OnPropertyChanged(new PropertyChangedEventArgs("Comment"));
      }
    }

    public override uint Year {
      get { return (uint)TrackInfo.Product.ReleaseDate.Year; }
      set { throw new NotSupportedException(); }
    }

    public override uint Track {
      get { return (uint)TrackInfo.TrackNumber; }
      set { throw new NotSupportedException(); }
    }

    public override uint TrackCount {
      get { return (uint)TrackInfo.Product.Tracks.Count; }
      set { throw new NotSupportedException(); }
    }

    public override string Copyright {
      get { return string.Format("{0} {1}", TrackInfo.Product.ReleaseDate.Year, TrackInfo.Product.Creator); }
      set { throw new NotSupportedException(); }
    }

    public TrackTag(TrackInfo trackInfo, ITrackFormatProvider trackFormatProvider)
    {
      if (trackInfo == null)
        throw new ArgumentNullException("trackInfo");
      if (trackFormatProvider == null)
        throw new ArgumentNullException("trackFormatProvider");

      this.TrackInfo = trackInfo;
      this.trackFormatProvider = trackFormatProvider;

      Reset();
    }

    public virtual TrackTag Clone(ITrackFormatProvider trackFormatProvider)
    {
      var cloned = (TrackTag)MemberwiseClone();

      cloned.PropertyChanged = null;
      cloned.trackFormatProvider = trackFormatProvider;
      cloned.performers   = (string[])performers.Clone();
      cloned.albumArtists = (string[])albumArtists.Clone();
      cloned.composers    = (string[])composers.Clone();
      cloned.genres       = (string[])genres.Clone();

      return cloned;
    }

    public override void Clear()
    {
      // do nothing
    }

    public void Reset()
    {
      outputFileName = Config.TrackFormatInfoDefaults.DefaultOutputFileName;
      title = TrackInfo.Title;
      album = TrackInfo.Product.Title;
      performers    = new[] {TrackInfo.Product.Creator};
      albumArtists  = new[] {TrackInfo.Product.Creator};
      composers     = new[] {TrackInfo.Product.Creator};
      comment = string.Empty;
      genres  = EmptyArray<string>.Instance;
    }

    public void ClearDirty()
    {
      IsDirty = false;
    }

    protected virtual void OnPropertyChanged(PropertyChangedEventArgs e)
    {
      IsDirty = true;

      var ev = this.PropertyChanged;

      if (ev != null)
        ev(this, e);
    }

    internal static string GetIniFormatJoined(string[] values)
    {
      if (values == null || values.Length == 0)
        return string.Empty;
      else
        return string.Join(";", values);
    }

    internal static string[] GetIniFormatSplitted(string values)
    {
      if (values == null)
        return EmptyArray<string>.Instance;
      else
        return values.Split(';');
    }

    internal static string GetJoined(string[] values)
    {
      if (values == null || values.Length == 0)
        return string.Empty;
      else
        return string.Join("; ", values);
    }

    internal static string[] GetSplitted(string values)
    {
      if (values == null)
        return EmptyArray<string>.Instance;
      else
        return values.Split(new [] {"; "}, StringSplitOptions.None);
    }
  }
}
